import { ROUTE_URL } from 'constant/other';
import CloudContext from 'model/CloudContext';
import DataColumn from 'model/DataColumn';
import DataRecord from 'model/DataRecord';
import DataTableData from 'molecules/DataTableData';
import { useContext } from 'react';
import { ButtonGroup, Dropdown } from 'react-bootstrap';
import { useTranslation } from 'react-i18next';
import { AppContext } from 'service/state';

/**
 * Wrapper of DataTableData.
 * @param dataColumn Column data of the element type.
 * @param dataRecord Data record of the element.
 * @param detailInfo Element type information.
 */
const DataTableDataWrapper = ({ dataColumn, dataRecord, detailInfo }: {
  dataColumn: DataColumn,
  dataRecord: DataRecord,
  detailInfo?: {
    column: string,
    path: string,
  },
}) => {
  let link = undefined;
  if (dataColumn.key === detailInfo?.column) {
    const temp = detailInfo.path.split('/');
    link = `/${temp[1]}/${dataRecord.cloudContext}/${temp[2]}/${dataRecord.id}`;
  }

  return <DataTableData key={dataColumn.key} text={
    dataColumn.key in dataRecord.value
      ? dataRecord.value[dataColumn.key]
      : ''
  } link={link} />;
}

/**
 * Find and return the list of items to be added to Operation links.
 *
 * @param dataRecord Record of data.
 * @returns The list of items.
 */
const getOperations = (dataRecord: DataRecord): string[] => {
  switch (dataRecord.entityTypeId) {
    case 'aws_cloud_instance': {
      // TODO: support 'Associate Elastic IP' menu.
      const list: string[] = ['Create Image'];
      if (dataRecord.value['instance_state'] === 'running') {
        list.push('Stop');
        list.push('Reboot');
      } else if (dataRecord.value['instance_state'] === 'stopped') {
        list.push('Start');
      }
      list.push('Associate Elastic IP');
      return list;
    }
    case 'aws_cloud_volume': {
      const list: string[] = ['Create Snapshot'];
      if (dataRecord.value['state'] === 'available') {
        list.push('Attach');
      } else if (dataRecord.value['state'] === 'in-use') {
        list.push('Detach');
      }
      return list;
    }
    case 'aws_cloud_elastic_ip':
      return dataRecord.value['association_id'] === ''
        || dataRecord.value['association_id'] === null
        || dataRecord.value['association_id'] === undefined
        ? ['Associate']
        : ['Disassociate'];
    case 'aws_cloud_snapshot':
      return ['Create Volume'];
    case 'aws_cloud_security_group':
      return ['Copy'];
    case 'aws_cloud_vpc_peering_connection':
      return dataRecord.value['status_code'] === 'pending-acceptance'
        ? ['Accept'] : [];
    case 'aws_cloud_internet_gateway':
      return dataRecord.value['vpc_id'] === ''
        || dataRecord.value['vpc_id'] === null
        ? ['Attach']
        : ['Detach'];
    case 'k8s_deployment':
      return ['Scale'];
    case 'cloud_launch_template': {
      const list: string[] = [];
      if (dataRecord.value['field_workflow_status'] === 'Approved') {
        list.push('Launch');
      }
      list.push('Copy');
      if (dataRecord.value['field_workflow_status'] === 'Review') {
        list.push('Approve');
      }
      if (dataRecord.value['field_workflow_status'] === 'Draft') {
        list.push('Review');
      }
      return list;
    }
    case 'cloud_project':
      return ['Launch', 'Copy'];
    case 'k8s_cost_store':
      return ['Copy'];
    case 'k8s_namespace_resource_store':
      return ['Copy'];
    case 'k8s_node_resource_store':
      return ['Copy'];
    case 'k8s_pod_resource_store':
      return ['Copy'];
    case 'openstack_instance': {
      const list: string[] = [];
      if (dataRecord.value['instance_state'] === 'running') {
        list.push('Stop');
        list.push('Reboot');
      } else if (dataRecord.value['instance_state'] === 'stopped') {
        list.push('Start');
      }
      return list;
    }
    case 'openstack_floating_ip':
      return dataRecord.value['association_id'] === ''
        || dataRecord.value['association_id'] === null
        || dataRecord.value['association_id'] === undefined
        ? ['Associate']
        : ['Disassociate'];
    case 'openstack_volume':
      return dataRecord.value['attachment_information'] === ''
        || dataRecord.value['attachment_information'] === null
        || dataRecord.value['attachment_information'] === undefined
        ? ['Attach']
        : ['Detach'];
    case 'vmware_vm':
      const list: string[] = [];
      if (dataRecord.value['power_state'] === 'POWERED_OFF') {
        list.push('Start');
      }
      if (dataRecord.value['power_state'] === 'POWERED_ON') {
        list.push('Stop');
        list.push('Suspend');
        list.push('Reboot');
      }
      if (dataRecord.value['power_state'] === 'SUSPENDED') {
        list.push('Start');
        list.push('Stop');
      }
      return list;
  }
  return [];
}

/**
 * Action logic of updating and deleting.
 *
 * @param dataRecord Data record of the element.
 * @param action Type of CxUD.
 * @param cloudContextList CloudContext list.
 */
const updateAndDeleteData = (dataRecord: DataRecord, action: string, cloudContextList: CloudContext[]) => {
  let urlParts = dataRecord.entityTypeId;
  const modules = ['aws_cloud', 'k8s', 'openstack', 'vmware'];
  for (const module of modules) {
    urlParts = urlParts.replace(`${module}_`, `${module}/${dataRecord.cloudContext}/`);
  }
  if (dataRecord.entityTypeId === 'cloud_launch_template') {
    const moduleList = cloudContextList.filter((c) => {
      return c.name === dataRecord.cloudContext;
    });
    if (moduleList.length >= 1) {
      urlParts = urlParts.replace(`${dataRecord.entityTypeId}`, `${moduleList[0].cloudServiceProvider}/${dataRecord.cloudContext}/${dataRecord.entityTypeId}`);
    }
  }
  window.location.href = `${ROUTE_URL}/${urlParts}/${dataRecord.id}/${action}`;
}

/**
 * Row data of DataTable.
 * Note: The OperationLinks column is added.
 *
 * @param dataRecord Record of data.
 * @param dataColumn List of DataColumn.
 * @param className Parameter of className.
 * @param detailInfo Information required to create a link to more information.
*/
const DataTableRowWithOperationLinks = ({ dataRecord, dataColumnList, className, detailInfo }: {
  dataRecord: DataRecord,
  dataColumnList: DataColumn[],
  className?: string,
  detailInfo?: {
    column: string,
    path: string,
  },
}) => {

  const { cloudContextList } = useContext(AppContext);
  const { t } = useTranslation();

  if (['k8s_node', 'k8s_event'].includes(dataRecord.entityTypeId)) {
    return <tr key={dataRecord.id} className={className}>
      {
        dataColumnList.map((dataColumn) => {
          return <DataTableDataWrapper
            dataColumn={dataColumn}
            dataRecord={dataRecord}
            detailInfo={detailInfo} />;
        })
      }
    </tr>;
  }

  return <tr key={dataRecord.id} className={className}>
    {
      dataColumnList.map((dataColumn) => {
        return <DataTableDataWrapper
          dataColumn={dataColumn}
          dataRecord={dataRecord}
          detailInfo={detailInfo} />;
      })
    }
    <td>
      <div className="dropbutton-wrapper dropbutton-multiple">
        <div className="dropbutton-widget">
          <Dropdown as={ButtonGroup} className="custom-dropdown">
            <button type="button" className="dropdown-toggle links ripple-effect" onClick={(e) => {
              e.preventDefault();
              updateAndDeleteData(dataRecord, 'edit', cloudContextList);
            }}>{t('Edit')}</button>

            <button type="button" className="uparrow" data-bs-toggle="dropdown" aria-expanded="false">
              <span className="caret"></span>
            </button>

            <ul data-drupal-selector="edit-entities-1-operations-data"
              className="dropbutton dropdown-menu"
              style={{
                position: 'absolute',
                inset: '0px auto auto 0px',
                margin: 0,
                transform: 'translate3d(1424px, 444px, 0px)',
              }}
              data-popper-placement="bottom-start"
            >
              <li className="dropbutton-action">
                {/* eslint-disable-next-line  */}
                <a href="#" onClick={(e) => {
                  e.preventDefault();
                  updateAndDeleteData(dataRecord, 'delete', cloudContextList);
                }}>
                  <span className="delete glyphicon glyphicon-"></span>
                  {t('Delete')}
                </a>
              </li>
              <li className="dropbutton-toggle">
                <button type="button"></button>
              </li>

              {
                getOperations(dataRecord).map((menuInfo) => {
                  return <li className="dropbutton-action secondary-action" key={menuInfo}>
                    {/* eslint-disable-next-line  */}
                    <a href="#" onClick={(e) => {
                      e.preventDefault();
                      switch (menuInfo) {
                        // By default, the name of the operation, lowercased as it is,
                        // becomes part of the URL in CxUD. If it does not match,
                        // set it individually with a switch-case statement.
                        case 'Start':
                          updateAndDeleteData(dataRecord, 'start', cloudContextList);
                          break;
                        default:
                          updateAndDeleteData(dataRecord, menuInfo.toLowerCase().replaceAll(' ', '_'), cloudContextList);
                          break;
                      }
                    }}>
                      {t(menuInfo)}
                    </a>
                  </li>;
                })
              }
            </ul>
          </Dropdown>
        </div>
      </div>
    </td>
  </tr>;

}

export default DataTableRowWithOperationLinks;
