import useDrupalJsonApi, { GetJsonDataType } from 'hooks/drupal_jsonapi';
import CloudContextItem from 'model/CloudContextItem';
import { MapData } from 'model/MapData';
import RawCloudContextItem from 'model/RawCloudContextItem';
import CloudContextItemPopup from 'molecules/CloudContextItemPopup';
import LoadingSpinner from 'molecules/LoadingSpinner';
import MapPolygonLayer from 'molecules/MapPolygonLayer';
import { useEffect, useState } from 'react';
import { MapContainer } from 'react-leaflet';
import { convertCloudContextItemList } from 'service/utility';

/**
 * Load List of CloudContextItem.
 */
const LoadCloudContextItemList = async (getJsonData: GetJsonDataType) => {
  const response = await fetch('/clouds/cloud_dashboard/config/marker_icon_uri');
  if (!response.ok) {
    return [];
  }
  const resJson = await response.json();
  const defaultIconUri: string = resJson['uri'];
  const rawCloudContextItemList = await getJsonData<RawCloudContextItem[]>('/clouds/cloud_config_location');
  const cloudContextItemList = convertCloudContextItemList(rawCloudContextItemList, defaultIconUri);
  return cloudContextItemList;
};

/**
 * Load MapData.
 */
const LoadMapData = async (
  getJsonData: GetJsonDataType
) => {
  const response = await fetch('/clouds/cloud_dashboard/config/map_geojson_uri');
  if (!response.ok) {
    return {
      features: []
    };
  }
  const responseJson = await response.json();
  return await getJsonData<MapData>(responseJson['uri']);
};

/**
 * Map for cloud service providers.
 */
const CloudServiceProviderMap = () => {

  const [cloudContextItemList, setCloudContextItemList] = useState<CloudContextItem[]>([]);
  const [mapData, setMapData] = useState<MapData>({
    features: []
  });
  const { getJsonData } = useDrupalJsonApi();

  useEffect(() => {
    const init = async () => {
      // Load CloudContextItemList.
      setCloudContextItemList(await LoadCloudContextItemList(getJsonData));

      // Load MapData.
      setMapData(await LoadMapData(getJsonData));
    }

    init();
    // eslint-disable-next-line react-hooks/exhaustive-deps
  }, []);

  return mapData.features.length === 0
    ? <LoadingSpinner />
    : <MapContainer
      center={[0, 0]}
      zoom={2}
      scrollWheelZoom={false}
      style={{ height: 500, backgroundColor: '#4e5d6c' }}>
      <MapPolygonLayer mapData={mapData} />
      {
        cloudContextItemList.map((item, index) => {
          return <CloudContextItemPopup item={item} key={index} />;
        })
      }
    </MapContainer>;

};

export default CloudServiceProviderMap;
