import PanelWrapper from 'atoms/PanelWrapper';
import EntityFormColumn from 'model/EntityFormColumn';
import EntityFormRecord from 'model/EntityFormRecord';
import EntityFormBlock from 'molecules/EntityFormBlock';
import { calcPanelName } from 'service/utility';

/**
 * Determines whether or not to display the form.
 *
 * Some pages require fine control over the display of forms.
 * Therefore, this function determines whether or not to display the form.
 * @param cloudServiceProvider The cloud service provider.
 * @param entityName The entity's type name.
 * @param actionType The CxUD action Type.
 * @param entityFormColumn The entity form column.
 * @param formData The data of form values.
 *
 * @returns If TRUE, the block is displayed.
 */
const showEntityFormBlockFlg = (
  cloudServiceProvider: string,
  entityName: string,
  actionType: string,
  entityFormColumn: EntityFormColumn,
  formData: Record<string, any>
) => {
  // Associate AWS Cloud Elastic IP
  if (actionType === 'associate'
    && cloudServiceProvider === 'aws_cloud'
    && entityName === 'elastic_ip') {
    if (entityFormColumn.type === 'select') {
      switch (entityFormColumn.name) {
        case 'instance_id':
          return formData['resource_type'] === 'instance';
        case 'instance_private_ip':
          return formData['resource_type'] === 'instance';
        case 'network_interface_id':
          return formData['resource_type'] === 'network_interface';
        case 'network_interface_private_ip':
          return formData['resource_type'] === 'network_interface';
      }
    }
  }

  // Create K8s deployment
  if (actionType === 'create'
    && cloudServiceProvider === 'k8s'
    && entityName === 'deployment') {
    if (entityFormColumn.labelName === 'Detail') {
      return formData['time_scheduler_option'] === 'cloud_orchestrator_scheduler';
    }
    if (entityFormColumn.labelName === 'YAML URL') {
      return formData['time_scheduler_option'] === 'cronjob_scheduler';
    }
    if (['Scheduler Use Type', 'Start-up Time', 'Stop Time'].includes(entityFormColumn.labelName)) {
      return formData['enableTimeScheduler'];
    }
  }

  // Create K8s pod
  if (actionType === 'create'
    && cloudServiceProvider === 'k8s'
    && entityName === 'pod') {
    if (entityFormColumn.labelName === 'Detail') {
      return formData['time_scheduler_option'] === 'cloud_orchestrator_scheduler';
    }
    if (entityFormColumn.labelName === 'YAML URL') {
      return formData['time_scheduler_option'] === 'cronjob_scheduler';
    }
    if (['Scheduler Use Type', 'Start-up Time', 'Stop Time'].includes(entityFormColumn.labelName)) {
      return formData['enableTimeScheduler'];
    }
  }

  // Associate OpenStack Floating IP
  if (actionType === 'associate'
    && cloudServiceProvider === 'openstack'
    && entityName === 'floating_ip') {
    if (entityFormColumn.type === 'select') {
      switch (entityFormColumn.name) {
        case 'instance_id':
          return formData['resource_type'] === 'instance';
        case 'instance_private_ip':
          return formData['resource_type'] === 'instance';
        case 'network_interface_id':
          return formData['resource_type'] === 'network_interface';
        case 'network_interface_private_ip':
          return formData['resource_type'] === 'network_interface';
      }
    }
  }

  // Create AWS Cloud VPC Peering Connection
  if (actionType === 'create'
    && cloudServiceProvider === 'aws_cloud'
    && entityName === 'vpc_peering_connection') {

    if (entityFormColumn.type === 'default'
      || entityFormColumn.type === 'select'
      || entityFormColumn.type === 'hidden') {

      // If the regions and accound_ids are different, display the textfield
      // `accepter_vpc_id_text` so users can enter a freeform VPC Id belonging
      // to a different region.
      if (entityFormColumn.name === 'accepter_vpc_id') {
        if (formData['accepter_region'] !== formData['default_region'] ||
          formData['accepter_account_id'] !== formData['default_account_id']
        ) {
          formData['accepter_vpc_id'] = '';
          return false;
        }
      }

      if (entityFormColumn.name === 'accepter_vpc_id_text') {
        if (formData['accepter_region'] === formData['default_region'] &&
          formData['accepter_account_id'] === formData['default_account_id']
        ) {
          formData['accepter_vpc_id_text'] = '';
          return false;
        }
      }
    }
  }

  if (actionType === 'edit'
    && cloudServiceProvider === 'aws_cloud'
    && entityName === 'transit_gateway'
  ) {
    if (entityFormColumn.type === 'boolean' || entityFormColumn.type === 'select') {
      if (entityFormColumn.name === 'association_default_route_table_id') {
        return formData['default_route_table_association'];
      }
      if (entityFormColumn.name === 'propagation_default_route_table_id') {
        return formData['default_route_table_propagation'];
      }
    }
  }

  return true;
}

const EntityFormPanel = ({
  cloudServiceProvider,
  cloudContext,
  entityName,
  actionType,
  entityRecord,
  formData,
  setFormData
}: {
  cloudServiceProvider: string,
  cloudContext: string,
  entityName: string,
  actionType: string,
  entityRecord: EntityFormRecord,
  formData: Record<string, any>,
  setFormData: (f: Record<string, any>) => void,
}) => {
  if (entityRecord.type !== 'panel') {
    return <></>;
  }

  return <PanelWrapper title={calcPanelName(entityRecord.panelName, entityName)}>
    {
      entityRecord.keyValueRecords.map((keyValueRecord, index) => {
        return (showEntityFormBlockFlg(
          cloudServiceProvider,
          entityName,
          actionType,
          keyValueRecord,
          formData
        ))
          ? <EntityFormBlock key={index}
            keyValueRecord={keyValueRecord}
            cloudContext={cloudContext}
            formData={formData}
            setFormData={setFormData} />
          : <></>;
      })
    }
  </PanelWrapper>;
}

export default EntityFormPanel;
