import ENTITY_FORM_LIST from 'constant/form_template/entity_form_template';
import useDrupalJsonApi from 'hooks/drupal_jsonapi';
import EntityFormTemplate from 'model/EntityFormTemplate';
import EntityFormLabel from 'molecules/EntityFormLabel';
import EntityFormPanel from 'organisms/EntityFormPanel';
import SideBar from 'organisms/SideBar';
import { useEffect, useState } from 'react';
import { Button, Col, Container, Form, Row } from 'react-bootstrap';
import { useTranslation } from 'react-i18next';
import { BsCheckLg } from 'react-icons/bs';
import { useHistory, useParams } from 'react-router-dom';
import { formDataToSendData } from 'service/utility';

const EntityCreatePageImpl = ({ cloudContext, entityFormTemplate, entityName, action }: {
  cloudContext: string,
  entityFormTemplate: EntityFormTemplate,
  entityName: string,
  action: string,
}) => {
  const [formData, setFormData] = useState<Record<string, any>>({});
  const { t } = useTranslation();
  const { getJsonData } = useDrupalJsonApi();
  const history = useHistory();

  // Initialize form data.
  useEffect(() => {
    initForm();
    // eslint-disable-next-line react-hooks/exhaustive-deps
  }, []);

  useEffect(() => {
    console.log('FormData:');
    console.log(formData);
  }, [formData]);

  const initForm = async () => {
    // Create a list of key strings and default values.
    const formInfoList: { key: string, type: string, defaultValue: any }[] = [];
    for (const entityRecord of entityFormTemplate.entityRecords) {
      // Process only forms contained in Panel.
      if (entityRecord.type !== 'panel') {
        continue;
      }

      // Load the form inside each Panel.
      for (const keyValueRecord of entityRecord.keyValueRecords) {
        if (keyValueRecord.type === 'time') {
          formInfoList.push({ key: keyValueRecord.hourName, type: keyValueRecord.type, defaultValue: keyValueRecord.defaultValue[0] });
          formInfoList.push({ key: keyValueRecord.minuteName, type: keyValueRecord.type, defaultValue: keyValueRecord.defaultValue[1] });
          continue;
        }

        formInfoList.push({
          key: keyValueRecord.name,
          type: keyValueRecord.type,
          defaultValue: keyValueRecord.type === 'file'
            ? null
            : 'defaultValue' in keyValueRecord
              ? keyValueRecord.defaultValue
              : undefined
        });
      }
    }

    // Load entity data.
    const newFormData: Record<string, any> = {};
    for (const formInfo of formInfoList) {
      // Handling of types that are not entered in the form, such as 'cost' types.
      if (formInfo.defaultValue === undefined) {
        continue;
      }

      newFormData[formInfo.key] = formInfo.defaultValue;
    }
    for (const entityRecord of entityFormTemplate.entityRecords) {
      if (entityRecord.type !== 'panel') {
        continue;
      }
      for (const keyValueRecord of entityRecord.keyValueRecords) {
        if (keyValueRecord.type === 'default'
          || keyValueRecord.type === 'textarea'
          || keyValueRecord.type === 'hidden'
          || keyValueRecord.type === 'select') {
          if (keyValueRecord.defaultValueUrl !== undefined) {
            const replacedUrl = keyValueRecord.defaultValueUrl.replace('{cloud_context}', cloudContext);
            if (!replacedUrl.endsWith('/')) {
              const jsonData = await getJsonData<string>(replacedUrl);
              if (jsonData.length > 0) {
                newFormData[keyValueRecord.name] = jsonData;
              }
            }
          }
        }
      }
    }

    setFormData(newFormData);
  }

  /**
   * Save entiy's data.
   */
  const save = async () => {
    const url = `/cloud_dashboard/${entityFormTemplate.cloudServiceProvider}/${cloudContext}/${entityFormTemplate.cloudServiceProvider}_${entityName}/${action}`;

    const response = await fetch(url, {
      method: 'POST',
      body: formDataToSendData(formData, entityFormTemplate),
    });
    if (!response.ok) {
      console.log('Create Request : No');
      console.error('Create failed.');
      console.error(await response.json());
      window.alert('Create failed.');
    } else {
      console.log('Create Request : Yes');
      console.log(await response.json());

      history.push(`/${entityFormTemplate.cloudServiceProvider}/${cloudContext}/${entityName}`);
    }
  }

  return <>
    <SideBar />
    <Container fluid className="px-0">
      <Row className="mx-0">
        <Col>
          <Form>
            {
              entityFormTemplate.entityRecords.map((entityRecord) => {
                switch (entityRecord.type) {
                  case 'panel':
                    return <EntityFormPanel
                      cloudServiceProvider={entityFormTemplate.cloudServiceProvider}
                      cloudContext={cloudContext}
                      entityName={entityName}
                      actionType={entityFormTemplate.actionType}
                      entityRecord={entityRecord}
                      formData={formData}
                      setFormData={setFormData} />;
                  case 'label':
                    return <EntityFormLabel label={entityRecord.text} entityName={entityName} />;
                  default:
                    return <></>;
                }
              })
            }
            <Button className="ripple-effect" onClick={save}>
              <BsCheckLg />{' ' + t(
                entityFormTemplate.submitButtonLabel
                  ? entityFormTemplate.submitButtonLabel
                  : t('Save')
              )}
            </Button>
          </Form>
        </Col>
      </Row>
    </Container>
  </>;
}

const EntityCreatePage = ({ action }: {
  action: string
}) => {
  const params: {
    cloudServiceProvider: string,
    cloudContext: string,
    entityName: string,
  } = useParams();

  const entityFormTemplate = ENTITY_FORM_LIST.filter((template) => {
    return template.cloudServiceProvider === params.cloudServiceProvider
      && (template.entityName === params.entityName || template.entityName === '')
      && template.actionType === action
  });

  if (entityFormTemplate.length >= 1) {
    return <EntityCreatePageImpl
      cloudContext={params.cloudContext}
      entityFormTemplate={entityFormTemplate[0]}
      entityName={params.entityName}
      action={action} />;
  }

  // Default Page
  return <>
    <SideBar />
    <Container fluid className="px-0">
      <Row>
        <Col>
          <span>EntityCxudPage</span><br />
          <span>cloudServiceProvider={params.cloudServiceProvider}</span><br />
          <span>cloudContext={params.cloudContext}</span><br />
          <span>entityName={params.entityName}</span><br />
          <span>action=create</span>
        </Col>
      </Row>
    </Container>
  </>;
}

export default EntityCreatePage;
