import ENTITY_INFO_LIST from 'constant/entity_info_template/entity_info_template';
import useDrupalJsonApi, { GetEntityListAllType, GetJsonDataType } from 'hooks/drupal_jsonapi';
import EntityColumn from 'model/EntityColumn';
import EntityData from 'model/EntityData';
import EntityInfoPanelData from 'model/EntityInfoPanelData';
import EntityInfoRecord from 'model/EntityInfoRecord';
import EntityInfoRecordData from 'model/EntityInfoRecordData';
import EntityInfoTemplate from 'model/EntityInfoTemplate';
import LoadingSpinner from 'molecules/LoadingSpinner';
import EntityInfoPanel from 'organisms/EntityInfoPanel';
import SideBar from 'organisms/SideBar';
import { useEffect, useState } from 'react';
import { Col, Container, Row } from 'react-bootstrap';
import { useParams } from 'react-router-dom';
import { convertDataForUI, getReplacedUrl, readDataCache } from 'service/utility';

const getEntityInfoRecordData = async (
  getJsonData: GetJsonDataType,
  dataCache: Record<string, EntityData[]>,
  tableRecordList: string[],
  cloudServiceProvider: string,
  entityName: string,
  entityColumn: EntityColumn,
  entityData: EntityData
) => {
  switch (entityColumn.type) {
    case 'metrics': {
      const url = `/clouds/${cloudServiceProvider}/${entityData.attributes['cloud_context']}/${entityName}/${entityData.attributes['drupal_internal__id']}/metrics`;
      const metricsData = await getJsonData<Record<string, number>[]>(url);
      return {
        type: 'metrics',
        record: entityColumn.column.map((column) => {
          return {
            title: column.title,
            yLabel: column.yLabel,
            record: metricsData.map((metrics) => {
              return { x: metrics['timestamp'], y: metrics[column.name] };
            })
          }
        })
      } as EntityInfoRecordData;
    }
    case 'custom-table': {
      const keyVal: Record<string, string>[] = [];
      entityData.attributes[entityColumn.name].forEach((record: any, index: number) => {
        const temp: Record<string, string> = {};
        for (const column of entityColumn.column) {
          const convertedText = convertDataForUI(
            record[column.name],
            column,
            dataCache
          );
          const convertedText2 = column.type === 'link'
            ? getReplacedUrl(
                convertedText,
                entityData.attributes['cloud_context'],
                entityName,
                entityData.attributes['drupal_internal__id'],
                cloudServiceProvider,
                index
              )
            : convertedText;
          temp[column.name] = convertedText2;
        }
        keyVal.push(temp);
      });
      return {
        type: 'table',
        title: entityColumn.labelName,
        record: keyVal
      } as EntityInfoRecordData;
    }
    case 'json-table': {
      const entityColumnList = entityColumn.column;
      const keyVal: Record<string, string>[] = [];

      for (const record of entityData.attributes[entityColumn.name]) {
        const record2: {
          item_key: string,
          item_value: string
        } = record;
        const record3: Record<string, any> = JSON.parse(record2.item_value);

        const temp: Record<string, string> = {};
        for (const entityColumn of entityColumnList) {
          if (entityColumn.name in record3) {
            const convertedText = convertDataForUI(
              record3[entityColumn.name],
              entityColumn,
              dataCache
            );
            temp[entityColumn.name] = convertedText;
          }
        }
        keyVal.push(temp);
      }
      return {
        type: 'table',
        title: entityColumn.labelName,
        record: keyVal
      } as EntityInfoRecordData;
    }
    default:
      if (tableRecordList.includes(entityColumn.name)) {
        const keyVal: Record<string, string>[] = [];
        for (const record of entityData.attributes[entityColumn.name]) {
          const temp: Record<string, string> = {};
          for (const key of Object.keys((record))) {
            temp[key] = record[key] ? `${record[key].replace(/(\r\n)|\n/, '')}` : '';
          }
          keyVal.push(temp);
        }
        return {
          type: 'table',
          title: entityColumn.labelName,
          record: keyVal
        } as EntityInfoRecordData;
      } else {
        const convertedText = convertDataForUI(
          entityData.attributes[entityColumn.name],
          entityColumn,
          dataCache
        );
        const convertedText2 = entityColumn.type === 'join' && convertedText !== entityData.attributes[entityColumn.name]
          ? `${convertedText} (${entityData.attributes[entityColumn.name]})`
          : convertedText;
        return {
          type: 'div',
          key: entityColumn.labelName,
          value: convertedText2
        } as EntityInfoRecordData;
      }
  }
}

const getPanelData = async (
  getEntityListAll: GetEntityListAllType,
  getJsonData: GetJsonDataType,
  entityInfoRecord: EntityInfoRecord,
  cloudServiceProvider: string,
  entityName: string,
  entityData: EntityData
) => {
  const newPanelData: EntityInfoPanelData = {
    title: entityInfoRecord.panelName,
    records: []
  };

  const dataCache = await (readDataCache(getEntityListAll, entityInfoRecord.keyValueRecords));
  for (const entityColumn of entityInfoRecord.keyValueRecords) {
    newPanelData.records.push(await getEntityInfoRecordData(
      getJsonData,
      dataCache,
      entityInfoRecord.tableRecordList,
      cloudServiceProvider,
      entityName,
      entityColumn,
      entityData
    ));
  }

  return newPanelData;
}

const getPanelDataList = async (
  getEntityListAll: GetEntityListAllType,
  getJsonData: GetJsonDataType,
  entityInfoTemplate: EntityInfoTemplate,
  entityId: string
) => {
  const entityTypeId = `${entityInfoTemplate.cloudServiceProvider}_${entityInfoTemplate.entityName}`;
  const url = `/jsonapi/${entityTypeId}/${entityTypeId}?filter[drupal_internal__id]=${entityId}`;
  const entityData = await (await getJsonData<{ data: EntityData[] }>(url)).data[0];

  const panelDataList: EntityInfoPanelData[] = [];

  for (const entityInfoRecord of entityInfoTemplate.entityRecords) {
    panelDataList.push(await getPanelData(
      getEntityListAll,
      getJsonData,
      entityInfoRecord,
      entityInfoTemplate.cloudServiceProvider,
      entityInfoTemplate.entityName,
      entityData
    ));
  }

  return panelDataList;
}

const EntityDetailPage = () => {
  const params: {
    cloudServiceProvider: string,
    cloudContext: string,
    entityName: string,
    entityId: string
  } = useParams();

  const { getEntityListAll, getJsonData } = useDrupalJsonApi();
  const [panelDataList, setPanelDataList] = useState<EntityInfoPanelData[]>([]);
  const [loadingFlg, setLoadingFlg] = useState(true);

  useEffect(() => {
    setLoadingFlg(true);
    const entityInfoTemplate = ENTITY_INFO_LIST.filter((e) => {
      return e.cloudServiceProvider === params.cloudServiceProvider
        && e.entityName === params.entityName;
    })[0];

    getPanelDataList(getEntityListAll, getJsonData, entityInfoTemplate, params.entityId).then((data) => {
      setPanelDataList(data);
      setLoadingFlg(false);
    });
    // eslint-disable-next-line react-hooks/exhaustive-deps
  }, [params]);

  return <>
  <SideBar />
  <Container fluid className="px-0">
    <Row className="mx-0">
      <Col>
        {
          loadingFlg
            ? <LoadingSpinner />
            : panelDataList.map((panelData, index) => {
              return <EntityInfoPanel index={index} key={index} panelData={panelData} />;
            })
        }
      </Col>
    </Row>
  </Container>
  </>;
}

export default EntityDetailPage;
