import ENTITY_FORM_LIST from 'constant/form_template/entity_form_template';
import useDrupalJsonApi from 'hooks/drupal_jsonapi';
import EntityData from 'model/EntityData';
import EntityFormTemplate from 'model/EntityFormTemplate';
import EntityFormLabel from 'molecules/EntityFormLabel';
import EntityFormPanel from 'organisms/EntityFormPanel';
import SideBar from 'organisms/SideBar';
import LoadingDataPage from 'pages/LoadingDataPage';
import { useEffect, useState } from 'react';
import { Button, Col, Container, Form, Row } from 'react-bootstrap';
import { useTranslation } from 'react-i18next';
import { BsCheckLg } from 'react-icons/bs';
import { useHistory, useParams } from 'react-router-dom';
import { capitalize, formDataToSendData } from 'service/utility';

const EntityXxudPageImpl = ({ cloudContext, entityFormTemplate, entityName, entityId }: {
  cloudContext: string,
  entityFormTemplate: EntityFormTemplate,
  entityName: string,
  entityId: string
}) => {
  const [formData, setFormData] = useState<Record<string, any>>({});
  const [loadingFlg, setLoadingFlg] = useState(true);
  const [entityData, setEntityData] = useState<EntityData | undefined>();
  const { getJsonData } = useDrupalJsonApi();
  const { t } = useTranslation();
  const history = useHistory();

  // Initialize form data.
  useEffect(() => {
    setLoadingFlg(true);
    initEntity().then(() => {
      setLoadingFlg(false);
    });
    // eslint-disable-next-line react-hooks/exhaustive-deps
  }, []);

  useEffect(() => {
    if (entityData !== undefined) {
      initForm(entityData);
    }
    // eslint-disable-next-line react-hooks/exhaustive-deps
  }, [entityData]);

  useEffect(() => {
    console.log('FormData:');
    console.log(formData);
  }, [formData]);

  const initEntity = async () => {
    let url = '';
    if (entityName !== "cloud_launch_template") {
      const entityTypeId = `${entityFormTemplate.cloudServiceProvider}_${entityName}`;
      url = `/jsonapi/${entityTypeId}/${entityTypeId}?filter[drupal_internal__id]=${entityId}`;
    } else {
      url = `/jsonapi/cloud_launch_template/${entityFormTemplate.cloudServiceProvider}?filter[drupal_internal__id]=${entityId}`;
    }
    const rawEntityData = await getJsonData<{ data: EntityData[] }>(url);
    if (rawEntityData.data.length >= 1) {
      setEntityData(rawEntityData.data[0]);
    }
  }

  const initForm = async (entityData: EntityData) => {
    // Create a list of key strings and default values.
    const formInfoList: { key: string, type: string, defaultValue: any }[] = [];
    for (const entityRecord of entityFormTemplate.entityRecords) {
      // Process only forms contained in Panel.
      if (entityRecord.type !== 'panel') {
        continue;
      }

      // Load the form inside each Panel.
      for (const keyValueRecord of entityRecord.keyValueRecords) {
        if (keyValueRecord.type === 'time') {
          formInfoList.push({ key: keyValueRecord.hourName, type: keyValueRecord.type, defaultValue: keyValueRecord.defaultValue[0] });
          formInfoList.push({ key: keyValueRecord.minuteName, type: keyValueRecord.type, defaultValue: keyValueRecord.defaultValue[1] });
          continue;
        }

        // Load default value.
        let defaultValue = keyValueRecord.type === 'file'
          ? null
          : 'defaultValue' in keyValueRecord
            ? keyValueRecord.defaultValue
            : undefined;
        if (keyValueRecord.type === 'default'
          || keyValueRecord.type === 'textarea'
          || keyValueRecord.type === 'hidden'
          || keyValueRecord.type === 'select') {
          if (keyValueRecord.defaultValueUrl !== undefined) {
            const replacedUrl = keyValueRecord.defaultValueUrl
              .replace('{cloud_context}', cloudContext)
              .replace('{entity_id}', entityId);
            if (!replacedUrl.endsWith('/')) {
              const jsonData = await getJsonData<string>(replacedUrl);
              if (jsonData.length > 0) {
                defaultValue = jsonData;
              }
            }
          }
        }

        formInfoList.push({
          key: keyValueRecord.name,
          type: keyValueRecord.type,
          defaultValue,
        });
      }
    }

    // Load entity data.
    const attributes = entityData.attributes;
    const newFormData: Record<string, any> = {};
    console.log('attributes:');
    console.log(attributes);
    console.log('formInfoList:');
    console.log(formInfoList);
    for (const formInfo of formInfoList) {
      // Handling of types that are not entered in the form, such as 'cost' types.
      if (formInfo.defaultValue === undefined) {
        continue;
      }

      // Processing when the value corresponding to a key cannot be read.
      if (!(formInfo.key in attributes)) {
        newFormData[formInfo.key] = formInfo.defaultValue;
        continue;
      }
      const value = attributes[formInfo.key];
      if (value === null || (Array.isArray(value) && value.length === 0)) {
        newFormData[formInfo.key] = formInfo.defaultValue;
        continue;
      }


      // Processing in 'multi-select' type forms.
      if (formInfo.type === 'multi-select' || formInfo.type === 'array') {
        newFormData[formInfo.key] = typeof value === 'string'
          ? [value] : value;
          continue;
      }

      // Processing in 'radio' type forms.
      if (typeof value === 'boolean' && formInfo.type === 'radio') {
        newFormData[formInfo.key] = value ? '1' : '0';
        continue;
      }

      // Other type.
      newFormData[formInfo.key] = value;
    }

    console.log('newFormData:');
    console.log(newFormData);
    setFormData(newFormData);
  }

  const save = async () => {
    const url = entityName !== 'cloud_launch_template'
      ? `/cloud_dashboard/${entityFormTemplate.cloudServiceProvider}/${cloudContext}/${entityFormTemplate.cloudServiceProvider}_${entityName}/${entityId}/${entityFormTemplate.actionType}`
      : `/cloud_dashboard/${entityFormTemplate.cloudServiceProvider}/${cloudContext}/${entityName}/${entityId}/${entityFormTemplate.actionType}`;

    const response = await fetch(url, {
      method: 'POST',
      body: formDataToSendData(formData, entityFormTemplate),
    });
    if (!response.ok) {
      console.log('Create Request : No');
      console.error(capitalize(entityFormTemplate.actionType) + ' failed.');
      console.error(await response.json());
      window.alert(capitalize(entityFormTemplate.actionType) + ' failed.');
    } else {
      console.log('Create Request : Yes');
      console.log(await response.json());

      history.push(`/${entityFormTemplate.cloudServiceProvider}/${cloudContext}/${entityName}`);
    }
  }

  return loadingFlg
    ? <LoadingDataPage />
    : <>
      <SideBar />
      <Container fluid className="px-0">
        <Row className="mx-0">
          <Col>
            <Form>
              {
                entityFormTemplate.entityRecords.map((entityRecord) => {
                  switch (entityRecord.type) {
                    case 'panel':
                      return <EntityFormPanel
                        cloudServiceProvider={entityFormTemplate.cloudServiceProvider}
                        cloudContext={cloudContext}
                        entityName={entityName}
                        actionType={entityFormTemplate.actionType}
                        entityRecord={entityRecord}
                        formData={formData}
                        setFormData={setFormData} />;
                    case 'label':
                      return <EntityFormLabel label={entityRecord.text} entityName={entityName} entityData={entityData} />;
                    default:
                      return <></>;
                  }
                })
              }
              <Button onClick={save}>
                <BsCheckLg />{' ' + t(
                  entityFormTemplate.submitButtonLabel
                    ? entityFormTemplate.submitButtonLabel
                    : 'Save'
                )}
              </Button>
            </Form>
          </Col>
        </Row>
      </Container>
    </>;
}

const EntityXxudPage = () => {
  const params: {
    cloudServiceProvider: string,
    cloudContext: string,
    entityName: string,
    entityId: string,
    action: string
  } = useParams();

  const entityFormTemplate = ENTITY_FORM_LIST.filter((template) => {
    return template.cloudServiceProvider === params.cloudServiceProvider
      && (template.entityName === params.entityName || template.entityName === '')
      && template.actionType === params.action
  });

  if (entityFormTemplate.length >= 1) {
    console.log(entityFormTemplate);
    return <EntityXxudPageImpl
      cloudContext={params.cloudContext}
      entityFormTemplate={entityFormTemplate[0]}
      entityName={params.entityName}
      entityId={params.entityId} />;
  }

  // Default Page
  return <>
    <SideBar />
    <Container fluid className="px-0">
      <Row>
        <Col>
          <span>EntityCxudPage</span><br />
          <span>cloudServiceProvider={params.cloudServiceProvider}</span><br />
          <span>cloudContext={params.cloudContext}</span><br />
          <span>entityName={params.entityName}</span><br />
          <span>entityId={params.entityId}</span><br />
          <span>action={params.action}</span>
        </Col>
      </Row>
    </Container>
  </>;
}

export default EntityXxudPage;
