import { CLOUD_SERVICE_PROVIDER_LIST, DEFAULT_CLOUD_CONTEXTS, ROUTE_URL } from 'constant/other';
import useDrupalJsonApi from 'hooks/drupal_jsonapi';
import CloudContext from 'model/CloudContext';
import CloudServiceProvider from 'model/CloudServiceProvider';
import EntityData from 'model/EntityData';
import EntityInfoTemplate from 'model/EntityInfoTemplate';
import { createContext, useEffect, useState } from 'react';
import { useTranslation } from 'react-i18next';
import { useParams } from 'react-router-dom';

type Action = {
  type: 'setLanguage',
  message: string
};

interface AppState {
  cloudContextList: CloudContext[];
  loadingFlg: boolean;
  dispatch: (action: Action) => void;
}

export const useAppState = (): AppState => {
  const { getEntityListAll } = useDrupalJsonApi();
  const [cloudContextList, setCloudContextList] = useState<CloudContext[]>([...DEFAULT_CLOUD_CONTEXTS]);
  const [loadingFlg, setLoadingFlg] = useState(true);
  const { i18n } = useTranslation();

  useEffect(() => {
    const languageName = window.localStorage.getItem('languageName');
    if (languageName !== null) {
      i18n.changeLanguage(languageName);
    }
    // eslint-disable-next-line react-hooks/exhaustive-deps
  }, []);

  // Set cloud context list.
  useEffect(() => {
    const init = async () => {
      let newCloudContextList = [...DEFAULT_CLOUD_CONTEXTS];
      for (const cloudServiceProvider of CLOUD_SERVICE_PROVIDER_LIST) {
        try {
          const data = (await getEntityListAll('cloud_config', {}, cloudServiceProvider))
            .map((record: any) => {
              return {
                cloudServiceProvider: cloudServiceProvider as CloudServiceProvider,
                name: record.attributes.cloud_context,
                labelName: record.attributes.name
              };
            });
          newCloudContextList = [...newCloudContextList, ...data];
        } catch {
          // pass
        }
      }
      setCloudContextList(newCloudContextList);
      setLoadingFlg(false);
    };
    init();
    // eslint-disable-next-line react-hooks/exhaustive-deps
  }, []);

  const dispatch = (action: Action) => {
    switch (action.type) {
      case 'setLanguage':
        i18n.changeLanguage(action.message);
        window.localStorage.setItem('languageName', action.message);
        break;
    }
  };

  return {
    cloudContextList,
    loadingFlg,
    dispatch
  };
};

export const useEntityData = (entityInfoTemplate: EntityInfoTemplate) => {
  const { getJsonData } = useDrupalJsonApi();
  const [entityData, setEntityData] = useState<EntityData | null>(null);
  const params: {
    cloudServiceProvider: string,
    cloudContext: string,
    entityName: string,
    entityId?: string
  } = useParams();

  useEffect(() => {
    if (typeof params.entityId !== 'undefined') {
      const entityTypeId = `${entityInfoTemplate.cloudServiceProvider}_${entityInfoTemplate.entityName}`;
      const url = `/jsonapi/${entityTypeId}/${entityTypeId}?filter[drupal_internal__id]=${params.entityId}`;
      getJsonData<{ data: EntityData[] }>(url).then((response) => {
        if (response.data.length >= 1) {
          setEntityData(response.data[0]);
        }
      });
    }
    // eslint-disable-next-line react-hooks/exhaustive-deps
  }, [params]);

  const titleKey = entityInfoTemplate.entityName;
  const title = titleKey.substring(0, 1).toUpperCase() + titleKey.substring(1);

  const action = async (parameter: {
    [key: string]: any,
  }, command: string, entityId: string = '') => {
    if (entityData === null) {
      return;
    }
    const formData = new FormData();
    for (const pair of Object.entries(parameter)) {
      switch (typeof pair[1]) {
        case 'object':
          formData.append(pair[0], JSON.stringify(pair[1]));
          break;
        default:
          formData.append(pair[0], `${pair[1]}`);
          break;
      }
    }

    const url = entityId !== ''
      ? `/cloud_dashboard/${entityInfoTemplate.cloudServiceProvider}/${entityData.attributes['cloud_context']}/${entityInfoTemplate.cloudServiceProvider}_${entityInfoTemplate.entityName}/${entityId}/${command}`
      : `/cloud_dashboard/${entityInfoTemplate.cloudServiceProvider}/${entityData.attributes['cloud_context']}/${entityInfoTemplate.cloudServiceProvider}_${entityInfoTemplate.entityName}/${command}`;
    const response = await fetch(url, {
      method: 'POST',
      body: formData,
    });
    if (!response.ok) {
      console.log(command + ' Request : No');
      console.error(command + ' failed.');
      console.groupEnd();
      console.error(await response.json());
      window.alert(command + ' failed.');
    }
    console.log(command + ' Request : Yes');
    console.log(await response.json());
    window.location.href = `${ROUTE_URL}/${entityInfoTemplate.cloudServiceProvider}/${entityInfoTemplate.entityName}`;
  }

  return {
    entityData,
    titleKey,
    title,
    action
  };
}

export const AppContext = createContext<AppState>({
  cloudContextList: [],
  loadingFlg: true,
  dispatch: () => { }
});
