import { VMWARE_OS_DICT } from 'constant/other';
import { GetEntityListAllType } from 'hooks/drupal_jsonapi';
import L from 'leaflet';
import CloudContext from 'model/CloudContext';
import CloudContextItem from 'model/CloudContextItem';
import DataRecord from 'model/DataRecord';
import EntityColumn from 'model/EntityColumn';
import EntityData from 'model/EntityData';
import EntityFormTemplate from 'model/EntityFormTemplate';
import MenuTemplate from 'model/MenuTemplate';
import RawCloudContextItem from 'model/RawCloudContextItem';
import { useEffect, useRef } from 'react';

/**
 * Padding Zero String.
 * @param data data
 * @param length data length
 * @returns Padded value
 */
export const paddingZero = (data: any, length: number) => {
  return `000000${data}`.slice(-length);
}

/**
 * Convert to datetime string for UI.
 * @param dateString datetime string
 * @returns datetime string for UI
 */
export const convertDateString = (dateString: string) => {
  const date = new Date(dateString);
  const year = date.getFullYear();
  const month = date.getMonth() + 1;
  const day = date.getDate();
  const hour = date.getHours();
  const minute = date.getMinutes();
  let output = `${year}/${paddingZero(month, 2)}/${paddingZero(day, 2)}`;
  output += ` - ${paddingZero(hour, 2)}:${paddingZero(minute, 2)}`;
  return output;
}

/**
 * Round number method.
 * @param value number
 * @param digit digit
 * @returns fixed value string
 */
const roundNumber = (value: number, digit: number) => {
  // This code means "x = 10 ^ digit".
  let x = 1.0;
  for (let i = 0; i < digit; i += 1) {
    x *= 10;
  }

  // Use guard clause.
  if (Math.floor(value) !== value) {
    return `${Math.round(value * x) / x}`;
  }

  let textBuffer = `${value}`;
  if (digit !== 0) {
    textBuffer += '.';
    for (let i = 0; i < digit; i += 1) {
      textBuffer += '0';
    }
  }
  return textBuffer;
};

const convertKeyValueCrLfData = (data: string) => {
  if (data === '{  }') {
    return '';
  }

  const records = data.split('\n');
  const CONVERT_KEY_VALUE_CRLF_DICT: Record<string, string> = {
    'on_demand_hourly: ': 'On-demand hourly ($): ',
    'on_demand_daily: ': 'On-demand daily ($): ',
    'on_demand_monthly: ': 'On-demand monthly ($): ',
    'on_demand_yearly: ': 'On-demand yearly ($): ',
    'ri_one_year: ': 'RI 1 year ($): ',
    'ri_three_year: ': 'RI 3 year ($): ',
    'cpu: ': 'CPU (Usage): ',
    'memory: !!float ': 'Memory (Usage): ',
    'memory: ': 'Memory (Usage): ',
    'pod_allocation: ': 'Pods (Allocation): ',
    'cpu_capacity: !!float ': 'CPU (Capacity): ',
    'memory_capacity: !!float ': 'Memory (Capacity): ',
    'pod_capacity: ': 'Pods (Capacity): ',
  };
  const fixedRecords: string[] = [];
  for (const record of records) {
    if (record === '') {
      continue;
    }
    let fixedRecord = record;
    for (const key in CONVERT_KEY_VALUE_CRLF_DICT) {
      if (record.includes(key)) {
        const value = record.replace(key, '');
        const valueFloat = parseFloat(value);
        if (key.includes('!!float') || key.includes('memory: ')) {
          if (valueFloat >= 1024 * 1024 * 1024) {
            fixedRecord = `${CONVERT_KEY_VALUE_CRLF_DICT[key]}${roundNumber(valueFloat / (1024 * 1024 * 1024), 2)}Gi`;
          } else if (valueFloat >= 1024 * 1024) {
            fixedRecord = `${CONVERT_KEY_VALUE_CRLF_DICT[key]}${roundNumber(valueFloat / (1024 * 1024), 2)}Mi`;
          } else if (valueFloat >= 1024) {
            fixedRecord = `${CONVERT_KEY_VALUE_CRLF_DICT[key]}${roundNumber(valueFloat / (1024), 2)}Ki`;
          } else {
            fixedRecord = `${CONVERT_KEY_VALUE_CRLF_DICT[key]}${roundNumber(valueFloat, 2)}`;
          }
        } else {
          fixedRecord = CONVERT_KEY_VALUE_CRLF_DICT[key].includes('Pods ')
            ? CONVERT_KEY_VALUE_CRLF_DICT[key] + value
            : CONVERT_KEY_VALUE_CRLF_DICT[key] + roundNumber(valueFloat, 2);
        }
        break;
      }
    }
    fixedRecords.push(fixedRecord);
  }
  return fixedRecords.sort((a, b) => {
    return a < b ? -1 : a > b ? 1 : 0;
  }).join('\n');
}

/**
 * Converter of entity's data for UI.
 * @param data Entity's data
 * @param ec Information of Entity Column
 * @param dataCache Data cache for 'join' type
 */
export const convertDataForUI = (data: any, ec: EntityColumn, dataCache: { [key: string]: EntityData[] }): string => {
  // Null check.
  if (data === null) {
    return '';
  }

  switch (ec.type) {
    case 'datetime':
      return convertDateString(data);
    case 'cpu':
      return roundNumber(data, 2);
    case 'memory': {
      if (data >= 1024 * 1024 * 1024) {
        return `${roundNumber(data / (1024 * 1024 * 1024), 2)}Gi`;
      }
      if (data >= 1024 * 1024) {
        return `${roundNumber(data / (1024 * 1024), 2)}Mi`;
      }
      if (data >= 1024) {
        return `${roundNumber(data / (1024), 2)}Ki`;
      }
      return roundNumber(data, 2);
    }
    case 'key-value': {
      let records: string[] = [];
      for (const record of data) {
        records.push(`${record['item_key']}:${record['item_value']}`);
      }
      return records.join(', ');
    }
    case 'cost':
      return `$${data}`;
    case 'boolean':
      return data ? ec.value[0] : ec.value[1];
    case 'array':
      return data.map((r: any) => `${r}`).join(', ');
    case 'join':
      for (const record of dataCache[ec.info.entityTypeId]) {
        const recordData: EntityData = record;
        if (recordData.attributes[ec.info.keyColumn] === data) {
          return recordData.attributes[ec.info.valueColumn];
        }
      }
      return `${data}`;
    case 'key-value-crlf':
      return convertKeyValueCrLfData(data);
    case 'conditions':
      return data !== ec.value[0] ? data : ec.value[1];
    case 'cidr':
      return data.length > 0 ? data[0]['cidr'] : '';
    case 'number':
      return data.toString();
    case 'vmware_os':
      // source: VmwareVmGuestOsDataProvider.php
      return data in VMWARE_OS_DICT ? VMWARE_OS_DICT[data] : data;
    case 'link':
      return JSON.stringify({
        'url': ec.url,
        'label': ec.label,
        'type': ec.type,
      });
    default:
      return data;
  }
};

/**
 * Getter of EntityListView's URL for MenuTemplate.
 * @param menu MenuTemplate
 * @param cloudContext cloudContext
 * @returns URL
 */
export const getEntityListViewUrl = (menu: MenuTemplate, cloudContext?: string) => {
  return cloudContext === undefined
    ? `/${menu.cloudServiceProvider as string}/${menu.entityName}`
    : `/${menu.cloudServiceProvider as string}/${cloudContext}/${menu.entityName}`;
};

/**
 * Getter of EntityTypeId for MenuTemplate.
 * @param menu MenuTemplate
 * @returns EntityTypeId
 */
export const getEntityTypeId = (menu: MenuTemplate) => {
  return `${menu.cloudServiceProvider as string}_${menu.entityName}`;
};

export const usePrevious = <T>(value: T) => {
  const ref = useRef<T>();
  useEffect(() => {
    ref.current = value;
  });
  return ref.current;
}

/**
 * Getter of LaunchTemplateView's URL for CloudContext.
 * @param cloudContext CloudContext
 * @returns URL
 */
export const getLaunchTemplateViewUrl = (cloudContext: CloudContext) => {
  return cloudContext.name === 'ALL'
    ? `/${cloudContext.cloudServiceProvider as string}/server_template`
    : `/${cloudContext.cloudServiceProvider as string}/${cloudContext.name}/server_template`;
};

/**
 * Read entity's data for convert entity's data by JSON:API.
 *
 * @param getEntityListAll The getEntityListAll() function;
 * @param entityColumnList Information about the entities that will be loaded in advance.
 * @returns Data cache.
 */
export const readDataCache = async (
  getEntityListAll: GetEntityListAllType,
  entityColumnList: EntityColumn[]
) => {
  const dataCache: Record<string, EntityData[]> = {};
  for (const entityColumn of entityColumnList) {
    switch (entityColumn.type) {
      case 'join': {
        const entityTypeId = entityColumn.info.entityTypeId;
        if (!(entityTypeId in dataCache)) {
          dataCache[entityTypeId] = await getEntityListAll(entityTypeId);
        }
        break;
      }
      case 'relationship':
        const entityTypeId = entityColumn.info.entityTypeId;
        if (!(entityTypeId in dataCache)) {
          dataCache[entityTypeId] = await getEntityListAll(entityTypeId);
        }
        break;
    }
  }
  return dataCache;
};

/**
 * Convert EntityData to DataRecord.
 *
 * @param rawDataList List of EntityData.
 * @param entityColumnList List of EntiyColumn.
 * @param cloudContext cloud_context.
 * @param dataCache Data cache of EntityData.
 * @returns List of DataRecord.
 */
export const convertEntityData = (
  entityTypeId: string,
  rawDataList: EntityData[],
  entityColumnList: EntityColumn[],
  cloudContextList: CloudContext[],
  dataCache: { [key: string]: EntityData[] }) => {
  const newDataRecordList: DataRecord[] = [];
  for (const rawRecord of rawDataList) {
    // convert (rawRecord: EntityData) => (dataRecord: DataRecord)
    const dataRecord: DataRecord = {
      id: `${rawRecord.attributes['drupal_internal__id']}`,
      entityTypeId,
      value: {},
      cloudContext: ''
    };
    dataRecord.cloudContext = rawRecord.attributes['cloud_context'];

    // Read data column by launchTemplateColumn, convert as appropriate, and write it out.
    for (const launchTemplateColumn of entityColumnList) {
      if (launchTemplateColumn.name in rawRecord.attributes) {
        // Read data column by launchTemplateColumn.
        const rawValue = rawRecord.attributes[launchTemplateColumn.name];

        // Convert as appropriate.
        if (launchTemplateColumn.name !== 'cloud_context') {
          // If it is not 'cloud_context', convert it by convertDataForUI().
          dataRecord.value[launchTemplateColumn.name] = convertDataForUI(rawValue, launchTemplateColumn, dataCache);
          continue;
        }
        // Search for possible candidates.
        const candidate = cloudContextList.filter((r) => {
          return r.name !== 'ALL' && r.name === rawValue;
        });
        // Branching depending on the number of candidates.
        dataRecord.value[launchTemplateColumn.name] = candidate.length >= 1
          ? candidate[0].labelName
          : rawValue;
        continue;
      }

      if (launchTemplateColumn.type === 'relationship' && typeof rawRecord.relationships !== 'string') {
        if (launchTemplateColumn.name in rawRecord.relationships) {
          // Read data column by launchTemplateColumn.
          const rawValue = rawRecord.relationships[launchTemplateColumn.name];
          const relationshipsData = Array.isArray(rawValue.data) ? rawValue.data[0] : rawValue.data;

          const entityId = relationshipsData.meta[launchTemplateColumn.info.keyColumn1];
          const entityList = dataCache[launchTemplateColumn.info.entityTypeId];
          const selectedEntity = entityList.filter((e) => {
            return e.attributes[launchTemplateColumn.info.keyColumn2] === entityId;
          });
          dataRecord.value[launchTemplateColumn.name] = selectedEntity.length >= 1
            ? selectedEntity[0].attributes[launchTemplateColumn.info.valueColumn]
            : '';
        }
      }
    }
    // Write it out.
    newDataRecordList.push(dataRecord);
  }
  return newDataRecordList;
};

/**
 * Convert a list of RawCloudContextItem to a format more suitable for display.
 *
 * @param rawCloudContextItemList List of RawCloudContextItem.
 * @returns List of CloudContextItem.
 */
export const convertCloudContextItemList = (
  rawCloudContextItemList: RawCloudContextItem[],
  defaultIconUri: string
): CloudContextItem[] => {
  const hash: Record<string, CloudContextItem> = {};
  for (const rawCloudContextItem of rawCloudContextItemList) {
    const hashKey = `${rawCloudContextItem.Latitude},${rawCloudContextItem.Longitude}`;
    if (hashKey in hash) {
      hash[hashKey].icon = new L.Icon({
        iconUrl: defaultIconUri
      });
    } else {
      hash[hashKey] = {
        icon: new L.Icon({
          iconUrl: rawCloudContextItem.Items[0].Image
        }),
        cloudServiceProvider: rawCloudContextItem.Type,
        position: [
          parseFloat(rawCloudContextItem.Latitude),
          parseFloat(rawCloudContextItem.Longitude),
        ],
        item: []
      };
    }

    const itemList = rawCloudContextItem.Items.map((rawItem) => {
      return {
        iconUrl: rawItem.Image,
        entityViewUrl: rawItem.Url.replace(/^\/clouds/, ''),
        name: rawItem.Name,
        positionLabel: `${rawCloudContextItem.City}, ${rawCloudContextItem.Country}`,
      };
    });
    hash[hashKey].item = [...hash[hashKey].item, ...itemList];
  }
  return Object.values(hash);
}

/**
 * Getter of ProjectView's URL for CloudContext.
 * @param cloudContext CloudContext
 * @returns URL
 */
export const getProjectViewUrl = (cloudContext: CloudContext) => {
  return cloudContext.name === 'ALL'
    ? `/${cloudContext.cloudServiceProvider as string}/project`
    : `/${cloudContext.cloudServiceProvider as string}/${cloudContext.name}/project`;
};

/**
 * Wrapper of window.localStorage.getItem();
 *
 * @param key Key.
 * @param defaultValue Default value.
 * @returns Item value.
 */
export const getLocalStorageItem = (key: string, defaultValue: string) => {
  const temp = window.localStorage.getItem(key);
  return temp !== null ? temp : defaultValue;
}

/**
 * Wrapper of window.localStorage.setItem();
 *
 * @param key Key.
 * @param value Item value.
 */
export const setLocalStorageItem = (key: string, value: string) => {
  window.localStorage.setItem(key, value);
}

/**
 * Record<string, any> to FormData.
 *
 * @param formData Record<string, any>.
 * @param entityFormTemplate Template of entity form.
 *
 * @returns FormData type.
 */
export const formDataToSendData = (formData: Record<string, any>, entityFormTemplate: EntityFormTemplate) => {
  const fileTypeRecordList = [];
  for (const record1 of entityFormTemplate.entityRecords) {
    if (record1.type === 'panel') {
      for (const record2 of record1.keyValueRecords) {
        if (record2.type === 'file') {
          fileTypeRecordList.push(record2.name);
        }
      }
    }
  }

  const sendData = new FormData();
  for (const pair of Object.entries(formData)) {
    switch (typeof pair[1]) {
      case 'object':
        sendData.append(
          pair[0],
          fileTypeRecordList.includes(pair[0])
            ? pair[1]
            : JSON.stringify(pair[1])
        );
        break;
      default:
        sendData.append(pair[0], `${pair[1]}`);
        break;
    }
  }
  return sendData;
}

export const capitalize = (text: string) => {
  if (text.length === 0) {
    return '';
  }
  return text.slice(0, 1).toUpperCase() + text.slice(1);
}

/**
   * calc the panel name.
   * Note: If the title of the Panel element has not been determined,
   * it will be assigned a name based on the type of entity.
   * Ex.) 'instance' => 'Instance', 'replica_set' => 'Replica Set'
   *
   * @param name The panel name as described in the configuration file.
   * @param entityName The entity's type name.
   * @returns The Calculated panel name.
   */
export const calcPanelName = (name: string, entityName: string) => {
  if (name !== '') {
    return name;
  }

  return entityName.split('_').map((text) => {
    return capitalize(text);
  }).join(' ');
}

/**
 * Getter of replaced URL for EntityDetailPage.
 * @param url url before replace
 * @param cloudContext cloudContext
 * @param entityName entityName
 * @param entityId entityId
 * @param cloudServiceProvider cloudServiceProvider
 * @param index index
 * @returns URL
 */
export const getReplacedUrl = (
  url: string,
  cloudContext: string,
  entityName: string,
  entityId: string,
  cloudServiceProvider: string,
  index: number
) => {
  return  url
    .replace('{cloudContext}', cloudContext)
    .replace('{entityName}', entityName)
    .replace('{entityId}', entityId)
    .replace('{cloudServiceProvider}', cloudServiceProvider)
    .replace('{index}', `${index}`)
}
