<?php

namespace Drupal\aws_cloud\Form\Vpc;

use Drupal\aws_cloud\Form\Ec2\AwsDeleteForm;
use Drupal\cloud\Service\Util\EntityLinkWithNameHtmlGenerator;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\Markup;
use Drupal\Core\StringTranslation\TranslatableMarkup;

/**
 * Detach internet gateway form.
 */
class InternetGatewayDetachForm extends AwsDeleteForm {

  /**
   * {@inheritdoc}
   */
  public function getQuestion(): TranslatableMarkup {
    /** @var \Drupal\aws_cloud\Entity\Vpc\InternetGateway $entity */
    $entity = $this->entity;

    return $this->t('Are you sure you want to detach the internet gateway: %name?', [
      '%name' => $entity->getName(),
    ]);
  }

  /**
   * {@inheritdoc}
   */
  public function getDescription(): string {
    $entity = $this->entity;

    // Get module name.
    $module_name = $this->getModuleName($entity);

    $msg = '<h2>Internet gateway information:</h2>';

    $internet_gateway_id = $entity->getInternetGatewayId();
    $vpc_id = $entity->getVpcId();

    $internet_gateway_link = $this->entityLinkRenderer->renderViewElement(
      $internet_gateway_id,
      "{$module_name}_internet_gateway",
      'internet_gateway_id',
      [],
      '',
      EntityLinkWithNameHtmlGenerator::class
    );

    $vpc_link = $this->entityLinkRenderer->renderViewElement(
      $vpc_id,
      "{$module_name}_vpc",
      'vpc_id',
      [],
      '',
      EntityLinkWithNameHtmlGenerator::class
    );

    $msg .= $this->t(
      '<ul><li>Internet gateway: %internet_gateway</li><li>Attached to VPC: %vpc</li></ul>',
      [
        '%internet_gateway' => Markup::create($internet_gateway_link['#markup']),
        '%vpc' => Markup::create($vpc_link['#markup']),
      ]
    );

    return $msg;
  }

  /**
   * {@inheritdoc}
   */
  public function getConfirmText(): TranslatableMarkup {
    return $this->t('Detach');
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state): array {
    if (!$this->updateEntityBuildForm($this->entity)) {
      return $this->redirectUrl;
    }
    $form = parent::buildForm($form, $form_state);
    if ($this->entity->getVpcId() === NULL) {
      $form['error'] = [
        '#markup' => '<div>' . $this->t('Internet gateway is already detached') . '</div>',
      ];
      unset($form['description']);
      unset($form['actions']['submit']);
    }
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state): void {
    $this->ec2Service->setCloudContext($this->entity->getCloudContext());
    $entity = $this->entity;
    $vpc_id = $entity->getVpcId();

    // Get module name.
    $module_name = $this->getModuleName($entity);

    $result = $this->ec2Service->detachInternetGateway([
      'InternetGatewayId' => $entity->getInternetGatewayId(),
      'VpcId' => $entity->getVpcId(),
    ]);

    $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);

    if ($result === NULL) {
      $this->messenger->addError($this->t('Unable to detach internet gateway.'));
    }

    $internet_gateway_link = $this->entityLinkRenderer->renderViewElement(
      $entity->getInternetGatewayId(),
      "{$module_name}_internet_gateway",
      'internet_gateway_id',
      [],
      '',
      EntityLinkWithNameHtmlGenerator::class
    );

    $vpc_link = $this->entityLinkRenderer->renderViewElement(
      $vpc_id,
      "{$module_name}_vpc",
      'vpc_id',
      [],
      '',
      EntityLinkWithNameHtmlGenerator::class
    );

    if (!empty($result['SendToWorker'])) {
      $this->messenger->addStatus($this->t('The internet gateway %internet_gateway will be detached from VPC %vpc remotely', [
        '%internet_gateway' => Markup::create($internet_gateway_link['#markup']),
        '%vpc' => Markup::create($vpc_link['#markup']),
      ]));

      return;
    }

    $this->messenger->addStatus($this->t('The internet gateway %internet_gateway detached from VPC %vpc', [
      '%internet_gateway' => Markup::create($internet_gateway_link['#markup']),
      '%vpc' => Markup::create($vpc_link['#markup']),
    ]));

    // Update the internet gateway.
    $this->ec2Service->updateInternetGateways([
      'InternetGatewayIds' => [$entity->getInternetGatewayId()],
    ], FALSE);

    $this->clearCacheValues($entity->getCacheTags());
  }

}
