<?php

namespace Drupal\openstack\Controller;

use Drupal\aws_cloud\Entity\Ec2\IntermediateFormState;
use Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface;
use Drupal\cloud\Service\CloudService;
use Drupal\cloud\Service\CloudServiceInterface;
use Drupal\cloud\Traits\AccessCheckTrait;
use Drupal\cloud\Traits\CloudContentEntityTrait;
use Drupal\cloud\Traits\CloudResourceTrait;
use Drupal\Core\Access\AccessResult;
use Drupal\Core\Access\AccessResultInterface;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Messenger\Messenger;
use Drupal\Core\Render\RendererInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\openstack\Service\OpenStackOperationsServiceInterface;
use Drupal\openstack\Service\OpenStackServiceFactoryInterface;
use Drupal\openstack\Service\OpenStackServiceInterface;
use Drupal\views\Views;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\RedirectResponse;
use Symfony\Component\HttpFoundation\RequestStack;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Route;

/**
 * Api controller for interacting with OpenStack.
 */
class ApiController extends ControllerBase implements ApiControllerInterface {

  use CloudContentEntityTrait;
  use CloudResourceTrait;
  use AccessCheckTrait;

  /**
   * The OpenStack Service.
   *
   * @var \Drupal\openstack\Service\OpenStackServiceFactoryInterface
   */
  private $openStackServiceFactory;

  /**
   * The Messenger service.
   *
   * @var \Drupal\Core\Messenger\Messenger
   */
  protected $messenger;

  /**
   * Request stack.
   *
   * @var \Symfony\Component\HttpFoundation\RequestStack
   */
  private $requestStack;

  /**
   * The cloud service provider plugin manager (CloudConfigPluginManager).
   *
   * @var \Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface
   */
  protected $cloudConfigPluginManager;

  /**
   * Renderer service.
   *
   * @var \Drupal\Core\Render\RendererInterface
   */
  private $renderer;

  /**
   * Cloud service interface.
   *
   * @var \Drupal\cloud\Service\CloudServiceInterface
   */
  protected $cloudService;

  /**
   * The configuration data.
   *
   * @var array
   */
  protected $configuration;

  /**
   * The current user.
   *
   * @var \Drupal\Core\Session\AccountInterface
   */
  protected $currentUser;

  /**
   * The OpenStack Operations Service.
   *
   * @var \Drupal\openstack\Service\OpenStackOperationsServiceInterface
   */
  protected $openStackOperationsService;

  /**
   * The OpenStack Service.
   *
   * @var \Drupal\openstack\Service\Rest\OpenStackServiceInterface
   */
  protected $openStackService;

  /**
   * ApiController constructor.
   *
   * @param \Drupal\openstack\Service\OpenStackServiceFactoryInterface $openstack_service_factory
   *   Object for interfacing with OpenStack Service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The Entity type manager.
   * @param \Drupal\Core\Messenger\Messenger $messenger
   *   Messenger Object.
   * @param \Symfony\Component\HttpFoundation\RequestStack $request_stack
   *   Request stack object.
   * @param \Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface $cloud_config_plugin_manager
   *   The cloud service provider plugin manager (CloudConfigPluginManager).
   * @param \Drupal\Core\Render\RendererInterface $renderer
   *   The renderer service.
   * @param \Drupal\cloud\Service\CloudServiceInterface $cloud_service
   *   Cloud service.
   * @param \Drupal\Core\Session\AccountInterface $current_user
   *   The current user.
   * @param \Drupal\openstack\Service\OpenStackOperationsServiceInterface $openstack_operations_service
   *   The OpenStack Operations service.
   * @param \Drupal\openstack\Service\OpenStackServiceInterface $openstack_service
   *   The OpenStack Service.
   */
  public function __construct(
    OpenStackServiceFactoryInterface $openstack_service_factory,
    EntityTypeManagerInterface $entityTypeManager,
    Messenger $messenger,
    RequestStack $request_stack,
    CloudConfigPluginManagerInterface $cloud_config_plugin_manager,
    RendererInterface $renderer,
    CloudServiceInterface $cloud_service,
    AccountInterface $current_user,
    OpenStackOperationsServiceInterface $openstack_operations_service,
    OpenStackServiceInterface $openstack_service
  ) {
    $this->openStackServiceFactory = $openstack_service_factory;
    $this->entityTypeManager = $entityTypeManager;
    $this->messenger = $messenger;
    $this->requestStack = $request_stack;
    $this->cloudConfigPluginManager = $cloud_config_plugin_manager;
    $this->renderer = $renderer;
    $this->cloudService = $cloud_service;
    $this->currentUser = $current_user;
    $this->configuration = ['cloud_context' => ''];
    $this->openStackOperationsService = $openstack_operations_service;
    $this->openStackService = $openstack_service;
  }

  /**
   * Dependency Injection.
   *
   * @param \Symfony\Component\DependencyInjection\ContainerInterface $container
   *   Instance of ContainerInterface.
   *
   * @return ApiController
   *   return created object.
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('openstack.factory'),
      $container->get('entity_type.manager'),
      $container->get('messenger'),
      $container->get('request_stack'),
      $container->get('plugin.manager.cloud_config_plugin'),
      $container->get('renderer'),
      $container->get('cloud'),
      $container->get('current_user'),
      $container->get('openstack.operations'),
      $container->get('openstack.rest')
    );
  }

  /**
   * Checks user access for a specific request based on the cloud context.
   *
   * @param string $cloud_context
   *   Cloud context to check.
   * @param \Drupal\Core\Session\AccountInterface $account
   *   Run access checks for this account.
   * @param \Symfony\Component\Routing\Route $route
   *   The route object.
   * @param string $entity_type_id
   *   The entity type.
   * @param string $entity_id
   *   The entity id.
   * @param string $command
   *   The command name.
   *
   * @return \Drupal\Core\Access\AccessResultInterface
   *   The access result.
   */
  public function access(string $cloud_context, AccountInterface $account, Route $route, string $entity_type_id, string $entity_id, string $command): AccessResultInterface {
    $entity = $this->entityTypeManager
      ->getStorage($entity_type_id)
      ->load($entity_id);

    if (!empty($entity) && $entity instanceof EntityInterface) {
      return $entity->access($command, $account, TRUE);
    }
    // Unknown operation, no opinion.
    return AccessResult::neutral();
  }

  /**
   * Update message after refresh resources.
   *
   * @param string $resource_type
   *   Entity type name.
   * @param array $updated
   *   Updated statuses.
   */
  protected function updateMessages($resource_type, array $updated): void {
    $labels = $this->getDisplayLabels($resource_type);
    if (isset($updated[0])) {
      if ($updated[0]) {
        $this->messageUser($this->t('Updated @resource_name.', ['@resource_name' => $labels['plural']]));
      }
      else {
        $this->messageUser($this->t('Unable to update @resource_name.', ['@resource_name' => $labels['plural']]), 'error');
      }
    }
    $this->cloudService->invalidateCacheTags();
  }

  /**
   * Update instances.
   *
   * @param string $cloud_context
   *   Cloud context string.
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   *   Redirect to instance list.
   */
  public function updateInstanceList($cloud_context = ''): RedirectResponse {
    return $this->updateResourceList('openstack_instance', $cloud_context);
  }

  /**
   * Update images.
   *
   * @param string $cloud_context
   *   Cloud context string.
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   *   Redirect to image list.
   */
  public function updateImageList($cloud_context = ''): RedirectResponse {
    return $this->updateResourceList('openstack_image', $cloud_context);
  }

  /**
   * Update key pairs.
   *
   * @param string $cloud_context
   *   Cloud context string.
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   *   Redirect to keypair list.
   */
  public function updateKeyPairList($cloud_context = ''): RedirectResponse {
    return $this->updateResourceList('openstack_key_pair', $cloud_context);
  }

  /**
   * Update security groups.
   *
   * @param string $cloud_context
   *   Cloud context string.
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   *   Redirect to security group list.
   */
  public function updateSecurityGroupList($cloud_context = ''): RedirectResponse {
    return $this->updateResourceList('openstack_security_group', $cloud_context);
  }

  /**
   * Update volumes.
   *
   * @param string $cloud_context
   *   Cloud context string.
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   *   Redirect to volume list.
   */
  public function updateVolumeList($cloud_context = ''): RedirectResponse {
    return $this->updateResourceList('openstack_volume', $cloud_context);
  }

  /**
   * Update snapshots.
   *
   * @param string $cloud_context
   *   Cloud context string.
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   *   Redirect to snapshot list.
   */
  public function updateSnapshotList($cloud_context = ''): RedirectResponse {
    return $this->updateResourceList('openstack_snapshot', $cloud_context);
  }

  /**
   * Update network interfaces.
   *
   * @param string $cloud_context
   *   Cloud context string.
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   *   Redirect to network interface list.
   */
  public function updateNetworkInterfaceList($cloud_context = ''): RedirectResponse {
    return $this->updateResourceList('openstack_network_interface', $cloud_context);
  }

  /**
   * Update networks.
   *
   * @param string $cloud_context
   *   Cloud context string.
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   *   Redirect to network list.
   */
  public function updateNetworkList($cloud_context = ''): RedirectResponse {
    return $this->updateResourceList('openstack_network', $cloud_context);
  }

  /**
   * Update floating ips.
   *
   * @param string $cloud_context
   *   Cloud context string.
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   *   Redirect to floating ip list.
   */
  public function updateFloatingIpList($cloud_context = ''): RedirectResponse {
    return $this->updateResourceList('openstack_floating_ip', $cloud_context);
  }

  /**
   * {@inheritdoc}
   */
  public function updateAll(): RedirectResponse {
    $regions = $this->requestStack->getCurrentRequest()->query->get('regions');
    if ($regions === NULL) {
      $this->messageUser($this->t('No region specified'), 'error');
    }
    else {
      $regions_array = explode(',', $regions);

      foreach ($regions_array ?: [] as $region) {
        $entity = $this->entityTypeManager()->getStorage('cloud_config')
          ->loadByProperties(
            [
              'cloud_context' => $region,
            ]);
        if ($entity) {
          openstack_update_resources(array_shift($entity));
        }
      }

      $this->messageUser($this->t('Creating cloud service provider was performed successfully.'));
      drupal_flush_all_caches();
    }
    return $this->redirect('entity.cloud_config.collection');
  }

  /**
   * {@inheritdoc}
   */
  public function listImageCallback(): Response {
    return $this->getViewResponse('openstack_image');
  }

  /**
   * Helper method to get views output.
   *
   * @param string $view_id
   *   The ID of list view.
   *
   * @return \Symfony\Component\HttpFoundation\Response
   *   The response of list view.
   *
   * @throws \Exception
   *   Thrown when an object is passed in which cannot be printed.
   */
  private function getViewResponse($view_id): Response {
    $view = Views::getView($view_id);

    // Set the display machine name.
    $view->setDisplay('list');

    // Render the view as html, and return it as a response object.
    $build = $view->executeDisplay();
    return new Response($this->renderer->render($build));
  }

  /**
   * Helper method to add messages for the end user.
   *
   * @param string $message
   *   The message.
   * @param string $type
   *   The message type: error or message.
   */
  private function messageUser($message, $type = 'message'): void {
    switch ($type) {
      case 'error':
        $this->messenger->addError($message);
        break;

      case 'message':
        $this->messenger->addStatus($message);
        break;

      default:
        break;
    }
  }

  /**
   * Helper method to update entities.
   *
   * @param string $entity_type_name
   *   The entity type name.
   * @param string $cloud_context
   *   The cloud context.
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   *   An associative array with a redirect route and any parameters to build
   *   the route.
   */
  private function updateResourceList($entity_type_name, $cloud_context): RedirectResponse {
    if (empty($cloud_context)) {
      return $this->updateAllResourceList($entity_type_name);
    }

    $labels = $this->getDisplayLabels($entity_type_name);
    $update_method_name = 'update' . self::getCamelCaseWithoutWhitespace($labels['plural']);
    $openstack_service = $this->openStackServiceFactory->get($cloud_context);
    $updated = $openstack_service->$update_method_name();

    if ($updated !== FALSE) {
      $this->messageUser($this->t('Updated @name.', ['@name' => $labels['plural']]));
    }
    else {
      $this->messageUser($this->t('Unable to update @name.', ['@name' => $labels['plural']]), 'error');
    }

    // Update the cache.
    $this->cloudService->invalidateCacheTags();

    return $this->redirect("view.$entity_type_name.list", [
      'cloud_context' => $cloud_context,
    ]);
  }

  /**
   * Helper method to update all resource list.
   *
   * @param string $entity_type_name
   *   The entity type name.
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   *   An associative array with a redirect route and any parameters to build
   *   the route.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  private function updateAllResourceList($entity_type_name): RedirectResponse {
    $cloud_configs = $this->entityTypeManager()
      ->getStorage('cloud_config')
      ->loadByProperties([
        'type' => 'openstack',
      ]);

    $labels = $this->getDisplayLabels($entity_type_name);
    $update_method_name = 'update' . self::getCamelCaseWithoutWhitespace($labels['plural']);

    foreach ($cloud_configs ?: [] as $cloud_config) {

      $cloud_context = $cloud_config->getCloudContext();
      $resource_link = $cloud_config->getResourceLink($labels['plural'], $entity_type_name);
      $cloud_config_link = $cloud_config->getCloudConfigLink();
      try {
        $openstack_service = $this->openStackServiceFactory->get($cloud_context);
        $updated = $openstack_service->$update_method_name();

        $success_message = $this->t('Updated %resources of %cloud_config cloud service provider.', [
          '%resources' => $resource_link ?? $labels['plural'],
          '%cloud_config' => $cloud_config_link ?? $cloud_config->getName(),
        ]);

        $error_message = $this->t('Unable to update %resources of %cloud_config cloud service provider.', [
          '%resources' => $resource_link ?? $labels['plural'],
          '%cloud_config' => $cloud_config_link ?? $cloud_config->getName(),
        ]);

        $updated === TRUE
        ? $this->messageUser($success_message)
        : $this->messageUser($error_message, 'error');
      }
      catch (\Exception $e) {
        $this->messenger->addError($this->t('Unable to update @resources of @cloud_config cloud service provider.', [
          '@resources' => $labels['plural'],
          '@cloud_config' => $cloud_config->getName(),
        ]));
      }
    }

    // Update the cache.
    $this->cloudService->invalidateCacheTags();

    return $this->redirect("view.$entity_type_name.all");
  }

  /**
   * {@inheritdoc}
   */
  public function getEntityCount(string $cloud_context, string $entity_type_id): JsonResponse {
    $params = !empty($cloud_context)
      ? ['cloud_context' => $cloud_context] : [];
    if (!empty($cloud_context)) {
      $this->configuration['cloud_context'] = $cloud_context;
    }
    $count = $this->getResourceCount(
      $entity_type_id,
      // Use a static trait method through CloudService.
      'list ' . CloudService::convertUnderscoreToWhitespace($entity_type_id),
      $params
    );

    return new JsonResponse(['count' => $count]);
  }

  /**
   * {@inheritdoc}
   */
  public function operateEntity(Request $request, string $cloud_context, string $entity_type_id, string $entity_id, string $command): JsonResponse {
    // Create an instance of the entity.
    // The process splits between adding a new entity
    // and when reading an existing entity.
    $entity = $command !== 'create' && $command !== 'import'
      ? $this->entityTypeManager
        ->getStorage($entity_type_id)
        ->load($entity_id)
      : $this->entityTypeManager
        ->getStorage($entity_type_id)
        ->create([
          'cloud_context' => $cloud_context,
        ]);
    if (empty($entity)) {
      return new JsonResponse([
        'result' => 'NG',
        'reason' => 'The entity has already been deleted.',
      ], 404);
    }

    // Depending on the content of the process,
    // branching is performed with a switch statement.
    $form_state = new IntermediateFormState();
    $method_name = '';
    switch ($command . '_' . $entity_type_id) {
      case 'edit_openstack_instance':
        /** @var \Drupal\aws_cloud\Entity\Ec2\InstanceInterface $entity */
        $entity->setName($request->get('name', ''));
        $entity->setTags(json_decode($request->get('annotations', '[]'), TRUE));
        $form_state->set('name', $request->get('name', ''));
        $form_state->set('tags', json_decode($request->get('annotations', '[]'), TRUE));
        $method_name = 'editOpenStackInstance';
        break;

      case 'delete_openstack_instance':
        $method_name = 'deleteOpenStackInstance';
        break;

      case 'start_openstack_instance':
        $method_name = 'startOpenStackInstance';
        break;

      case 'stop_openstack_instance':
        $method_name = 'stopOpenStackInstance';
        break;

      case 'reboot_openstack_instance':
        $method_name = 'rebootOpenStackInstance';
        break;

      case 'create_openstack_image':
        /** @var \Drupal\aws_cloud\Entity\Ec2\ImageInterface $entity */
        $entity->setName($request->get('name', ''));
        $entity->setVisibility(!empty($request->get('visibility', '')) ? 'public' : 'private');
        $form_state->set('name', $request->get('name', ''));
        $method_name = 'createOpenStackImage';
        break;

      case 'edit_openstack_image':
        /** @var \Drupal\aws_cloud\Entity\Ec2\ImageInterface $entity */
        $entity->setName($request->get('name', ''));
        $entity->setVisibility(!empty($request->get('visibility', '')) ? 'public' : 'private');
        $method_name = 'editOpenStackImage';
        break;

      case 'delete_openstack_image':
        $method_name = 'deleteOpenStackImage';
        break;

      case 'create_openstack_key_pair':
        /** @var \Drupal\aws_cloud\Entity\Ec2\KeyPairInterface $entity */
        $entity->set('key_pair_name', $request->get('key_pair_name', ''));
        $method_name = 'createOpenStackKeyPair';
        break;

      case 'edit_openstack_key_pair':
        $method_name = 'editOpenStackKeyPair';
        break;

      case 'delete_openstack_key_pair':
        $method_name = 'deleteOpenStackKeyPair';
        break;

      case 'import_openstack_key_pair':
        /** @var \Drupal\aws_cloud\Entity\Ec2\KeyPairInterface $entity */
        $entity->set('key_pair_name', $request->get('key_pair_name', ''));
        /** @var \Symfony\Component\HttpFoundation\File\UploadedFile $file */
        $file = $request->files->get('key_pair_public_key');
        if (empty($file)) {
          return new JsonResponse([
            'result' => 'NG',
            'reason' => 'Internal Server Error',
          ], 500);
        }
        $form_state->setValue('key_pair_public_key', $file->getRealPath());
        $method_name = 'importOpenStackKeyPair';
        break;

      case 'create_openstack_network_interface':
        /** @var \Drupal\aws_cloud\Entity\Ec2\NetworkInterfaceInterface $entity */
        $entity->setName($request->get('name', ''));
        $entity->setDescription($request->get('description', ''));
        $method_name = 'createOpenStackNetworkInterface';
        break;

      case 'delete_openstack_network_interface':
        $method_name = 'deleteOpenStackNetworkInterface';
        break;

      case 'edit_openstack_network_interface':
        /** @var \Drupal\aws_cloud\Entity\Ec2\NetworkInterfaceInterface $entity */
        $entity->setName($request->get('name', ''));
        $entity->setDescription($request->get('description', ''));
        $method_name = 'editOpenStackNetworkInterface';
        break;

      case 'create_openstack_floating_ip':
        /** @var \Drupal\aws_cloud\Entity\Ec2\ElasticIpInterface $entity */
        $entity->setName($request->get('name', ''));
        $form_state->setValue('floating_network_id', $request->get('floating_network_id', ''));
        $method_name = 'createOpenStackFloatingIp';
        break;

      case 'edit_openstack_floating_ip':
        /** @var \Drupal\aws_cloud\Entity\Ec2\ElasticIpInterface $entity */
        $entity->setName($request->get('name', ''));
        $method_name = 'editOpenStackFloatingIp';
        break;

      case 'delete_openstack_floating_ip':
        $method_name = 'deleteOpenStackFloatingIp';
        break;

      case 'associate_openstack_floating_ip':
        $form_state->setValue('resource_type', $request->get('resource_type', ''));
        $form_state->setValue('instance_id', $request->get('instance_id', '-1'));
        $form_state->setValue('instance_private_ip', $request->get('instance_private_ip', ''));
        $form_state->setValue('network_interface_id', $request->get('network_interface_id', '-1'));
        $form_state->setValue('network_private_ip', $request->get('network_private_ip', ''));
        $method_name = 'associateOpenStackFloatingIp';
        break;

      case 'disassociate_openstack_floating_ip':
        $method_name = 'disassociateOpenStackFloatingIp';
        break;

      case 'create_openstack_security_group':
        /** @var \Drupal\aws_cloud\Entity\Ec2\SecurityGroupInterface $entity */
        $entity->setGroupName($request->get('group_name', ''));
        $entity->setDescription($request->get('description', ''));
        $method_name = 'createOpenStackSecurityGroup';
        break;

      case 'delete_openstack_security_group':
        $method_name = 'deleteOpenStackSecurityGroup';
        break;

      case 'edit_openstack_security_group':
        /** @var \Drupal\aws_cloud\Entity\Ec2\SecurityGroupInterface $entity */
        $entity->setGroupName($request->get('group_name', ''));
        $entity->set('ip_permission', json_decode($request->get('ip_permission', '[]'), TRUE));
        $entity->set('outbound_permission', json_decode($request->get('outbound_permission', '[]'), TRUE));
        $method_name = 'editOpenStackSecurityGroup';
        break;

      case 'revoke_openstack_security_group':
        $method_name = 'revokeOpenStackSecurityGroup';
        break;

      case 'create_openstack_snapshot':
        /** @var \Drupal\aws_cloud\Entity\Ec2\SnapshotInterface $entity */
        $entity->setName($request->get('name', ''));
        $entity->set('volume_id', $request->get('volume_id', ''));
        $entity->set('description', $request->get('description', ''));
        $method_name = 'createOpenStackSnapshot';
        break;

      case 'delete_openstack_snapshot':
        $method_name = 'deleteOpenStackSnapshot';
        break;

      case 'edit_openstack_snapshot':
        /** @var \Drupal\aws_cloud\Entity\Ec2\SnapshotInterface $entity */
        $entity->setName($request->get('name', ''));
        $method_name = 'editOpenStackSnapshot';
        break;

      case 'create_openstack_volume':
        /** @var \Drupal\aws_cloud\Entity\Ec2\VolumeInterface $entity */
        $entity->setName($request->get('name', ''));
        $entity->setSnapshotId($request->get('snapshot_id', ''));
        $entity->setSize($request->get('size', ''));
        $entity->setVolumeType($request->get('volume_type', ''));
        $entity->set('availability_zone', $request->get('availability_zone', ''));
        $method_name = 'createOpenStackVolume';
        break;

      case 'delete_openstack_volume':
        $method_name = 'deleteOpenStackVolume';
        break;

      case 'edit_openstack_volume':
        /** @var \Drupal\aws_cloud\Entity\Ec2\VolumeInterface $entity */
        $entity->setName($request->get('name', ''));
        $method_name = 'editOpenStackVolume';
        break;

      case 'attach_openstack_volume':
        $form_state->setValue('device_name', $request->get('device_name', ''));
        $form_state->setValue('instance_id', $request->get('instance_id', ''));
        $method_name = 'attachOpenStackVolume';
        break;

      case 'detach_openstack_volume':
        $method_name = 'detachOpenStackVolume';
        break;

    }

    // Execute the process.
    $result = NULL;
    if (method_exists($this->openStackOperationsService, $method_name)) {
      $form = [];
      $result = $this->openStackOperationsService->$method_name(
        $entity,
        $form,
        $form_state
      );
    }

    return !empty($result)
      ? new JsonResponse([
        'result' => 'OK',
      ], 200)
      : new JsonResponse([
        'result' => 'NG',
        'reason' => 'Internal Server Error',
      ], 500);
  }

  /**
   * {@inheritdoc}
   */
  public function getNetworkInterfaceOptions(string $cloud_context): JsonResponse {
    $options = [];
    $network_interfaces = $this->entityTypeManager
      ->getStorage('openstack_network_interface')
      ->loadByProperties([
        'cloud_context' => $cloud_context,
        'external' => TRUE,
      ]);

    foreach ($network_interfaces ?: [] as $network_interface) {
      $options[] = [
        'value' => $network_interface->getNetworkInterfaceId(),
        'label' => $network_interface->getName(),
      ];
    }

    return new JsonResponse($options);
  }

  /**
   * {@inheritdoc}
   */
  public function getVolumeOptions($cloud_context): JsonResponse {
    $options = [];
    $params = [
      'cloud_context' => $cloud_context,
    ];

    if (!$this->currentUser->hasPermission("view any openstack volume")) {
      $params['uid'] = $this->currentUser->id();
    }

    $volumes = $this->entityTypeManager
      ->getStorage("openstack_volume")
      ->loadByProperties($params);

    foreach ($volumes ?: [] as $volume) {
      $options[] = [
        'value' => $volume->getVolumeId(),
        'label' => $volume->getName() !== $volume->getVolumeId()
          ? "{$volume->getName()} ({$volume->getVolumeId()})"
          : $volume->getVolumeId(),
      ];
    }

    return new JsonResponse($options);
  }

  /**
   * {@inheritdoc}
   */
  public function getUnassociatedInstanceOptions(string $entity_id): JsonResponse {
    /** @var \Drupal\openstack\Entity\OpenStackFloatingIp $entity */
    $entity = $this->entityTypeManager
      ->getStorage('openstack_floating_ip')
      ->load($entity_id);
    if (empty($entity)) {
      return new JsonResponse([
        'result' => 'NG',
        'reason' => 'The Floating IP has already been deleted.',
      ], 404);
    }

    // Get module name, network border group.
    $module_name = $this->getModuleName($entity);
    $network_border_group = $entity->getNetworkBorderGroup();

    $options = [];
    $account = $this->currentUser();

    $query = $this->entityTypeManager
      ->getStorage("{$module_name}_instance")
      ->getQuery()
      ->accessCheck(TRUE)
      ->condition('cloud_context', $entity->getCloudContext());

    // Filter by network border group.
    if (!empty($network_border_group)) {
      $subnets = $this->entityTypeManager
        ->getStorage("{$module_name}_subnet")
        ->loadByProperties([
          'cloud_context' => $entity->getCloudContext(),
          'network_border_group' => $network_border_group,
        ]);
      $subnet_ids = array_map(function ($subnet) {
        return $subnet->getSubnetId();
      }, $subnets);
      if (!empty($subnet_ids)) {
        $query->condition('subnet_id', $subnet_ids, 'IN');
      }
    }

    // Get cloud service provider name.
    // Use a static trait method through CloudService.
    $cloud_name = CloudService::convertUnderscoreToWhitespace($module_name);

    if (!$account->hasPermission("view any {$cloud_name} instance")) {
      $query->condition('uid', $account->id());
    }

    $results = $query->execute();

    foreach ($results ?: [] as $result) {

      $instance_list = $this->entityTypeManager
        ->getStorage("{$module_name}_instance")
        ->loadByProperties([
          'id' => $result,
          'cloud_context' => $entity->getCloudContext(),
        ]);

      /** @var \Drupal\aws_cloud\Entity\Ec2\InstanceInterface */
      $instance = count($instance_list) === 1
        ? array_shift($instance_list)
        : $instance_list;

      $private_ips = explode(', ', $instance->getPrivateIps());

      foreach ($private_ips ?: [] as $private_ip) {

        $elastic_ips = $this->entityTypeManager
          ->getStorage($entity->getEntityTypeId())
          ->loadbyProperties([
            'private_ip_address' => $private_ip,
            'cloud_context' => $entity->getCloudContext(),
          ]);

        if (count($elastic_ips) === 0) {
          $options[] = [
            'value' => $instance->id(),
            'label' => "{$instance->getName()} ({$instance->getInstanceId()}) - {$instance->getInstanceState()}",
          ];
        }
      }
    }

    return new JsonResponse($options);
  }

  /**
   * {@inheritdoc}
   */
  public function getSnapshotOptions($cloud_context): JsonResponse {
    $options = [];
    $params = [
      'cloud_context' => $cloud_context,
    ];

    if (!$this->currentUser->hasPermission('view any openstack snapshot')) {
      $params['uid'] = $this->currentUser->id();
    }

    $snapshots = $this->entityTypeManager
      ->getStorage('openstack_snapshot')
      ->loadByProperties($params);

    foreach ($snapshots ?: [] as $snapshot) {
      $options[] = [
        'value' => $snapshot->getSnapshotId(),
        'label' => $snapshot->getName() !== $snapshot->getSnapshotId()
          ? "{$snapshot->getName()} ({$snapshot->getSnapshotId()})"
          : $snapshot->getSnapshotId(),
      ];
    }

    return new JsonResponse($options);
  }

  /**
   * {@inheritdoc}
   */
  public function getUnassociatedNetworkInterfaceOptions(string $entity_id): JsonResponse {
    /** @var \Drupal\openstack\Entity\OpenStackFloatingIp $entity */
    $entity = $this->entityTypeManager
      ->getStorage('openstack_floating_ip')
      ->load($entity_id);
    if (empty($entity)) {
      return new JsonResponse([
        'result' => 'NG',
        'reason' => 'The Floating IP has already been deleted.',
      ], 404);
    }

    $options = [];

    // Get module name, network border group.
    $module_name = $this->getModuleName($entity);
    $network_border_group = $entity->getNetworkBorderGroup();

    $query = $this->entityTypeManager
      ->getStorage("{$module_name}_network_interface")
      ->getQuery()
      ->accessCheck(TRUE)
      ->condition('cloud_context', $entity->getCloudContext());

    $group = $query
      ->orConditionGroup()
      ->notExists('association_id')
      ->notExists('secondary_association_id');

    $query->condition($group);

    // Filter by network border group.
    if (!empty($network_border_group)) {
      $subnets = $this->entityTypeManager
        ->getStorage("{$module_name}_subnet")
        ->loadByProperties([
          'cloud_context' => $entity->getCloudContext(),
          'network_border_group' => $network_border_group,
        ]);
      $subnet_ids = array_map(function ($subnet) {
        return $subnet->getSubnetId();
      }, $subnets);
      if (!empty($subnet_ids)) {
        $query->condition('subnet_id', $subnet_ids, 'IN');
      }
    }

    $results = $query->execute();

    foreach ($results ?: [] as $result) {
      $interfaces = $this->entityTypeManager
        ->getStorage("{$module_name}_network_interface")
        ->loadByProperties([
          'id' => $result,
          'cloud_context' => $entity->getCloudContext(),
        ]);

      /** @var \Drupal\aws_cloud\Entity\Ec2\NetworkInterface $interface */
      $interface = count($interfaces) === 1
        ? array_shift($interfaces)
        : $interfaces;

      if ($interface->getCloudContext() === $entity->getCloudContext()) {
        $options[] = [
          'value' => $interface->id(),
          'label' => "{$interface->getName()} - {$interface->getNetworkInterfaceId()}",
        ];
      }
    }

    return new JsonResponse($options);
  }

  /**
   * {@inheritdoc}
   */
  public function getVolumeTypes($cloud_context): JsonResponse {
    $dummy_options = [
      '' => 'No volume type',
      'lvmdriver-1' => 'lvmdriver-1',
      '__DEFAULT__' => '__DEFAULT__',
    ];

    $options = [];
    foreach ($dummy_options ?: [] as $key => $option) {
      $options[] = [
        'value' => $key,
        'label' => $option,
      ];
    }

    return new JsonResponse($options);
  }

  /**
   * {@inheritdoc}
   */
  public function getInstancePrivateIps(string $entity_id, string $instance_id): JsonResponse {
    if (empty($instance_id)) {
      return new JsonResponse([]);
    }
    /** @var \Drupal\openstack\Entity\OpenStackFloatingIp $entity */
    $entity = $this->entityTypeManager
      ->getStorage('openstack_floating_ip')
      ->load($entity_id);
    if (empty($entity)) {
      return new JsonResponse([
        'result' => 'NG',
        'reason' => 'The Floating IP has already been deleted.',
      ], 404);
    }

    $options = [];
    $ips = $this->openStackOperationsService->getPrivateIps($entity, $instance_id);
    foreach ($ips ?: [] as $value => $label) {
      $options[] = [
        'value' => $value,
        'label' => $label,
      ];
    }
    return new JsonResponse($options);
  }

  /**
   * {@inheritdoc}
   */
  public function getNetworkInterfacePrivateIps(string $cloud_context, string $network_interface_id): JsonResponse {
    if (empty($network_interface_id)) {
      return new JsonResponse([]);
    }

    $ips = $this->openStackRestService->getNetworkPrivateIps($network_interface_id, $cloud_context);

    $options = [];
    foreach ($ips ?: [] as $value => $label) {
      $options[] = [
        'value' => $value,
        'label' => $label,
      ];
    }
    return new JsonResponse($options);
  }

  /**
   * {@inheritdoc}
   */
  public function getAvailabilityZones($cloud_context): JsonResponse {
    $options = [];
    $params = [
      'cloud_context' => $cloud_context,
    ];

    $availability_zones = $this->entityTypeManager
      ->getStorage('openstack_availability_zone')
      ->loadByProperties($params);

    foreach ($availability_zones ?: [] as $availability_zone) {
      $options[] = [
        'value' => $availability_zone->getZoneName(),
        'label' => $availability_zone->getZoneName(),
      ];
    }

    return new JsonResponse($options);
  }

  /**
   * {@inheritdoc}
   */
  public function getInstanceOptions(string $cloud_context, string $entity_id): JsonResponse {
    /** @var \Drupal\openstack\Entity\OpenStackVolume $entity */
    $entity = $this->entityTypeManager
      ->getStorage('openstack_volume')
      ->load($entity_id);

    if (empty($entity)) {
      return new JsonResponse([
        'result' => 'NG',
        'reason' => 'The volume has already been deleted.',
      ], 404);
    }

    $account = $this->currentUser();
    $properties = [
      'availability_zone' => $entity->getAvailabilityZone(),
      'cloud_context' => $cloud_context,
    ];

    if (!$account->hasPermission("view any openstack instance")) {
      $properties['uid'] = $account->id();
    }

    $results = $this->entityTypeManager
      ->getStorage('openstack_instance')
      ->loadByProperties($properties);

    $options = [];
    foreach ($results ?: [] as $result) {
      $options[] = [
        'value' => $result->getInstanceId(),
        'label' => $this->t('@name - @instance_id', [
          '@name' => $result->getName(),
          '@instance_id' => $result->getInstanceId(),
        ]),
      ];
    }

    return new JsonResponse($options);
  }

}
