<?php

namespace Drupal\openstack\Service;

use Drupal\aws_cloud\Entity\Ec2\ElasticIpInterface;
use Drupal\aws_cloud\Entity\Ec2\ImageInterface;
use Drupal\aws_cloud\Entity\Ec2\InstanceInterface;
use Drupal\aws_cloud\Entity\Ec2\KeyPairInterface;
use Drupal\aws_cloud\Entity\Ec2\NetworkInterfaceInterface;
use Drupal\aws_cloud\Entity\Ec2\SecurityGroupInterface;
use Drupal\aws_cloud\Entity\Ec2\SnapshotInterface;
use Drupal\aws_cloud\Entity\Ec2\VolumeInterface;
use Drupal\aws_cloud\Service\AwsCloud\AwsCloudOperationsServiceInterface;
use Drupal\aws_cloud\Service\Ec2\Ec2ServiceInterface;
use Drupal\aws_cloud\Traits\AwsCloudEntityCheckTrait;
use Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface;
use Drupal\cloud\Service\CloudServiceInterface;
use Drupal\cloud\Service\EntityLinkRendererInterface;
use Drupal\cloud\Entity\CloudContentEntityBase;
use Drupal\cloud\Traits\CloudContentEntityTrait;
use Drupal\Core\Entity\EntityDeleteFormTrait;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Messenger\Messenger;
use Drupal\Core\Render\Markup;
use Drupal\openstack\Entity\OpenStackNetworkExInterface;
use Drupal\openstack\Service\Rest\OpenStackService as OpenStackRestService;
use Drupal\openstack\Service\Ec2\OpenStackService as OpenStackEc2Service;
use Symfony\Component\HttpFoundation\RequestStack;

/**
 * Interacts with OpenStack using OpenStackService API.
 */
class OpenStackOperationsService implements OpenStackOperationsServiceInterface {

  use AwsCloudEntityCheckTrait;
  use CloudContentEntityTrait;
  use EntityDeleteFormTrait;

  public const SECURITY_GROUP_DELIMITER = ', ';

  /**
   * The AWS Cloud Operations service.
   *
   * @var \Drupal\aws_cloud\Service\AwsCloud\AwsCloudOperationsServiceInterface
   */
  private $awsCloudOperationsService;

  /**
   * The cloud service provider plugin manager (CloudConfigPluginManager).
   *
   * @var \Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface
   */
  private $cloudConfigPluginManager;

  /**
   * The Cloud service.
   *
   * @var \Drupal\cloud\Service\CloudServiceInterface
   */
  private $cloudService;

  /**
   * The AWS Cloud EC2 service.
   *
   * @var \Drupal\aws_cloud\Service\Ec2\Ec2ServiceInterface
   */
  private $ec2Service;

  /**
   * The entity being used by this form.
   *
   * @var \Drupal\Core\Entity\ContentEntityInterface|\Drupal\Core\Entity\RevisionLogInterface
   */
  private $entity;

  /**
   * The Entity Type Manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  private $entityTypeManager;

  /**
   * The OpenStack Service Factory.
   *
   * @var \Drupal\openstack\Service\OpenStackServiceFactoryInterface
   */
  private $openStackServiceFactory;

  /**
   * Entity link renderer object.
   *
   * @var \Drupal\cloud\Service\EntityLinkRendererInterface
   */
  private $entityLinkRenderer;

  /**
   * Request stack.
   *
   * @var \Symfony\Component\HttpFoundation\RequestStack
   */
  protected $requestStack;

  /**
   * OpenStackOperationsService constructor.
   *
   * @param \Drupal\aws_cloud\Service\AwsCloud\AwsCloudOperationsServiceInterface $aws_cloud_operations_service
   *   The AWS Cloud Operations service.
   * @param \Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface $cloud_config_plugin_manager
   *   The cloud service provider plugin manager (CloudConfigPluginManager).
   * @param \Drupal\cloud\Service\CloudServiceInterface $cloud_service
   *   The Cloud service.
   * @param \Drupal\aws_cloud\Service\Ec2\Ec2ServiceInterface $ec2_service
   *   The AWS Cloud EC2 service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The Entity Type Manager.
   * @param \Drupal\Core\Messenger\Messenger $messenger
   *   The messenger service.
   * @param \Drupal\openstack\Service\OpenStackServiceFactoryInterface $openstack_service_factory
   *   Object for interfacing with OpenStack Service.
   * @param \Symfony\Component\HttpFoundation\RequestStack $request_stack
   *   Request stack object.
   * @param \Drupal\cloud\Service\EntityLinkRendererInterface $entity_link_renderer
   *   The entity link render service.
   */
  public function __construct(
    AwsCloudOperationsServiceInterface $aws_cloud_operations_service,
    CloudConfigPluginManagerInterface $cloud_config_plugin_manager,
    CloudServiceInterface $cloud_service,
    Ec2ServiceInterface $ec2_service,
    EntityTypeManagerInterface $entity_type_manager,
    Messenger $messenger,
    OpenStackServiceFactoryInterface $openstack_service_factory,
    RequestStack $request_stack,
    EntityLinkRendererInterface $entity_link_renderer
  ) {
    $this->awsCloudOperationsService = $aws_cloud_operations_service;
    $this->cloudConfigPluginManager = $cloud_config_plugin_manager;
    $this->cloudService = $cloud_service;
    $this->ec2Service = $ec2_service;
    $this->entityTypeManager = $entity_type_manager;
    $this->messenger = $messenger;
    $this->openStackServiceFactory = $openstack_service_factory;
    $this->requestStack = $request_stack;
    $this->entityLinkRenderer = $entity_link_renderer;
  }

  /**
   * Gets the entity of this form.
   *
   * Provided by \Drupal\Core\Entity\EntityForm.
   *
   * @return \Drupal\Core\Entity\EntityInterface
   *   The entity.
   */
  public function getEntity(): EntityInterface {
    return $this->entity;
  }

  /**
   * {@inheritdoc}
   */
  public function deleteOpenStackInstance(InstanceInterface $entity, array &$form, FormStateInterface $form_state): bool {
    try {
      $this->entity = $entity;
      // Switch OpenStack service based on $entity->getCloudContext().
      $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
      $this->awsCloudOperationsService->setEc2Service($this->ec2Service);

      if ($this->ec2Service instanceof OpenStackEc2Service) {
        return $this->awsCloudOperationsService->deleteInstance($this->entity, $form, $form_state);
      }

      if (!$this->updateEntitySubmitForm($form_state, $entity)) {
        return FALSE;
      }
      $this->ec2Service->setCloudContext($entity->getCloudContext());

      $result = $this->ec2Service->terminateInstances([
        'InstanceId' => $entity->getInstanceId(),
      ]);

      $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);

      if (empty($result)) {
        $this->processOperationErrorStatus($entity, 'deleted');
        return FALSE;
      }

      if (!empty($result['SendToWorker'])) {
        $this->messenger->addStatus($this->t('The @type @label has been deleted remotely.', [
          '@type'  => $entity->getEntityType()->getSingularLabel(),
          '@label' => $entity->getName(),
        ]));
        return TRUE;
      }

      $entity->delete();
      $this->messenger->addStatus($this->getDeletionMessage());
      $this->logDeletionMessage();
      $this->clearCacheValues($entity->getCacheTags());

      return TRUE;
    }
    catch (\Exception $e) {
      $this->handleException($e);
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function startOpenStackInstance(InstanceInterface $entity, array &$form, FormStateInterface $form_state): bool {
    try {
      $this->entity = $entity;

      // Switch OpenStack EC2 or Rest service based on
      // $entity->getCloudContext().
      $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
      $this->awsCloudOperationsService->setEc2Service($this->ec2Service);

      $params = [
        'InstanceId' => $entity->getInstanceId(),
      ];

      $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);

      $result = $this->ec2Service->startInstances($params);
      if (empty($result)) {
        return FALSE;
      }

      if (!empty($result['SendToWorker'])) {
        $this->processOperationStatus($entity, 'started remotely');
        return TRUE;
      }

      $current_state = "running";
      /** @var \Drupal\aws_cloud\Entity\Ec2\InstanceInterface */
      $instance = $this->entityTypeManager
        ->getStorage($entity->getEntityTypeId())
        ->load($entity->id());
      $instance->setInstanceState($current_state);
      $instance->save();

      $this->processOperationStatus($instance, 'started');
      $this->clearCacheValues($entity->getCacheTags());
      return TRUE;
    }
    catch (\Exception $e) {
      $this->handleException($e);
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function stopOpenStackInstance(InstanceInterface $entity, array &$form, FormStateInterface $form_state): bool {
    try {
      $this->entity = $entity;

      // Switch OpenStack EC2 or Rest service based on
      // $entity->getCloudContext().
      $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
      $this->awsCloudOperationsService->setEc2Service($this->ec2Service);

      if (!$this->updateEntitySubmitForm($form_state, $entity)) {
        return FALSE;
      }

      $params = [
        'InstanceId' => $entity->getInstanceId(),
      ];

      $result = $this->ec2Service->stopInstances($params);

      $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);

      if (empty($result)) {
        return FALSE;
      }

      if (!empty($result['SendToWorker'])) {
        $this->processOperationStatus($entity, 'stopped remotely');
        return TRUE;
      }

      $current_state = "stopped";

      /** @var \Drupal\aws_cloud\Entity\Ec2\InstanceInterface */
      $instance = $this->entityTypeManager
        ->getStorage($entity->getEntityTypeId())
        ->load($entity->id());
      $instance->setInstanceState($current_state);
      $instance->save();

      $this->processOperationStatus($instance, 'stopped');

      $this->clearCacheValues($entity->getCacheTags());
      return TRUE;
    }
    catch (\Exception $e) {
      $this->handleException($e);
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function rebootOpenStackInstance(InstanceInterface $entity, array &$form, FormStateInterface $form_state): bool {
    /** @var \Drupal\aws_cloud\Entity\Ec2\InstanceInterface */
    $this->entity = $entity;

    $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
    $this->awsCloudOperationsService->setEc2Service($this->ec2Service);

    if ($this->ec2Service instanceof OpenStackEc2Service) {
      $this->awsCloudOperationsService->setEc2Service($this->ec2Service);
      return $this->awsCloudOperationsService->rebootInstance($this->entity, $form, $form_state);
    }

    if (!$this->updateEntitySubmitForm($form_state, $entity)) {
      return FALSE;
    }

    $params = [
      'InstanceId' => $entity->getInstanceId(),
    ];

    $form_state->setRedirect("view.{$this->entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);

    $result = $this->ec2Service->rebootInstances($params);
    if (!empty($result['SendToWorker'])) {
      $this->processOperationStatus($entity, 'rebooted remotely');
      return TRUE;
    }

    $this->processOperationStatus($entity, 'rebooted');
    $this->clearCacheValues($this->entity->getCacheTags());
    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function createOpenStackImage(ImageInterface $entity, array &$form, FormStateInterface $form_state): bool {
    try {
      $this->entity = $entity;

      // Switch OpenStack service based on $entity->getCloudContext().
      $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
      $this->awsCloudOperationsService->setEc2Service($this->ec2Service);

      if (!$this->awsCloudOperationsService->createImage($this->entity, $form, $form_state)) {
        return FALSE;
      }

      if (!empty($result['SendToWorker'])) {
        return TRUE;
      }

      if ($this->ec2Service instanceof OpenStackRestService) {
        $this->awsCloudOperationsService->updateNameAndCreatedByTags($entity, $entity->getImageId());
      }

      return TRUE;
    }
    catch (\Exception $e) {
      $this->handleException($e);
      return FALSE;
    }
  }

  /**
   * Save an Image.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @throws \Drupal\Core\Entity\EntityMalformedException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  private function saveImage(array $form, FormStateInterface $form_state): void {
    /** @var \Drupal\aws_cloud\Entity\Ec2\ImageInterface */
    $entity = $this->entity;

    /** @var \Drupal\aws_cloud\Entity\Ec2\ImageInterface */
    $original_entity = $this->entityTypeManager
      ->getStorage($entity->getEntityTypeId())
      ->load($entity->id());

    $this->awsCloudOperationsService->saveAwsCloudContent($entity, $form, $form_state);

    $this->awsCloudOperationsService->updateNameAndCreatedByTags($entity, $entity->getImageId());

    $this->cloudConfigPluginManager->setCloudContext($entity->getCloudContext());
    $cloud_config = $this->cloudConfigPluginManager->loadConfigEntity();

    // Call 'modifyImageAttribute'
    // only when OwnerId is same as Account ID of Cloud Config.
    // It's to avoid AuthFailure error from AWS API.
    if ($cloud_config->get('field_account_id')->value !== $entity->getAccountId()) {
      return;
    }

    $form_values = $form_state->getValues();
    $visibility = $form_values['visibility'] ?? 0;

    $permission = [
      [
        'Group' => 'all',
        'UserId' => $entity->getAccountId(),
      ],
    ];

    // 1 => Public, 0 => Private.
    $launch_permission =
      $visibility
        ? ['Add' => $permission]
        : ['Remove' => $permission];

    // Update image.
    $this->ec2Service->setCloudContext($entity->getCloudContext());

    if (!method_exists($this->ec2Service, 'modifyImageAttribute')) {
      return;
    }

    // Update image description.
    $this->ec2Service->modifyImageAttribute([
      'ImageId' => $entity->getImageId(),
      'Description' => ['Value' => $entity->getDescription()],
    ]);

    // These fields (Description , LaunchPermission , or ProductCode) cannot
    // be specified at the same time.
    // See also:
    // https://docs.aws.amazon.com/cli/latest/reference/ec2/modify-image-attribute.html
    // Update image launch permission.
    $this->ec2Service->modifyImageAttribute([
      'ImageId' => $entity->getImageId(),
      'LaunchPermission' => $launch_permission,
    ]);

    $new_account_ids = [];
    foreach ($entity->getLaunchPermissionAccountIds() ?: [] as $account_id) {
      if (empty($account_id->value)) {
        continue;
      }
      $new_account_ids[] = $account_id->value;
    }

    $old_account_ids = [];
    foreach ($original_entity->getLaunchPermissionAccountIds() ?: [] as $account_id) {
      if (empty($account_id->value)) {
        continue;
      }
      $old_account_ids[] = $account_id->value;
    }

    $account_ids_to_remove = array_diff($old_account_ids, $new_account_ids);
    $account_ids_to_add = array_diff($new_account_ids, $old_account_ids);

    if (!empty($account_ids_to_remove)) {
      $this->ec2Service->modifyImageAttribute([
        'ImageId' => $entity->getImageId(),
        'LaunchPermission' => [
          'Remove' => array_map(function ($item) {
            return ['UserId' => $item];
          }, $account_ids_to_remove),
        ],
      ]);
    }

    if (!empty($account_ids_to_add)) {
      $this->ec2Service->modifyImageAttribute([
        'ImageId' => $entity->getImageId(),
        'LaunchPermission' => [
          'Add' => array_map(function ($item) {
            return ['UserId' => $item];
          }, $account_ids_to_add),
        ],
      ]);
    }

    $this->clearCacheValues($entity->getCacheTags());
    $this->dispatchSaveEvent($entity);
  }

  /**
   * {@inheritdoc}
   */
  public function editOpenStackImage(ImageInterface $entity, array &$form, FormStateInterface $form_state): bool {
    try {
      $this->entity = $entity;
      $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
      $this->awsCloudOperationsService->setEc2Service($this->ec2Service);
      $this->saveImage($form, $form_state);

      $visibility = !empty($entity->getVisibility()) ? 'public' : 'private';

      // Update Image of OpenStack REST API.
      if ($this->ec2Service instanceof OpenStackRestService) {
        // Make an API call to update image.
        $result = $this->ec2Service->updateImage([
          'ImageId' => $entity->getImageId(),
          'Name' => $entity->getName(),
          'Visibility' => $visibility,
        ]);

        $old_tags = $result['Tags'] ?? [];

        foreach ($old_tags ?: [] as $tag) {
          // Delete old tags.
          $this->ec2Service->deleteTags(
            [
              'ImageId'    => $entity->getImageId(),
              'Tag'        => $tag,
              'EntityType' => $entity->getEntityTypeId(),
            ]);
        }

        $this->awsCloudOperationsService->updateNameAndCreatedByTags($entity, $entity->getImageId());
      }

      return TRUE;
    }
    catch (\Exception $e) {
      $this->handleException($e);
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function deleteOpenStackImage(ImageInterface $entity, array &$form, FormStateInterface $form_state): bool {
    try {
      $this->entity = $entity;
      $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
      $this->awsCloudOperationsService->setEc2Service($this->ec2Service);
      return $this->awsCloudOperationsService->deleteImage($this->entity, $form, $form_state);
    }
    catch (\Exception $e) {
      $this->handleException($e);
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function createOpenStackKeyPair(KeyPairInterface $entity, array &$form, FormStateInterface $form_state): bool {
    try {
      $this->entity = $entity;
      $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
      $this->awsCloudOperationsService->setEc2Service($this->ec2Service);
      return $this->awsCloudOperationsService->createKeyPair($this->entity, $form, $form_state);
    }
    catch (\Exception $e) {
      $this->handleException($e);
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function editOpenStackKeyPair(KeyPairInterface $entity, array &$form, FormStateInterface $form_state): bool {
    try {
      $this->entity = $entity;
      $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
      $this->awsCloudOperationsService->setEc2Service($this->ec2Service);
      return $this->awsCloudOperationsService->editKeyPair($this->entity, $form, $form_state);
    }
    catch (\Exception $e) {
      $this->handleException($e);
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function deleteOpenStackKeyPair(KeyPairInterface $entity, array &$form, FormStateInterface $form_state): bool {
    try {
      $this->entity = $entity;
      $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
      $this->awsCloudOperationsService->setEc2Service($this->ec2Service);
      return $this->awsCloudOperationsService->deleteKeyPair($this->entity, $form, $form_state);
    }
    catch (\Exception $e) {
      $this->handleException($e);
      return FALSE;
    }
  }

  /**
   * Helper function to update field tags.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state object.
   */
  private function updateTagsField(FormStateInterface $form_state): void {
    /** @var \Drupal\aws_cloud\Entity\Ec2\InstanceInterface */
    $entity = $this->entity;

    $form_tags = $form_state->getValue('tags');
    usort($form_tags, static function ($a, $b) {
      return (int) $a['_weight'] - (int) $b['_weight'];
    });
    /** @var \Drupal\openstack\Service\OpenStackServiceInterface|\Drupal\cloud\Service\CloudResourceTagInterface $ec2_service */
    $ec2_service = $this->openStackServiceFactory->get($entity->getCloudContext());
    $this->ec2Service = $ec2_service;
    $uid_key_name = $ec2_service->getTagKeyCreatedByUid(
      'openstack',
      $entity->getCloudContext()
    );

    $fixed_tags = [];
    $fixed_tags[$uid_key_name] = $entity->getOwner() === NULL
    ? 0
    : $entity->getOwner()->id();

    $tags = [];
    foreach ($form_tags ?: [] as $form_tag) {
      $item_key = $form_tag['item_key'];
      if ($item_key === '') {
        continue;
      }

      // Skip special tags.
      if (strpos($item_key, 'openstack:') === 0) {
        continue;
      }

      $item_value = $form_tag['item_value'];

      if ($form_tag['item_key'] !== $uid_key_name) {
        $tags[] = ['item_key' => $item_key, 'item_value' => $item_value];
      }
    }

    if (!empty($fixed_tags)) {
      foreach ($fixed_tags ?: [] as $item_key => $item_value) {
        if ($item_value) {
          $tags[] = ['item_key' => $item_key, 'item_value' => $item_value];
        }
      }
    }

    $entity->setTags($tags);
    $entity->save();
  }

  /**
   * Helper function to update OpenStack tags.
   */
  private function updateOpenStackTags(): void {
    /** @var \Drupal\aws_cloud\Entity\Ec2\InstanceInterface */
    $entity = $this->entity;
    /** @var \Drupal\openstack\Service\Rest\OpenStackServiceInterface $ec2_service */
    $ec2_service = $this->ec2Service;

    $params = [
      'Resources' => [$entity->getInstanceId()],
    ];

    // Delete old tags.
    $this->ec2Service->deleteTags(
      [
        'InstanceId' => $entity->getInstanceId(),
        'EntityType' => $entity->getEntityTypeId(),
        'UidKeyName' => $ec2_service->getTagKeyCreatedByUid(
          'openstack',
          $entity->getCloudContext()
        ),
      ]);

    foreach ($entity->getTags() ?: [] as $tag) {
      $params['Tags'][] = [
        'Key' => $tag['item_key'],
        'Value' => $tag['item_value'],
      ];
    }

    // Create Tags with different parameters for OpenStack.
    $this->ec2Service->createTags([
      'Resources' => [$entity->getInstanceId()],
      'Tags' => $params['Tags'],
      'EntityType' => $entity->getEntityTypeId(),
    ]);

    $this->ec2Service->updateInstances();

  }

  /**
   * {@inheritdoc}
   */
  public function editOpenStackInstance(InstanceInterface $entity, array $form, FormStateInterface $form_state): bool {
    try {
      $this->entity = $entity;
      $this->awsCloudOperationsService->copyFormItemValues($this->entity, $form);
      // Switch OpenStack EC2 or Rest service
      // based on $entity->getCloudContext().
      $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
      $this->awsCloudOperationsService->setEc2Service($this->ec2Service);

      if (!($this->ec2Service instanceof OpenStackRestService)) {
        return $this->awsCloudOperationsService->editInstance($this->entity, $form, $form_state);
      }

      $termination_timestamp = $form_state->getValue('termination_timestamp')[0]['value'] ?? '';

      $this->updateTagsField($form_state);

      $this->updateOpenStackTags();

      if ($termination_timestamp === NULL) {
        // Unset the termination timestamp.
        $entity->set('termination_timestamp', NULL);
      }

      // Update Instance of OpenStack REST API.
      $result = $this->ec2Service->modifyLaunchTemplate([
        'InstanceId' => $entity->getInstanceId(),
        'Name' => $entity->getName(),
      ]);

      if (!empty($result['SendToWorker'])) {
        $this->processOperationStatus($entity, 'updated remotely');
        return TRUE;
      }

      $entity->save();

      $this->processOperationStatus($entity, 'updated');

      $this->clearCacheValues($entity->getCacheTags());
      return TRUE;
    }
    catch (\Exception $e) {
      $this->handleException($e);
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function createOpenStackNetworkInterface(NetworkInterfaceInterface $entity, array &$form, FormStateInterface $form_state): bool {
    try {
      $this->entity = $entity;
      $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
      $this->awsCloudOperationsService->setEc2Service($this->ec2Service);
      return $this->awsCloudOperationsService->createNetworkInterface($this->entity, $form, $form_state);
    }
    catch (\Exception $e) {
      $this->handleException($e);
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function createOpenStackNetwork(OpenStackNetworkExInterface $entity, array &$form, FormStateInterface $form_state): bool {
    $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
    $this->awsCloudOperationsService->setEc2Service($this->ec2Service);

    $this->awsCloudOperationsService->trimTextfields($entity, $form, $form_state);

    $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);

    $zone_names = [];
    foreach ($entity->getAvailabilityZones() ?: [] as $availability_zone) {
      $zone_names[] = array_keys($availability_zone)[0];
    }

    $this->ec2Service->setCloudContext($entity->getCloudContext());
    $result = $this->ec2Service->createNetwork([
      'Name' => $entity->getName(),
      'AdminStateUp' => $entity->getAdminStateUp(),
      'Shared' => $entity->getShared(),
      'AvailabilityZones' => $zone_names,
    ]);

    if ($result === NULL) {
      $this->processOperationErrorStatus($entity, 'created');
      return FALSE;
    }

    if (!empty($result['SendToWorker'])) {
      $this->processOperationStatus($entity, 'created remotely');
      return TRUE;
    }

    $entity->setNetworkId($result['Network']['NetworkId']);
    $entity->save();

    if ($this->ec2Service instanceof OpenStackRestService) {
      $this->awsCloudOperationsService->updateNameAndCreatedByTags($entity, $entity->getNetworkId());
    }

    // Update the Network entity.
    $this->ec2Service->updateNetworks([
      'NetworkId' => $entity->getNetworkId(),
    ]);

    $this->processOperationStatus($entity, 'created');

    $this->clearCacheValues($entity->getCacheTags());
    $this->dispatchSaveEvent($entity);

    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function deleteOpenStackNetworkInterface(NetworkInterfaceInterface $entity, array &$form, FormStateInterface $form_state): bool {
    try {
      $this->entity = $entity;
      $this->ec2Service = $this->openStackServiceFactory->get($this->entity->getCloudContext());
      $this->awsCloudOperationsService->setEc2Service($this->ec2Service);
      return $this->awsCloudOperationsService->deleteNetworkInterface($this->entity, $form, $form_state);
    }
    catch (\Exception $e) {
      $this->handleException($e);
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function editOpenStackNetworkInterface(NetworkInterfaceInterface $entity, array &$form, FormStateInterface $form_state): bool {
    try {
      $this->entity = $entity;
      $this->ec2Service = $this->openStackServiceFactory->get($this->entity->getCloudContext());
      $this->awsCloudOperationsService->setEc2Service($this->ec2Service);
      $this->saveNetworkInterface($form, $form_state);
      return TRUE;
    }
    catch (\Exception $e) {
      $this->handleException($e);
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function editOpenStackNetwork(OpenStackNetworkExInterface $entity, array &$form, FormStateInterface $form_state): bool {
    $this->entity = $entity;

    $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
    $this->awsCloudOperationsService->setEc2Service($this->ec2Service);

    $this->awsCloudOperationsService->trimTextfields($entity, $form, $form_state);

    $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);

    $this->ec2Service->setCloudContext($entity->getCloudContext());
    $result = $this->ec2Service->updateNetwork([
      'NetworkId' => $entity->getNetworkId(),
      'Name' => $entity->getName(),
      'AdminStateUp' => $entity->getAdminStateUp(),
      'Shared' => $entity->getShared(),
    ]);

    if ($result === NULL) {
      $this->processOperationErrorStatus($entity, 'updated');
      return FALSE;
    }

    if (!empty($result['SendToWorker'])) {
      $this->processOperationStatus($entity, 'updated remotely');
      return TRUE;
    }

    $entity->save();

    if ($this->ec2Service instanceof OpenStackRestService) {
      $this->awsCloudOperationsService->updateNameAndCreatedByTags($entity, $entity->getNetworkId());
    }

    // Update the Network entity.
    $this->ec2Service->updateNetworks([
      'NetworkId' => $entity->getNetworkId(),
    ]);

    $this->processOperationStatus($entity, 'updated');

    $this->clearCacheValues($entity->getCacheTags());
    $this->dispatchSaveEvent($entity);

    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function deleteOpenStackNetwork(OpenStackNetworkExInterface $entity, array &$form, FormStateInterface $form_state): bool {
    $this->entity = $entity;

    $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
    $this->awsCloudOperationsService->setEc2Service($this->ec2Service);

    $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);

    $this->ec2Service->setCloudContext($entity->getCloudContext());
    $result = $this->ec2Service->deleteNetwork([
      'NetworkId' => $entity->getNetworkId(),
    ]);

    if ($result === NULL) {
      $this->processOperationErrorStatus($entity, 'deleted');
      return FALSE;
    }

    if (!empty($result['SendToWorker'])) {
      $this->processOperationStatus($entity, 'deleted remotely');
      return TRUE;
    }

    $entity->delete();

    $this->messenger->addStatus($this->getDeletionMessage());
    $this->logDeletionMessage();

    $this->clearCacheValues($entity->getCacheTags());
    $this->dispatchSaveEvent($entity);

    return TRUE;
  }

  /**
   * Save a Network Interface.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @throws \Drupal\Core\Entity\EntityMalformedException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  private function saveNetworkInterface(array $form, FormStateInterface $form_state): void {
    /** @var \Drupal\aws_cloud\Entity\Ec2\NetworkInterface */
    $entity = $this->entity;

    $this->ec2Service->setCloudContext($entity->getCloudContext());

    $this->awsCloudOperationsService->saveAwsCloudContent($entity, $form, $form_state);

    $params = [
      'NetworkInterfaceId' => $entity->getNetworkInterfaceId(),
      'Description' => ['Value' => $entity->getDescription()],
    ];

    if (method_exists($this->ec2Service, 'modifyNetworkInterfaceAttribute')) {
      $this->ec2Service->modifyNetworkInterfaceAttribute($params);
    }

    $this->awsCloudOperationsService->updateNameAndCreatedByTags($entity, $entity->getNetworkInterfaceId());

    // Update OpenStack REST NetworkInterfaces.
    if ($this->ec2Service instanceof OpenStackRestService) {
      // Delete old tags.
      $this->ec2Service->deleteTags(
        [
          'NetworkInterfaceId' => $entity->getNetworkInterfaceId(),
          'EntityType'         => $entity->getEntityTypeId(),
        ]);
      $this->awsCloudOperationsService->updateNameAndCreatedByTags($entity, $entity->getNetworkInterfaceId());

      $this->ec2Service->updateNetworkInterface([
        'NetworkInterfaceId' => $entity->getNetworkInterfaceId(),
        'Name' => $entity->getName(),
        'Description' => $entity->getDescription(),
      ]);
    }

    $this->clearCacheValues($entity->getCacheTags());
    $this->dispatchSaveEvent($entity);
  }

  /**
   * {@inheritdoc}
   */
  public function createOpenStackFloatingIp(ElasticIpInterface $entity, array &$form, FormStateInterface $form_state): bool {
    $this->entity = $entity;

    // Switch OpenStack EC2 or REST service based on $entity->getCloudContext().
    $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
    $this->awsCloudOperationsService->setEc2Service($this->ec2Service);

    if (!($this->ec2Service instanceof OpenStackRestService)) {
      return $this->awsCloudOperationsService->createElasticIp($this->entity, $form, $form_state);
    }

    $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
    $params = [
      'AllocationId' => $form_state->getValue('floating_network_id'),
    ];

    $result = $this->ec2Service->allocateAddress($params);

    if (empty($result)) {
      $this->processOperationErrorStatus($entity, 'created');
      return FALSE;
    }

    if (!empty($result['SendToWorker'])) {
      $this->processOperationStatus($entity, 'created remotely');
      return TRUE;
    }

    $floating_network_id = $form_state->getValue('floating_network_id');
    $entity->setAllocationId($floating_network_id);
    $entity->save();

    $public_ip = $result['PublicIp'] ?? NULL;
    $floating_ip_type = !empty($result['PublicIp']) ? 'Public IP' : '';
    /** @var \Drupal\openstack\Entity\OpenStackFloatingIp $entity */
    if (!empty($public_ip)
      && $entity->setFloatingIpId($result['FloatingIpId'] ?? '')
      && $entity->setPublicIp($public_ip)
      && $entity->setElasticIpType($floating_ip_type)
      && $entity->save()) {

      // Create uid tag.
      $this->awsCloudOperationsService->updateNameAndCreatedByTags($entity, $entity->getFloatingIpId());
      $this->processOperationStatus($entity, 'created');
      $this->clearCacheValues($entity->getCacheTags());
    }
    else {
      $this->processOperationErrorStatus($entity, 'created');
    }
    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function editOpenStackFloatingIp(ElasticIpInterface $entity, array &$form, FormStateInterface $form_state): bool {
    $this->entity = $entity;

    // Copy values of the form elements whose type are item to entity.
    // If not, the properties corresponding to the form elements
    // will be saved as NULL.
    $this->awsCloudOperationsService->copyFormItemValues($this->entity, $form);

    $this->awsCloudOperationsService->trimTextfields($this->entity, $form, $form_state);

    // Switch OpenStack EC2 or REST service based on $entity->getCloudContext().
    $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
    $this->awsCloudOperationsService->setEc2Service($this->ec2Service);
    if (!$this->ec2Service instanceof OpenStackRestService) {
      $this->saveFloatingIp($form, $form_state);
      return TRUE;
    }

    // Delete old tags.
    /** @var \Drupal\openstack\Entity\OpenStackFloatingIp $entity */
    $this->ec2Service->deleteTags(
      [
        'FloatingIpId' => $entity->getFloatingIpId(),
        'EntityType'   => $entity->getEntityTypeId(),
      ]);

    $this->awsCloudOperationsService->updateNameAndCreatedByTags($entity, $entity->getFloatingIpId());

    $this->cloudConfigPluginManager->setCloudContext($entity->getCloudContext());
    $cloud_config = $this->cloudConfigPluginManager->loadConfigEntity();
    if ($cloud_config->isRemote()) {
      $this->processOperationStatus($entity, 'updated remotely');
      $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
      return TRUE;
    }

    if ($entity->save()) {
      $this->processOperationStatus($entity, 'updated');
      $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
      return TRUE;
    }

    $this->processOperationErrorStatus($entity, 'updated');
    return TRUE;
  }

  /**
   * Save a floating IP.
   *
   * @throws \Drupal\Core\Entity\EntityMalformedException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  private function saveFloatingIp(): void {
    /** @var \Drupal\aws_cloud\Entity\Ec2\ElasticIp */
    $entity = $this->entity;

    $this->entity->save();

    $this->ec2Service->setCloudContext($entity->getCloudContext());

    // Update the name.
    $this->awsCloudOperationsService->updateNameAndCreatedByTags($this->entity, $entity->getAllocationId());

    $this->processOperationStatus($this->entity, 'updated');

    $this->clearCacheValues($this->entity->getCacheTags());
    $this->dispatchSaveEvent($this->entity);
  }

  /**
   * {@inheritdoc}
   */
  public function deleteOpenStackFloatingIp(ElasticIpInterface $entity, array &$form, FormStateInterface $form_state): bool {
    $this->entity = $entity;
    $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
    $this->awsCloudOperationsService->setEc2Service($this->ec2Service);

    if ($this->ec2Service instanceof OpenStackEc2Service) {
      return $this->awsCloudOperationsService->deleteElasticIp($this->entity, $form, $form_state);
    }

    /** @var \Drupal\openstack\Entity\OpenStackFloatingIp $entity */
    $params = ['FloatingIpId' => $entity->getFloatingIpId()];

    $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);

    if (empty($entity)
      || empty($params)
      || empty($this->ec2Service->releaseAddress($params))) {
      $this->processOperationErrorStatus($entity, 'deleted');
      return FALSE;
    }

    $entity->delete();

    $this->messenger->addStatus($this->getDeletionMessage());
    $this->logDeletionMessage();
    $this->clearCacheValues($entity->getCacheTags());
    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function disassociateOpenStackFloatingIp(ElasticIpInterface $entity, array &$form, FormStateInterface $form_state): bool {
    $this->entity = $entity;
    // Switch OpenStack EC2 or REST service based on $entity->getCloudContext().
    $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());

    if (!($this->ec2Service instanceof OpenStackRestService)) {
      return $this->disassociateFloatingIp($form, $form_state);
    }

    // Get module name.
    $module_name = $this->getModuleName($entity);

    $instance_id = $entity->getInstanceId();
    $network_interface_id = $entity->getNetworkInterfaceId();

    $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);

    /** @var \Drupal\openstack\Entity\OpenStackFloatingIp $entity */
    $result = $this->ec2Service->disassociateAddress([
      'FloatingIpId' => $entity->getFloatingIpId(),
      'AssociationId' => $entity->getAssociationId(),
    ]);

    if (empty($result)) {
      $this->messenger->addError($this->t('Unable to disassociate Floating IP.'));
      return FALSE;
    }

    if (!empty($instance_id)) {
      $instance = $this->getInstanceById($instance_id, $module_name);
      $instance_link = $this->entityLinkRenderer->renderViewElement(
      $instance_id,
      "{$module_name}_instance",
      'instance_id',
      [],
      !empty($instance->getName()) !== $instance->getInstanceId()
        ? $this->t('@instance_name (@instance_id)', [
          '@instance_name' => $instance->getName(),
          '@instance_id' => $instance_id,
        ])
        : $instance_id
      );
    }

    if (!empty($network_interface_id)) {
      $network_interface = $this->getNetworkInterfaceById($network_interface_id, $module_name);
      $network_interface_link = $this->entityLinkRenderer->renderViewElement(
      $network_interface_id,
      "{$module_name}_network_interface",
      'network_interface_id',
      [],
      $network_interface->getName() !== $network_interface->getNetworkInterfaceId()
        ? $this->t('@network_interface_name (@network_interface_id)', [
          '@network_interface_name' => $network_interface->getName(),
          '@network_interface_id' => $network_interface_id,
        ])
        : $network_interface_id
      );
    }

    if (!empty($result['SendToWorker'])) {
      $this->messenger->addStatus($this->t('Floating IP disassociated remotely from: <ul><li>
        Instance: @instance_id </li> <li>Network: @network_id</li></ul>', [
          '@instance_id' => Markup::create($instance_link['#markup']),
          '@network_id' => Markup::create($network_interface_link['#markup']),
        ]));
      return TRUE;
    }

    $this->messenger->addStatus($this->t('Floating IP disassociated from: <ul><li>
      Instance: @instance_id </li> <li>Network: @network_id</li></ul>', [
        '@instance_id' => Markup::create($instance_link['#markup']),
        '@network_id' => Markup::create($network_interface_link['#markup']),
      ]));
    $this->ec2Service->updateFloatingIps();
    $this->ec2Service->updateInstances();
    $this->ec2Service->updateNetworkInterfaces();

    $this->clearCacheValues($entity->getCacheTags());

    // Refresh the instances.
    $this->ec2Service->updateInstances();
    $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
    return TRUE;
  }

  /**
   * Disassociate a floating IP.
   *
   * @param array $form
   *   Array of form object.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @throws \Drupal\Core\Entity\EntityMalformedException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  private function disassociateFloatingIp(array $form, FormStateInterface $form_state): bool {
    /** @var \Drupal\aws_cloud\Entity\Ec2\ElasticIpInterface @entity */
    $entity = $this->entity;
    $this->ec2Service->setCloudContext($entity->getCloudContext());

    $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);

    $result = $this->ec2Service->disassociateAddress([
      'AssociationId' => $entity->getAssociationId(),
    ]);

    if ($result === NULL) {
      $this->messenger->addError($this->t('Unable to disassociate Elastic IP.'));
      return FALSE;
    }

    // Get module name.
    $module_name = $this->getModuleName($entity);

    $instance_id = $entity->getInstanceId();
    $network_interface_id = $entity->getNetworkInterfaceId();

    if (!empty($instance_id)) {
      $instance = $this->getInstanceById($instance_id, $module_name, $entity->getCloudContext());
      $instance_link = $this->entityLinkRenderer->renderViewElement(
      $instance_id,
      "{$module_name}_instance",
      'instance_id',
      [],
      !empty($instance->getName()) !== $instance->getInstanceId()
        ? $this->t('@instance_name (@instance_id)', [
          '@instance_name' => $instance->getName(),
          '@instance_id' => $instance_id,
        ])
        : $instance_id
      );
    }
    if (!empty($network_interface_id)) {
      $network_interface = $this->getNetworkInterfaceById($network_interface_id, $module_name, $entity->getCloudContext());
      $network_interface_link = $this->entityLinkRenderer->renderViewElement(
      $network_interface_id,
      "{$module_name}_network_interface",
      'network_interface_id',
      [],
      $network_interface->getName() !== $network_interface->getNetworkInterfaceId()
        ? $this->t('@network_interface_name (@network_interface_id)', [
          '@network_interface_name' => $network_interface->getName(),
          '@network_interface_id' => $network_interface_id,
        ])
        : $network_interface_id
      );
    }

    if (!empty($result['SendToWorker'])) {
      $this->messenger->addStatus($this->t('Elastic IP disassociated from remotely: <ul><li>
        Instance: @instance_id </li> <li>Network: @network_id</li></ul>', [
          '@instance_id' => Markup::create($instance_link['#markup']),
          '@network_id' => Markup::create($network_interface_link['#markup']),
        ]));
      return TRUE;
    }

    $this->messenger->addStatus($this->t('Elastic IP disassociated from: <ul><li>
      Instance: @instance_id </li> <li>Network: @network_id</li></ul>', [
        '@instance_id' => Markup::create($instance_link['#markup']),
        '@network_id' => Markup::create($network_interface_link['#markup']),
      ]));
    $this->ec2Service->updateElasticIps();
    $this->ec2Service->updateInstances();
    $this->ec2Service->updateNetworkInterfaces();

    $this->clearCacheValues($entity->getCacheTags());
    $this->dispatchSubmitEvent($entity);
    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function createOpenStackSecurityGroup(SecurityGroupInterface $entity, array &$form, FormStateInterface $form_state): bool {
    try {
      $this->entity = $entity;
      $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
      $this->awsCloudOperationsService->setEc2Service($this->ec2Service);
      return $this->awsCloudOperationsService->createSecurityGroup($entity, $form, $form_state);
    }
    catch (\Exception $e) {
      $this->handleException($e);
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function deleteOpenStackSecurityGroup(SecurityGroupInterface $entity, array &$form, FormStateInterface $form_state): bool {
    try {
      $this->entity = $entity;
      $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
      $this->awsCloudOperationsService->setEc2Service($this->ec2Service);
      return $this->awsCloudOperationsService->deleteSecurityGroup($this->entity, $form, $form_state);
    }
    catch (\Exception $e) {
      $this->handleException($e);
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function revokeOpenStackSecurityGroup(
    SecurityGroupInterface $entity,
    array &$form,
    FormStateInterface $form_state
  ): bool {
    $this->entity = $entity;
    // Switch OpenStack EC2 or REST service based on $entity->getCloudContext().
    $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());

    $referer = $this->requestStack->getCurrentRequest()->headers->get('referer');
    $url = parse_url($referer);
    parse_str($url['query'], $results);
    $type = $results['type'];
    $position = $results['position'];

    // Use parent save function in case of OpenStack EC2.
    if (!($this->ec2Service instanceof OpenStackRestService)) {
      return $this->awsCloudOperationsService->revokeSecurityGroup($entity, $form, $form_state, $type, $position);
    }

    /** @var \Drupal\cloud\Entity\CloudContentEntityBase|\Drupal\aws_cloud\Entity\Ec2\SecurityGroupInterface $entity */
    $permission = $this->awsCloudOperationsService->getPermission($type, $position, $entity);
    if ($permission !== FALSE) {

      $form_state->setRedirect("entity.{$entity->getEntityTypeId()}.canonical", [
        'cloud_context' => $entity->getCloudContext(),
        $entity->getEntityTypeId() => $entity->id(),
      ]);

      // Revoke security group rule.
      $perm_array['IpPermissions'][] = [
        'RuleId' => $permission->rule_id,
      ];
      $param_array = [
        'RuleId' => $perm_array['IpPermissions'][0]['RuleId'],
      ];
      $result = $this->ec2Service->revokeSecurityGroupRule($param_array);

      if (!empty($result['SendToWorker'])) {
        $this->messenger->addStatus($this->t('Permission revoked remotely'));
        return TRUE;
      }

      // Have the system refresh the security group.
      $this->ec2Service->updateSecurityGroups([
        'GroupIds' => [$entity->getGroupId()],
      ], FALSE);

      $this->validateRevoke($type, $entity)
        ? $this->messenger->addStatus($this->t('Permission revoked'))
        : $this->messenger->addError($this->t('Permission not revoked'));

      $this->clearCacheValues($entity->getCacheTags());
    }

    return TRUE;
  }

  /**
   * Verify that the revoke was successful.
   *
   * Since OpenStack REST does not return any error codes in the revoke API
   * calls, the only way to verify is to count the permissions array from the
   * current entity, and the entity that is newly updated from the
   * updateSecurityGroups API call.
   *
   * @param string $type
   *   The type of permission.
   * @param \Drupal\cloud\Entity\CloudContentEntityBase $entity
   *   The SecurityGroup.
   *
   * @return bool
   *   True or false.
   */
  private function validateRevoke(string $type, CloudContentEntityBase $entity): bool {
    /** @var \Drupal\openstack\Entity\OpenStackSecurityGroup $updated_group */
    $updated_group = $this->entityTypeManager
      ->getStorage($entity->getEntityTypeId())
      ->load($entity->id());

    /** @var \Drupal\aws_cloud\Entity\Ec2\SecurityGroupInterface $entity */
    return $type === 'outbound_permission'
      ? $entity->getOutboundPermission()->count() === $updated_group->getOutboundPermission()->count()
      : $entity->getIpPermission()->count() === $updated_group->getIpPermission()->count();
  }

  /**
   * {@inheritdoc}
   */
  public function editOpenStackSecurityGroup(SecurityGroupInterface $entity, array $form, FormStateInterface $form_state): bool {
    try {
      $this->entity = $entity;

      // Switch OpenStack EC2 or REST service based on $cloud_context.
      $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
      $this->awsCloudOperationsService->setEc2Service($this->ec2Service);
      if ($this->ec2Service instanceof OpenStackEc2Service) {
        return $this->awsCloudOperationsService->editSecurityGroup($entity, $form, $form_state);
      }
      /** @var \Drupal\openstack\Service\Rest\OpenStackServiceInterface $ec2_service */
      $ec2_service = $this->ec2Service;
      // Call copyFormItemValues() to ensure the form array is intact.
      $this->awsCloudOperationsService->copyFormItemValues($entity, $form);

      $this->awsCloudOperationsService->trimTextfields($entity, $form, $form_state);

      $ec2_service->setCloudContext($entity->getCloudContext());

      // Delete old tags.
      $ec2_service->deleteTags(
        [
          'SecurityGroupId' => $entity->getGroupId(),
          'EntityType'      => $entity->getEntityTypeId(),
        ]);

      $this->awsCloudOperationsService->updateNameAndCreatedByTags($entity, $entity->getGroupId());

      /** @var \Drupal\aws_cloud\Entity\Ec2\SecurityGroupInterface $existing_group */
      $existing_group = $this->entityTypeManager
        ->getStorage($entity->getEntityTypeId())
        ->load($entity->id());

      if (!empty($existing_group)) {

        // Update the inbound/outbound permissions.
        $existing_permissions_inbound = $existing_group->getIpPermission() ?? [];
        // Clear out the existing permissions.
        foreach ($existing_permissions_inbound ?: [] as $inbound_permission) {
          if (empty($inbound_permission->rule_id)) {
            continue;
          }

          $ip_param_array = [
            'RuleId' => $inbound_permission->rule_id,
          ];
          $ec2_service->revokeSecurityGroupRule($ip_param_array);
        }

        // Update the inbound/outbound permissions.
        $existing_permissions_outbound = $existing_group->getOutboundPermission() ?? [];
        // Clear out the existing permissions.
        foreach ($existing_permissions_outbound ?: [] as $outbound_permission) {
          if (empty($outbound_permission->rule_id)) {
            continue;
          }

          $ip_param_array = [
            'RuleId' => $outbound_permission->rule_id,
          ];
          $ec2_service->revokeSecurityGroupRule($ip_param_array);
        }

        // Have the system refresh the security group.
        $ec2_service->updateSecurityGroups([
          'GroupIds' => [$entity->getGroupId()],
        ], FALSE);

        $this->clearCacheValues($entity->getCacheTags());
      }
      else {
        $this->messenger->addError($this->t('Cannot update security group.'));
      }

      if ($entity->save()) {

        $ip_perm_arr = ($entity->get('ip_permission') !== NULL)
          ? $entity->get('ip_permission')->getValue() : [];
        $outbound_perm_arr = ($entity->get('outbound_permission') !== NULL)
          ? $entity->get('outbound_permission')->getValue() : [];

        $ec2_service->updateSecurityGroup([
          'GroupId' => $entity->getGroupId(),
          'GroupName' => $entity->getName(),
        ]);

        // Create inbound security group rule.
        foreach ($ip_perm_arr ?: [] as $ip_perm) {
          if (!empty($ip_perm)) {
            $inbound_rule_param = [
              'Direction' => 'ingress',
              'FromPort' => $ip_perm['from_port'],
              'ToPort' => $ip_perm['to_port'],
              'IpProtocol' => $ip_perm['ip_protocol'],
              'GroupId' => $entity->getGroupId(),
              'Description' => $ip_perm['description'],
            ];

            switch ($ip_perm['source']) {
              case 'ip4':
                $inbound_rule_param += [
                  'Source' => 'IPv4',
                  'CidrIp' => $ip_perm['cidr_ip'],
                ];
                $ec2_service->createSecurityGroupRule($inbound_rule_param);
                break;

              case 'group':
                $inbound_rule_param += [
                  'Source' => 'IPv4',
                  'RemoteGroupId' => $ip_perm['group_id'],
                  'RemoteUserId' => $ip_perm['user_id'],
                ];
                $ec2_service->createSecurityGroupRule($inbound_rule_param);
                break;

              case 'ip6':
                $inbound_rule_param += [
                  'Source' => 'IPv6',
                  'CidrIpv6' => $ip_perm['cidr_ip_v6'],
                ];
                $ec2_service->createSecurityGroupRule($inbound_rule_param);
                break;
            }
          }
        }

        // Create outbound security group rule.
        foreach ($outbound_perm_arr ?: [] as $outbound_perm) {
          if (!empty($outbound_perm)) {
            $outbound_rule_param = [
              'Direction' => 'egress',
              'FromPort' => $outbound_perm['from_port'],
              'ToPort' => $outbound_perm['to_port'],
              'IpProtocol' => $outbound_perm['ip_protocol'],
              'GroupId' => $entity->getGroupId(),
              'Description' => $outbound_perm['description'],
            ];

            switch ($outbound_perm['source']) {
              case 'ip4':
                $outbound_rule_param += [
                  'Source' => 'IPv4',
                  'CidrIp' => $outbound_perm['cidr_ip'],
                ];
                $ec2_service->createSecurityGroupRule($outbound_rule_param);
                break;

              case 'group':
                $outbound_rule_param += [
                  'Source' => 'IPv4',
                  'RemoteGroupId' => $outbound_perm['group_id'],
                  'RemoteUserId' => $outbound_perm['user_id'],
                ];
                $ec2_service->createSecurityGroupRule($outbound_rule_param);
                break;

              case 'ip6':
                $outbound_rule_param += [
                  'Source' => 'IPv6',
                  'CidrIpv6' => $outbound_perm['cidr_ip_v6'],
                ];
                $ec2_service->createSecurityGroupRule($outbound_rule_param);
                break;
            }
          }
        }

        // Have the system refresh the security group.
        $ec2_service->updateSecurityGroups([
          'GroupIds' => [$entity->getGroupId()],
        ], FALSE);

        if (count($this->messenger->messagesByType('error')) === 0) {
          $this->processOperationStatus($entity, $ec2_service->isWorkerResource() ? 'updated remotely' : 'updated');
          $form_state->setRedirect("entity.{$entity->getEntityTypeId()}.canonical", [
            'cloud_context' => $entity->getCloudContext(),
            $entity->getEntityTypeId() => $entity->id(),
          ]);
        }
        return TRUE;
      }
      else {
        $this->processOperationErrorStatus($this->entity, 'updated');
        return FALSE;
      }
    }
    catch (\Exception $e) {
      $this->handleException($e);
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function createOpenStackSnapshot(SnapshotInterface $entity, array &$form, FormStateInterface $form_state): bool {
    $this->entity = $entity;
    $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
    $this->awsCloudOperationsService->setEc2Service($this->ec2Service);
    return $this->awsCloudOperationsService->createSnapshot($this->entity, $form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function deleteOpenStackSnapshot(SnapshotInterface $entity, array &$form, FormStateInterface $form_state): bool {
    $this->entity = $entity;
    $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
    $this->awsCloudOperationsService->setEc2Service($this->ec2Service);
    return $this->awsCloudOperationsService->deleteSnapshot($this->entity, $form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function editOpenStackSnapshot(SnapshotInterface $entity, array &$form, FormStateInterface $form_state): bool {
    $this->entity = $entity;

    // Switch OpenStack EC2 or REST service based on $entity->getCloudContext().
    $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
    $this->awsCloudOperationsService->setEc2Service($this->ec2Service);
    $this->awsCloudOperationsService->editSnapshot($this->entity, $form, $form_state);

    // Update Snapshot of OpenStack REST API.
    if ($this->ec2Service instanceof OpenStackRestService) {
      $this->ec2Service->updateSnapshot([
        'SnapshotId'  => $entity->getSnapshotId(),
        'Name'        => $entity->getName(),
        'Description' => $entity->getDescription(),
      ]);
    }
    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function createOpenStackVolume(VolumeInterface $entity, array &$form, FormStateInterface $form_state): bool {
    /** @var \Drupal\aws_cloud\Entity\Ec2\VolumeInterface $entity */
    $this->entity = $entity;

    $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
    $this->awsCloudOperationsService->setEc2Service($this->ec2Service);
    return $this->awsCloudOperationsService->createVolume($entity, $form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function deleteOpenStackVolume(VolumeInterface $entity, array &$form, FormStateInterface $form_state): bool {
    /** @var \Drupal\aws_cloud\Entity\Ec2\VolumeInterface $entity */
    $this->entity = $entity;

    $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
    $this->awsCloudOperationsService->setEc2Service($this->ec2Service);
    return $this->awsCloudOperationsService->deleteVolume($entity, $form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function editOpenStackVolume(VolumeInterface $entity, array &$form, FormStateInterface $form_state): bool {
    /** @var \Drupal\aws_cloud\Entity\Ec2\VolumeInterface $entity */
    $this->entity = $entity;

    $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
    $this->awsCloudOperationsService->setEc2Service($this->ec2Service);
    $this->awsCloudOperationsService->editVolume($entity, $form, $form_state);

    // Update Volume of OpenStack REST API.
    if ($this->ec2Service instanceof OpenStackRestService) {
      $this->ec2Service->updateVolume([
        'VolumeId' => $entity->getVolumeId(),
        'Name' => $entity->getName(),
      ]);
    }
    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function attachOpenStackVolume(VolumeInterface $entity, array &$form, FormStateInterface $form_state): bool {
    /** @var \Drupal\aws_cloud\Entity\Ec2\VolumeInterface $entity */
    $this->entity = $entity;

    $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
    $this->awsCloudOperationsService->setEc2Service($this->ec2Service);
    return $this->awsCloudOperationsService->attachVolume($entity, $form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function detachOpenStackVolume(VolumeInterface $entity, array &$form, FormStateInterface $form_state): bool {
    /** @var \Drupal\aws_cloud\Entity\Ec2\VolumeInterface $entity */
    $this->entity = $entity;

    $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
    $this->awsCloudOperationsService->setEc2Service($this->ec2Service);
    $this->awsCloudOperationsService->detachVolume($entity, $form, $form_state);

    // Detach Volume of OpenStack REST API.
    if ($this->ec2Service instanceof OpenStackRestService) {
      $this->ec2Service->detachVolume([
        'VolumeId' => $entity->getVolumeId(),
        'InstanceId' => $entity->getAttachmentInformation(),
      ]);
    }
    return TRUE;
  }

  /**
   * Helper method to load instance by ID.
   *
   * @param string $instance_id
   *   Instance ID to load.
   * @param string $module_name
   *   Module name.
   *
   * @return \Drupal\aws_cloud\Entity\Ec2\InstanceInterface
   *   The Instance entity.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  private function getInstanceById($instance_id, $module_name): InstanceInterface {
    $instances = $this->entityTypeManager
      ->getStorage("{$module_name}_instance")
      ->loadByProperties([
        'instance_id' => $instance_id,
      ]);
    return array_shift($instances);
  }

  /**
   * Helper method to load network interface by ID.
   *
   * @param string $network_interface_id
   *   Network interface ID to load.
   * @param string $module_name
   *   Module name.
   *
   * @return \Drupal\aws_cloud\Entity\Ec2\NetworkInterfaceInterface
   *   The NetworkInterface entity.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  private function getNetworkInterfaceById($network_interface_id, $module_name): NetworkInterfaceInterface {
    $network_interface = $this->entityTypeManager
      ->getStorage("{$module_name}_network_interface")
      ->loadByProperties([
        'network_interface_id' => $network_interface_id,
      ]);
    return array_shift($network_interface);
  }

  /**
   * {@inheritdoc}
   */
  public function associateOpenStackFloatingIp(ElasticIpInterface $entity, array &$form, FormStateInterface $form_state): bool {
    try {
      $this->entity = $entity;

      // Switch OpenStack EC2 or REST service
      // based on $entity->getCloudContext().
      $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
      $this->awsCloudOperationsService->setEc2Service($this->ec2Service);

      if ($this->ec2Service instanceof OpenStackEc2Service) {
        return $this->awsCloudOperationsService->associateElasticIp($entity, $form, $form_state);
      }

      $form_state->setRedirect("entity.{$entity->getEntityTypeId()}.canonical", [
        'cloud_context' => $entity->getCloudContext(),
        "{$entity->getEntityTypeId()}" => $entity->id(),
      ]);

      // Determine if an IP address is attaching to
      // instance or network_interface.
      return $form_state->getValue('resource_type') === 'instance'
        ? $this->associateOpenStackFloatingIpToInstance($entity, $form, $form_state)
        : $this->associateOpenStackFloatingIpToNetworkInterface($entity, $form, $form_state);

    }
    catch (\Exception $e) {
      $this->handleException($e);
      $this->processOperationErrorStatus($entity, 'associated');
      return FALSE;
    }
  }

  /**
   * Associate an OpenStack Floating IP to instance.
   *
   * @param \Drupal\aws_cloud\Entity\Ec2\ElasticIpInterface $entity
   *   The OpenStack Floating IP entity.
   * @param array $form
   *   Array of form object.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   *
   * @return bool
   *   TRUE when the process succeeds.
   */
  private function associateOpenStackFloatingIpToInstance(ElasticIpInterface $entity, array &$form, FormStateInterface $form_state): bool {
    // Get module name.
    $module_name = $this->getModuleName($entity);

    $instance_entity_id = $form_state->getValue('instance_id');
    $private_ip = $form_state->getValue('instance_private_ip');

    if ($instance_entity_id === -1) {
      $this->messenger->addError($this->t('Unable to load instance ID. No association performed.'));
      return FALSE;
    }

    $instances = $this->entityTypeManager
      ->getStorage("{$module_name}_instance")
      ->loadByProperties(['id' => $instance_entity_id]);

    if (empty($instances)) {
      $this->messenger->addError(
        $this->t('Unable to load the instance %instance_entity_id.', [
          '%instance_entity_id' => $instance_entity_id,
        ]
      ));

      return FALSE;
    }

    /** @var \Drupal\aws_cloud\Entity\Ec2\InstanceInterface $instance */
    $instance = array_shift($instances);
    $instance_id = $instance->getInstanceId();

    $instance_link = $this->entityLinkRenderer->renderViewElement(
      $instance_id,
      "{$module_name}_instance",
      'instance_id',
      [],
      $instance->getName() !== $instance->getInstanceId()
        ? $this->t('@instance_name (@instance_id)', [
          '@instance_name' => $instance->getName(),
          '@instance_id' => $instance->getInstanceId(),
        ])
        : $instance_id
    );

    $entity->setInstanceId($instance->getInstanceId());
    $entity->save();

    // Associate instance.
    /** @var \Drupal\openstack\Entity\OpenStackFloatingIp $entity */
    $result = $this->ec2Service->associateAddress([
      'FloatingIpId' => $entity->getFloatingIpId(),
      'PortId' => $private_ip,
    ]);

    if (empty($result)) {
      $this->messenger->addError($this->t('Unable to associate @label.', [
        '@label' => $entity->getEntityType()->getSingularLabel(),
      ]));
      return FALSE;
    }

    if (!empty($result['SendToWorker'])) {
      $message = $this->t("@label @ip_address assoicated remotely with @private_ip for instance: @instance_id", [
        '@ip_address' => $entity->getPublicIp(),
        '@private_ip' => $private_ip,
        '@instance_id' => Markup::create($instance_link['#markup']),
        '@label' => $entity->getEntityType()->getSingularLabel(),
      ]);

      $this->messenger->addStatus($message);
      return TRUE;
    }

    $message = $this->t("@label @ip_address assoicated with @private_ip for instance: @instance_id", [
      '@ip_address' => $entity->getPublicIp(),
      '@private_ip' => $private_ip,
      '@instance_id' => Markup::create($instance_link['#markup']),
      '@label' => $entity->getEntityType()->getSingularLabel(),
    ]);

    $this->updateFloatingIpEntity($message);
    $this->clearCacheValues($entity->getCacheTags());

    // Refresh the instances.
    $this->ec2Service->updateInstances();
    return TRUE;
  }

  /**
   * Associate an OpenStack Floating IP to network interface.
   *
   * @param \Drupal\aws_cloud\Entity\Ec2\ElasticIpInterface $entity
   *   The OpenStack Floating IP entity.
   * @param array $form
   *   Array of form object.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   *
   * @return bool
   *   TRUE when the process succeeds.
   */
  private function associateOpenStackFloatingIpToNetworkInterface(ElasticIpInterface $entity, array &$form, FormStateInterface $form_state): bool {
    // Get module name.
    $module_name = $this->getModuleName($entity);

    $network_interface_id = $form_state->getValue('network_interface_id');
    $network_private_ip = $form_state->getValue('network_private_ip');

    if ($network_interface_id === -1) {
      $this->messenger->addError($this->t('Unable to load instance ID. No association performed.'));
      return FALSE;
    }

    $network_interfaces = $this->entityTypeManager
      ->getStorage("{$module_name}_network_interface")
      ->loadByProperties(['id' => $network_interface_id]);

    /** @var \Drupal\aws_cloud\Entity\Ec2\NetworkInterfaceInterface $network_interface */
    $network_interface = count($network_interfaces) === 1
      ? array_shift($network_interfaces)
      : $network_interfaces;

    // Associate network interface.
    /** @var \Drupal\openstack\Entity\OpenStackFloatingIp $entity */
    $result = $this->ec2Service->associateAddress([
      'FloatingIpId' => $entity->getFloatingIpId(),
      'PortId' => $network_private_ip,
    ]);

    if (empty($result)) {
      $this->messenger->addError($this->t("Unable to associate @label.", [
        '@label' => $entity->getEntityType()->getSingularLabel(),
      ]));
    }

    if (!empty($result['SendToWorker'])) {
      $message = $this->t('@label @ip_address associated with @private_ip for network interface: @network_interface_id', [
        '@ip_address' => $entity->getPublicIp(),
        '@network_interface_id' => $network_interface->getNetworkInterfaceId(),
        '@private_ip' => $network_private_ip,
        '@label' => $entity->getEntityType()->getSingularLabel(),
      ]);

      $this->messenger->addStatus($message);
      return TRUE;
    }

    $message = $this->t('@label @ip_address associated with @private_ip for network interface: @network_interface_id', [
      '@ip_address' => $entity->getPublicIp(),
      '@network_interface_id' => $network_interface->getNetworkInterfaceId(),
      '@private_ip' => $network_private_ip,
      '@label' => $entity->getEntityType()->getSingularLabel(),
    ]);

    $this->updateFloatingIpEntity($message);
    $this->clearCacheValues($entity->getCacheTags());

    // Refresh the instances.
    $this->ec2Service->updateInstances();
    return TRUE;
  }

  /**
   * Helper function to update the current openstack_elastic_ip entity.
   *
   * @param string $message
   *   Message to display to use.
   */
  private function updateFloatingIpEntity($message): void {
    /** @var \Drupal\openstack\Service\OpenStackServiceInterface $ec2_service */
    $ec2_service = $this->ec2Service;

    // Update the following entities from OpenStack.
    $ec2_service->updateFloatingIps();
    $ec2_service->updateNetworkInterfaces();
    $ec2_service->updateInstances();
    $this->messenger->addStatus($message);
  }

  /**
   * {@inheritdoc}
   */
  public function getPrivateIps(ElasticIpInterface $entity, $instance_id): array {

    // Get module name.
    $module_name = $this->getModuleName($entity);

    $instances = $this->entityTypeManager
      ->getStorage("{$module_name}_instance")
      ->loadByProperties(['id' => $instance_id]);

    /** @var \Drupal\aws_cloud\Entity\Ec2\InstanceInterface $instance */
    $instance = count($instances) === 1
      ? array_shift($instances)
      : $instances;

    $ips = !empty($instance) && !empty($instance->getPrivateSecondaryIp())
      ? explode(', ', $instance->getPrivateSecondaryIp())
      : [];
    $private_ips = [];
    foreach ($ips ?: [] as $ip) {
      $private_ips[$ip] = $ip;
    }

    return $private_ips;
  }

  /**
   * {@inheritdoc}
   */
  public function importOpenStackKeyPair(KeyPairInterface $entity, array $form, FormStateInterface $form_state): bool {
    try {
      // Switch OpenStack EC2 or REST service
      // based on $entity->getCloudContext().
      $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
      $this->awsCloudOperationsService->setEc2Service($this->ec2Service);
      return $this->awsCloudOperationsService->importKeyPair($this->entity, $form, $form_state);
    }
    catch (\Exception $e) {
      $this->handleException($e);
      return FALSE;
    }
  }

}
