<?php

namespace Drupal\openstack\Service\Rest;

use Drupal\cloud\Entity\CloudConfigInterface;
use Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface;
use Drupal\cloud\Service\CloudServiceBase;
use Drupal\cloud\Service\CloudServiceInterface;
use Drupal\cloud\Traits\CloudContentEntityTrait;
use Drupal\Component\Render\FormattableMarkup;
use Drupal\Core\Batch\BatchBuilder;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\core\Field\FieldTypePluginManagerInterface;
use Drupal\Core\Http\ClientFactory;
use Drupal\Core\Lock\LockBackendInterface;
use Drupal\Core\Queue\QueueFactory;
use Drupal\Core\Template\TwigEnvironment;
use Drupal\cloud\Service\CloudResourceTagInterface;
use Drupal\openstack\Entity\OpenStackKeyPair;
use Drupal\openstack\Service\OpenStackBatchOperations;
use Drupal\openstack\Service\OpenStackServiceInterface;
use GuzzleHttp\Exception\GuzzleException;

/**
 * Interacts with OpenStack using Native API.
 */
class OpenStackService extends CloudServiceBase implements OpenStackServiceInterface, CloudResourceTagInterface {

  use CloudContentEntityTrait;

  /**
   * Drupal\Core\Entity\EntityTypeManagerInterface definition.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The cloud service provider plugin manager (CloudConfigPluginManager).
   *
   * @var \Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface
   */
  protected $cloudConfigPluginManager;

  /**
   * Guzzle client.
   *
   * @var \GuzzleHttp\Client
   */
  protected $httpClient;

  /**
   * The lock interface.
   *
   * @var \Drupal\Core\Lock\LockBackendInterface
   */
  protected $lock;

  /**
   * The queue factory.
   *
   * @var \Drupal\Core\Queue\QueueFactory
   */
  protected $queueFactory;

  /**
   * Credentials array.
   *
   * @var array
   */
  protected $credentials;

  /**
   * Entity field manager interface.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * Field type plugin manager.
   *
   * @var \Drupal\core\Field\FieldTypePluginManagerInterface
   */
  protected $fieldTypePluginManager;

  /**
   * The config factory.
   *
   * Subclasses should use the self::config() method, which may be overridden to
   * address specific needs when loading config, rather than this property
   * directly. See \Drupal\Core\Form\ConfigFormBase::config() for an example of
   * this.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The Cloud service.
   *
   * @var \Drupal\cloud\Service\CloudServiceInterface
   */
  protected $cloudService;

  /**
   * The Twig service.
   *
   * @var \Drupal\Core\Template\TwigEnvironment
   */
  protected $twig;

  /**
   * Constructs a new OpenStackService object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   An entity type manager instance.
   * @param \Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface $cloud_config_plugin_manager
   *   The cloud service provider plugin manager (CloudConfigPluginManager).
   * @param \Drupal\Core\Http\ClientFactory $client_factory
   *   The http client.
   * @param \Drupal\Core\Lock\LockBackendInterface $lock
   *   The lock interface.
   * @param \Drupal\Core\Queue\QueueFactory $queue_factory
   *   The queue factory.
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $entity_field_manager
   *   The entity field manager.
   * @param \Drupal\core\Field\FieldTypePluginManagerInterface $field_type_plugin_manager
   *   The field type plugin manager.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   A configuration factory.
   * @param \Drupal\cloud\Service\CloudServiceInterface $cloud_service
   *   The Cloud service.
   * @param \Drupal\Core\Template\TwigEnvironment $twig
   *   The twig service.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager,
                              CloudConfigPluginManagerInterface $cloud_config_plugin_manager,
                              ClientFactory $client_factory,
                              LockBackendInterface $lock,
                              QueueFactory $queue_factory,
                              EntityFieldManagerInterface $entity_field_manager,
                              FieldTypePluginManagerInterface $field_type_plugin_manager,
                              ConfigFactoryInterface $config_factory,
                              CloudServiceInterface $cloud_service,
                              TwigEnvironment $twig) {

    // The parent constructor takes care of $this->messenger object.
    parent::__construct();

    // Setup the entity type manager for querying entities.
    $this->entityTypeManager = $entity_type_manager;

    $this->cloudConfigPluginManager = $cloud_config_plugin_manager;
    $this->httpClient = $client_factory->fromOptions();
    $this->lock = $lock;
    $this->queueFactory = $queue_factory;
    $this->entityFieldManager = $entity_field_manager;
    $this->fieldTypePluginManager = $field_type_plugin_manager;
    $this->configFactory = $config_factory;
    $this->cloudService = $cloud_service;
    $this->twig = $twig;
  }

  /**
   * Helper method to call API.
   *
   * @param array $credentials
   *   The credentials array.
   *
   * @return array
   *   Array interpretation of the response body.
   */
  public function getClient(array $credentials = []) {
    $http_method = $credentials['http_method'] ?? '';
    $endpoint = $credentials['endpoint'] ?? '';
    $data = $credentials['params'] ?? '';

    if (strtolower($http_method) !== 'get' && !empty($data)) {
      $params['body'] = json_encode($data);
    }

    $this->loadCredentials();
    $auth_token = $this->generateAuthToken();
    $params['headers']['X-Auth-Token'] = $auth_token;

    $params['headers']['Content-Type'] = !empty($credentials['Content-Type'])
      ? $credentials['Content-Type'] : 'application/json';

    if (!empty($credentials['X-OpenStack-Nova-API-Version'])) {
      $params['headers']['X-OpenStack-Nova-API-Version'] = $credentials['X-OpenStack-Nova-API-Version'];
    }

    $params['verify'] = !empty($this->credentials['self_signed_cert_path'])
      ? $this->credentials['self_signed_cert_path']
      : TRUE;

    return $this->request($http_method, $endpoint, $params);
  }

  /**
   * {@inheritdoc}
   */
  public function setCloudContext($cloud_context) {
    $this->cloudContext = $cloud_context;
    $this->cloudConfigPluginManager->setCloudContext($cloud_context);
  }

  /**
   * Load credentials.
   *
   * @param array $credentials
   *   The credentials array.
   *
   * @return array
   *   The credentials array.
   */
  private function loadCredentials(array $credentials = []) {
    if (!empty($credentials)) {
      $this->credentials = $credentials;
    }
    elseif (empty($this->credentials)) {
      $this->credentials = $this->cloudConfigPluginManager->loadCredentials();
    }

    return $this->credentials;
  }

  /**
   * Setup any default parameters for the Guzzle request.
   *
   * @param array $credentials
   *   The credentials array.
   *
   * @return array
   *   Array of parameters.
   */
  public function generateAuthToken(array $credentials = []) {
    $this->loadCredentials($credentials);

    $body = [
      'auth' => [
        'identity' => [
          'methods' => ['password'],
          'password' => [
            'user' => [
              'name' => $this->credentials['username'],
              'domain' => [
                'id' => $this->credentials['domain_id'],
              ],
              'password' => $this->credentials['password'],
            ],
          ],
        ],
        'scope' => [
          'project' => [
            'id' => $this->credentials['project_id'],
          ],
        ],
      ],
    ];

    $endpoint = $this->getEndpoint(OpenStackServiceInterface::IDENTITY_ENDPOINT, '/auth/tokens');
    $response = $this->request('POST', $endpoint, [
      'json' => $body,
      'verify' => !empty($this->credentials['self_signed_cert_path'])
        ? $this->credentials['self_signed_cert_path']
        : TRUE,
    ]);

    return $response['auth_token'] ?? $response;
  }

  /**
   * Get Instance List.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The Instance list API response.
   */
  public function describeInstances(array $params = []) {
    $this->loadCredentials();

    $credentials = [
      'http_method' => 'get',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::COMPUTE_ENDPOINT, '/servers/detail'),
      'params' => $params,
    ];

    $response = $this->getClient($credentials);

    $results = $volume_id_arr = $security_groups = [];
    $instances = $response['body']['servers'] ?? [];

    foreach ($instances ?: [] as $key => $instance) {
      $instance_tags = $this->getInstanceTags(['InstanceId' => $instance['id']]);

      foreach ($instance['security_groups'] ?? [] as $key_group => $group) {
        $security_groups[$key_group] = [
          'GroupName' => $group['name'] ?? NULL,
        ];
      }

      // Get Volume Response.
      $volume_ids = $instance['os-extended-volumes:volumes_attached'];
      foreach ($volume_ids ?: [] as $volume_id_val) {
        $volume_id = $volume_id_val['id'];
        $volume_data = $this->describeVolumeById([], $volume_id);
        $image_id = $volume_data['volume_image_metadata']['image_id'] ?? NULL;
        $instance['image'] = empty($instance['image'])
          ? ['id' => $image_id] : [];
        $volume_id_arr = [
          'Ebs' => ['VolumeId' => $volume_id ?? ''],
        ];
      }

      // Get flavor Response.
      $flavor_data = !empty($instance['flavor']['id'])
        ? $this->describeFlavorById([], $instance['flavor']['id'])
        : [];

      // Get port response.
      $port_result = !empty($instance['OS-EXT-STS:vm_state'])
      && $instance['OS-EXT-STS:vm_state'] !== 'building'
        ? $this->instancePortInterface([
          'InstanceId' => $instance['id'],
        ])
        : [];

      $port_ids = [];
      if (!empty($port_result)) {
        foreach ($port_result ?: [] as $port) {
          $port_ids[] = $port['port_id'];
        }
      }

      foreach ($instance['addresses'] ?: [] as $add_type => $address_val) {
        foreach ($address_val ?: [] as $add_value) {
          if ($add_type === 'private') {
            if ($add_value['OS-EXT-IPS:type'] === 'floating') {
              $public_ip = $add_value['addr'] ?? '';
            }
            elseif ($add_value['OS-EXT-IPS:type'] === 'fixed'
              && $add_value['version'] === OpenStackServiceInterface::IPV4) {
              $private_ip = $add_value['addr'] ?? '';
            }
          }
          elseif ($add_type === 'public'
            && $add_value['OS-EXT-IPS:type'] === 'fixed'
            && $add_value['version'] === OpenStackServiceInterface::IPV4) {
            $public_ip = $add_value['addr'] ?? '';
          }
        }
      }

      $state = strtolower($instance['status']);
      if ($state === 'active') {
        $state = 'running';
      }
      elseif ($state === 'shutoff') {
        $state = 'stopped';
      }

      $results['Reservations'][$key] = [
        'OwnerId'    => $instance['tenant_id'],
        'Instances'  => [
          [
            'AmiLaunchIndex'      => NULL,
            'ImageId'             => $instance['image']['id'] ?? NULL,
            'InstanceId'          => $instance['id'] ?? NULL,
            'InstanceName'        => $instance['name'] ?? NULL,
            'InstanceType'        => $flavor_data['name'] ?? NULL,
            'KeyName'             => $instance['key_name'] ?? NULL,
            'LaunchTime'          => !empty($instance['OS-SRV-USG:launched_at'])
              ? str_replace('T', ' ', $instance['OS-SRV-USG:launched_at'])
              : str_replace('Z', '', str_replace('T', ' ', $instance['created'])),
            'Placement'           => ['AvailabilityZone' => $instance['OS-EXT-AZ:availability_zone'] ?? ''] ?? NULL,
            'PrivateDnsName'      => $instance['name'] ?? NULL,
            'PrivateIpAddress'    => $private_ip ?? '',
            'PublicIpAddress'     => $public_ip ?? '',
            'PublicDnsName'       => $instance['dns_name'] ?? NULL,
            'State'               => ['Name' => $state],
            'PowerState'          => $instance['OS-EXT-STS:power_state'] ?? NULL,
            'RootDeviceName'      => NULL,
            'RootDeviceType'      => NULL,
            'SecurityGroups'      => $security_groups ?? [],
            'BlockDeviceMappings' => [$volume_id_arr ?? []],
            'PortId'              => $port_ids ?? [],
            'Tags'                => $instance_tags,
          ],
        ],
      ];
    }

    return $results;
  }

  /**
   * Update Instance.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The Instance update API response.
   */
  public function modifyLaunchTemplate(array $params = []): array {
    if ($this->isWorkerResource()) {
      $this->saveOperation(__FUNCTION__, [$params], 'updateInstances');

      return ['SendToWorker' => TRUE];
    }

    $this->loadCredentials();

    $body['server'] = [
      "name" => $params['Name'],
    ];

    $credentials = [
      'http_method' => 'put',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::COMPUTE_ENDPOINT, "/servers/{$params['InstanceId']}"),
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    $instance = $response['body']['server'] ?? [];

    $volume_id_arr = $security_groups = [];

    foreach ($instance['security_groups'] ?? [] as $key_group => $group) {
      $security_groups[$key_group] = [
        'GroupName' => $group['name'] ?? NULL,
      ];
    }

    // Get Volume Response.
    $volume_ids = $instance['os-extended-volumes:volumes_attached'] ?? [];
    foreach ($volume_ids ?: [] as $volume_id_val) {
      $volume_id = $volume_id_val['id'];
      $volume_data = $this->describeVolumeById([], $volume_id);
      $image_id = $volume_data['volume_image_metadata']['image_id'] ?? NULL;
      $instance['image'] = empty($instance['image'])
        ? ['id' => $image_id] : [];
      $volume_id_arr = [
        'Ebs' => ['VolumeId' => $volume_id ?? ''],
      ];
    }

    // Get flavor Response.
    $flavor_data = !empty($instance['flavor']['id'])
      ? $this->describeFlavorById([], $instance['flavor']['id'])
      : [];

    // Get port response.
    $port_result = !empty($instance['OS-EXT-STS:vm_state'])
    && $instance['OS-EXT-STS:vm_state'] !== 'building'
      ? $this->instancePortInterface([
        'InstanceId' => $instance['id'],
      ])
      : [];

    $port_ids = [];
    if (!empty($port_result)) {
      foreach ($port_result ?: [] as $port) {
        $port_ids[] = $port['port_id'];
      }
    }

    foreach ($instance['addresses'] ?: [] as $address_val) {
      foreach ($address_val ?: [] as $add_value) {
        if (!empty($add_value['OS-EXT-IPS:type'])) {
          if ($add_value['OS-EXT-IPS:type'] === 'floating') {
            $public_ip = $add_value['addr'] ?? '';
          }
          if ($add_value['OS-EXT-IPS:type'] === 'fixed'
            && $add_value['version'] === OpenStackServiceInterface::IPV4) {
            $private_ip = $add_value['addr'] ?? '';
          }
        }
      }
    }

    $instance_tags = $this->getInstanceTags(
      ['InstanceId' => $params['InstanceId']]
    );

    return [
      'Reservations' => [
        'OwnerId'    => $instance['tenant_id'],
        'Instances'  => [
          [
            'AmiLaunchIndex'      => NULL,
            'ImageId'             => $instance['image']['id'] ?? NULL,
            'InstanceId'          => $instance['id'] ?? NULL,
            'InstanceName'        => $instance['name'] ?? NULL,
            'InstanceType'        => $flavor_data['name'] ?? NULL,
            'KeyName'             => $instance['key_name'] ?? NULL,
            'LaunchTime'          => !empty($instance['OS-SRV-USG:launched_at'])
              ? str_replace('T', ' ', $instance['OS-SRV-USG:launched_at']) : NULL,
            'Placement'           => ['AvailabilityZone' => $instance['OS-EXT-AZ:availability_zone'] ?? ''] ?? NULL,
            'PrivateDnsName'      => $instance['name'] ?? NULL,
            'PrivateIpAddress'    => $private_ip ?? '',
            'PublicIpAddress'     => $public_ip ?? '',
            'PublicDnsName'       => $instance['dns_name'] ?? NULL,
            'State'               => ['Name' => $instance['status'] === 1 ? 'running' : 'stopped'],
            'RootDeviceName'      => NULL,
            'RootDeviceType'      => NULL,
            'SecurityGroups'      => $security_groups ?? [],
            'BlockDeviceMappings' => [$volume_id_arr ?? []],
            'PortId'              => $port_ids ?? [],
            'Tags'                => $instance_tags,
          ],
        ],
      ],
    ];
  }

  /**
   * Start Instance.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array|null
   *   The Instance Start API response.
   */
  public function startInstances(array $params = []): ?array {
    if ($this->isWorkerResource()) {
      $this->saveOperation(__FUNCTION__, [$params], 'updateInstances');

      return ['SendToWorker' => TRUE];
    }

    $this->loadCredentials();

    $body = [
      'os-start' => NULL,
    ];

    $credentials = [
      'http_method' => 'post',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::COMPUTE_ENDPOINT, "/servers/{$params['InstanceId']}/action"),
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    return $response['status_code'] === 202 ? ['Success' => TRUE] : NULL;
  }

  /**
   * Stop Instance.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array|null
   *   The Instance Stop API response.
   */
  public function stopInstances(array $params = []): ?array {
    if ($this->isWorkerResource()) {
      $this->saveOperation(__FUNCTION__, [$params], 'updateInstances');

      return ['SendToWorker' => TRUE];
    }

    $this->loadCredentials();

    $body = [
      'os-stop' => NULL,
    ];

    $credentials = [
      'http_method' => 'post',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::COMPUTE_ENDPOINT, "/servers/{$params['InstanceId']}/action"),
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    return $response['status_code'] === 202 ? ['Success' => TRUE] : NULL;
  }

  /**
   * Reboot Instance.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array|null
   *   The Instance Reboot API response.
   */
  public function rebootInstances(array $params = []): ?array {
    if ($this->isWorkerResource()) {
      $this->saveOperation(__FUNCTION__, [$params], 'updateInstances');

      return ['SendToWorker' => TRUE];
    }

    $this->loadCredentials();

    $body['reboot'] = [
      'type' => 'HARD',
    ];

    $credentials = [
      'http_method' => 'post',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::COMPUTE_ENDPOINT, "/servers/{$params['InstanceId']}/action"),
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    return $response['status_code'] === 202 ? ['Success' => TRUE] : NULL;
  }

  /**
   * Terminate|Delete Instance.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array|null
   *   The Instance Terminate|Delete API response.
   */
  public function terminateInstances(array $params = []): ?array {
    if ($this->isWorkerResource()) {
      $this->saveOperation(__FUNCTION__, [$params], 'updateInstances');

      return ['SendToWorker' => TRUE];
    }

    $this->loadCredentials();

    $credentials = [
      'http_method' => 'delete',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::COMPUTE_ENDPOINT, "/servers/{$params['InstanceId']}"),
    ];

    $response = $this->getClient($credentials);

    return $response['status_code'] === 204 ? ['Success' => TRUE] : NULL;
  }

  /**
   * Run instance.
   *
   * @param array $params
   *   Parameters to pass to the API.
   * @param array $tags
   *   The tags array.
   *
   * @return array|null
   *   The create instance list API response.
   */
  public function runInstances(array $params = [], array $tags = []): ?array {
    if ($this->isWorkerResource()) {
      $this->saveOperation(__FUNCTION__, [$params], 'updateInstances');

      return ['SendToWorker' => TRUE];
    }

    $this->loadCredentials();

    $metadata = [];
    foreach ($tags as $tag) {
      $metadata[$tag['Key']] = (string) $tag['Value'];
    }

    $body = [
      'server' => [
        'name' => $params['Name'],
        'flavorRef' => $params['FlavorId'],
        'imageRef' => $params['ImageId'],
        'key_name' => $params['KeyName'],
        'availability_zone' => $params['Placement']['AvailabilityZone'],
        'security_groups' => $params['SecurityGroupIds'],
        'networks' => $params['NetworkInterfaces'],
        'metadata' => $metadata,
      ],
    ];

    $credentials = [
      'http_method' => 'post',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::COMPUTE_ENDPOINT, '/servers'),
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    // Update instances.
    $this->updateInstances();

    return $response['status_code'] === 202 ? ['Success' => TRUE] : NULL;
  }

  /**
   * Delete tags.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array|null
   *   The delete instance list API response.
   */
  public function deleteTags(array $params = []): ?array {
    if ($this->isWorkerResource()) {
      $this->saveOperation(__FUNCTION__, [$params]);

      return ['SendToWorker' => TRUE];
    }

    $this->loadCredentials();

    $endpoint = $this->getDeleteTagsEndpointUrl($params);

    $credentials = [
      'http_method' => 'delete',
      'endpoint' => $endpoint[$params['EntityType']],
    ];
    if ($params['EntityType'] === 'openstack_instance') {
      $credentials['X-OpenStack-Nova-API-Version'] = OpenStackServiceInterface::OPENSTACK_NOVA_API_VERSION;
    }

    $response = $this->getClient($credentials);

    return ($response['status_code'] === 202 || $response['status_code'] === 204) ? ['Success' => TRUE] : NULL;
  }

  /**
   * Get Image List.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The Image list API response.
   */
  public function describeImages(array $params = []) {
    $this->loadCredentials();

    $credentials = [
      'http_method' => 'get',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::IMAGE_ENDPOINT, '/images'),
      'params' => $params,
    ];

    $response = $this->getClient($credentials);
    $images = $response['body']['images'] ?? [];

    $results = [];

    foreach ($images ?: [] as $key => $value) {
      $results['Images'][$key]['Architecture'] = NULL;
      $results['Images'][$key]['CreationDate'] = $value['created_at'];
      $results['Images'][$key]['ImageId'] = $value['id'];
      $results['Images'][$key]['ImageLocation'] = NULL;
      $results['Images'][$key]['ImageType'] = $value['disk_format'];
      $results['Images'][$key]['Public'] = $value['visibility'] === 'public' ? 1 : 0;
      $results['Images'][$key]['OwnerId'] = $value['owner'];
      $results['Images'][$key]['State'] = $value['status'] === 'active' ? 'available' : 'pending';
      $results['Images'][$key]['Name'] = $value['name'];
      $results['Images'][$key]['RootDeviceType'] = $value['container_format'];
      $results['Images'][$key]['Tags'] = $value['tags'];
    }

    return $results;
  }

  /**
   * Create Image List.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The Image create API response.
   */
  public function createImage(array $params = []): array {
    if ($this->isWorkerResource()) {
      $this->saveOperation(__FUNCTION__, [$params], 'updateImages');

      return ['SendToWorker' => TRUE];
    }

    $this->loadCredentials();

    $body = [
      'container_format' => 'bare',
      'disk_format' => 'qcow2',
      'name' => $params['Name'],
      'visibility' => $params['Visibility'] ? 'public' : 'private',
    ];

    $credentials = [
      'http_method' => 'post',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::IMAGE_ENDPOINT, '/images'),
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    return !empty($response) && !empty($response['body']) ? [
      'ImageId'          => $response['body']['id'],
      'name'             => $response['body']['name'],
      'disk_format'      => $response['body']['disk_format'],
      'container_format' => $response['body']['container_format'],
      'visibility'       => $response['body']['visibility'],
      'created_at'       => $response['body']['created_at'],
      'updated_at'       => $response['body']['updated_at'],
      'Tags'             => $response['body']['tags'],
      'status'           => $response['body']['status'],
    ]
    : [];
  }

  /**
   * Update Image List.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The Image update API response.
   */
  public function updateImage(array $params = []): array {
    if ($this->isWorkerResource()) {
      $this->saveOperation(__FUNCTION__, [$params], 'updateImages');

      return ['SendToWorker' => TRUE];
    }

    $this->loadCredentials();

    $body = [
      [
        'op' => 'replace',
        'path' => '/name',
        'value' => $params['Name'],
      ],
      [
        'op' => 'replace',
        'path' => '/visibility',
        'value' => $params['Visibility'],
      ],
    ];

    $credentials = [
      'http_method' => 'patch',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::IMAGE_ENDPOINT, "/images/{$params['ImageId']}"),
      'params' => $body,
    ];

    $credentials['Content-Type'] = 'application/openstack-images-v2.1-json-patch';

    $response = $this->getClient($credentials);

    return !empty($response) && !empty($response['body']) ? [
      'ImageId'          => $response['body']['id'],
      'name'             => $response['body']['name'],
      'disk_format'      => $response['body']['disk_format'],
      'container_format' => $response['body']['container_format'],
      'visibility'       => $response['body']['visibility'],
      'created_at'       => $response['body']['created_at'],
      'updated_at'       => $response['body']['updated_at'],
      'Tags'             => $response['body']['tags'],
      'status'           => $response['body']['status'],
    ]
    : [];
  }

  /**
   * Register uid tag in the OpenStack resource's Tags.
   *
   * For resources that have Tags, register the uid with Tags.
   *
   * @param string $entity_type
   *   The entity type.
   * @param string $endpoint
   *   The endpoint.
   * @param array $tags
   *   The tags.
   *
   * @return bool
   *   TRUE when the process succeeds.
   */
  private function setUidToTags(string $entity_type, string $endpoint, array $tags): bool {
    $success = FALSE;
    foreach ($tags as $tag) {
      $key_value = "{$tag['Key']}.{$tag['Value']}";

      if (strlen($key_value) > OpenStackServiceInterface::MAX_TAG_LENGTH) {
        $this->messenger->addError($this->t('Failed to create tag @tag because it was longer than @max_length characters.', [
          '@tag' => $key_value,
          '@max_length' => OpenStackServiceInterface::MAX_TAG_LENGTH,
        ]));

        $success = FALSE;
        break;
      }

      $credentials = [
        'http_method' => 'put',
        'endpoint' => $endpoint . '/' . $key_value,
      ];

      if ($entity_type === 'openstack_instance') {
        $credentials['X-OpenStack-Nova-API-Version'] = OpenStackServiceInterface::OPENSTACK_NOVA_API_VERSION;
      }

      $response = $this->getClient($credentials);

      $success = $response['status_code'] === 204 || $response['status_code'] === 201 || $response['status_code'] === 200;
      if (!$success) {
        break;
      }
    }

    return $success;
  }

  /**
   * Register uid tag in the OpenStack resource's Metadata.
   *
   * Basically, the uid is registered in Tags.
   * But if the OpenStack resource's has no Tags, it is registered in Metadata.
   *
   * @param string $entity_type
   *   The entity type.
   * @param string $endpoint
   *   The endpoint.
   * @param array $tags
   *   The tags.
   *
   * @return bool
   *   TRUE when the process succeeds.
   */
  private function setUidToMetadata(string $entity_type, string $endpoint, array $tags): bool {
    $request_tags = [];
    foreach ($tags as $tag) {
      $request_tags['metadata'][$tag['Key']] = (string) $tag['Value'];
    }

    $credentials = [
      'http_method' => 'put',
      'endpoint' => $endpoint,
      'params' => $request_tags,
    ];

    if ($entity_type === 'openstack_instance') {
      $credentials['X-OpenStack-Nova-API-Version'] = OpenStackServiceInterface::OPENSTACK_NOVA_API_VERSION;
    }

    $response = $this->getClient($credentials);

    return $response['status_code'] === 204 || $response['status_code'] === 201 || $response['status_code'] === 200;
  }

  /**
   * Create Tags.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array|null
   *   The Tags create API response.
   */
  public function createTags(array $params = []): ?array {
    if ($this->isWorkerResource()) {
      $this->saveOperation(__FUNCTION__, [$params]);

      return ['SendToWorker' => TRUE];
    }

    $this->loadCredentials();

    if (empty($params['EntityType'])) {
      return NULL;
    }

    $endpoint = $this->getTagsEndpointUrl($params);

    // If the endpoint doesn't exist, skip.
    if (empty($endpoint[$params['EntityType']])) {
      return NULL;
    }

    $success = TRUE;
    if (empty($params['Tags'])) {
      return NULL;
    }
    $tags = empty($params['Tags']) ? [] : $params['Tags'];

    $success = NULL;
    if (empty($params['EntityType'])
      || !array_key_exists($params['EntityType'], OpenStackServiceInterface::OPENSTACK_RESOURCE_NAME_MAPPINGS_TO_SAVE_UID)) {
      return NULL;
    }

    switch (OpenStackServiceInterface::OPENSTACK_RESOURCE_NAME_MAPPINGS_TO_SAVE_UID[$params['EntityType']]) {
      case OpenStackServiceInterface::OPENSTACK_TAGS_TO_SAVE_UID:
        $success = $this->setUidToTags($params['EntityType'], $endpoint[$params['EntityType']], $tags);
        break;

      case OpenStackServiceInterface::OPENSTACK_METADATA_TO_SAVE_UID:
        $success = $this->setUidToMetadata($params['EntityType'], $endpoint[$params['EntityType']], $tags);
        break;

      default:
        return NULL;
    }

    return $success ? ['Success' => TRUE] : NULL;
  }

  /**
   * Get instance tags.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The instance tag list API response.
   */
  public function getInstanceTags(array $params = []): array {
    $this->loadCredentials();

    $credentials = [
      'http_method' => 'GET',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::COMPUTE_ENDPOINT, "/servers/{$params['InstanceId']}"),
      'X-OpenStack-Nova-API-Version' => OpenStackServiceInterface::OPENSTACK_NOVA_API_VERSION,
    ];

    $response = $this->getClient($credentials);

    $result = $response['body']['server']['metadata'] ?? [];

    $tags = [];
    if (empty($result)) {
      return $tags;
    }

    $uid_key_name = $this->getTagKeyCreatedByUid(
      'openstack',
      $this->cloudContext,
    );

    $tags = [
      [
        'Key' => $uid_key_name,
        'Value' => !empty($result[$uid_key_name]) ? intval($result[$uid_key_name]) : 0,
      ],
    ];

    return $tags;
  }

  /**
   * Create Tags.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The endpoint URL array.
   */
  public function getTagsEndpointUrl(array $params = []): array {
    $this->loadCredentials();
    return [
      'openstack_image'             => $this->getEndpoint(OpenStackServiceInterface::IMAGE_ENDPOINT, "/images/{$params['Resources'][0]}/tags"),
      'openstack_network_interface' => $this->getEndpoint(OpenStackServiceInterface::NETWORK_ENDPOINT, "/networks/{$params['Resources'][0]}/tags"),
      'openstack_security_group'    => $this->getEndpoint(OpenStackServiceInterface::NETWORK_ENDPOINT, "/security-groups/{$params['Resources'][0]}/tags"),
      'openstack_floating_ip'       => $this->getEndpoint(OpenStackServiceInterface::NETWORK_ENDPOINT, "/floatingips/{$params['Resources'][0]}/tags"),
      'openstack_instance'          => $this->getEndpoint(OpenStackServiceInterface::COMPUTE_ENDPOINT, "/servers/{$params['Resources'][0]}/metadata"),
      'openstack_volume'            => $this->getEndpoint(OpenStackServiceInterface::VOLUME_ENDPOINT, "/volumes/{$params['Resources'][0]}/metadata"),
      'openstack_snapshot'          => $this->getEndpoint(OpenStackServiceInterface::VOLUME_ENDPOINT, "/snapshots/{$params['Resources'][0]}/metadata"),
    ];
  }

  /**
   * Get delete tags endpoints.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The endpoint URL array.
   */
  public function getDeleteTagsEndpointUrl(array $params = []): array {
    $this->loadCredentials();
    $urls = [];

    $endpoints = [
      'openstack_image' => [
        'url' => OpenStackServiceInterface::IMAGE_ENDPOINT,
        'path' => '/images/{{ImageId}}/tags/{{Tag}}',
        'path_params' => [
          'ImageId',
          'Tag',
        ],
      ],
      'openstack_network_interface' => [
        'url' => OpenStackServiceInterface::NETWORK_ENDPOINT,
        'path' => '/networks/{{NetworkInterfaceId}}/tags/',
        'path_params' => [
          'NetworkInterfaceId',
        ],
      ],
      'openstack_security_group' => [
        'url' => OpenStackServiceInterface::NETWORK_ENDPOINT,
        'path' => '/security-groups/{{SecurityGroupId}}/tags/',
        'path_params' => [
          'SecurityGroupId',
        ],
      ],
      'openstack_floating_ip' => [
        'url' => OpenStackServiceInterface::NETWORK_ENDPOINT,
        'path' => '/floatingips/{{FloatingIpId}}/tags/',
        'path_params' => [
          'FloatingIpId',
        ],
      ],
      'openstack_instance' => [
        'url' => OpenStackServiceInterface::COMPUTE_ENDPOINT,
        'path' => '/servers/{{InstanceId}}/metadata/{{UidKeyName}}',
        'path_params' => [
          'InstanceId',
          'UidKeyName',
        ],
      ],
    ];

    foreach ($endpoints as $entity_type => $endpoint) {
      // If some path parameters aren't in $params, skip.
      if (!empty(array_diff($endpoint['path_params'], array_keys($params)))) {
        continue;
      }

      $urls[$entity_type] = $this->getEndpoint($endpoint['url'], $this->twig->renderInline($endpoint['path'], $params));
    }

    return $urls;
  }

  /**
   * Delete Image List.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array|null
   *   The Image delete API response.
   */
  public function deregisterImage(array $params = []): ?array {
    if ($this->isWorkerResource()) {
      $this->saveOperation(__FUNCTION__, [$params], 'updateImages');

      return ['SendToWorker' => TRUE];
    }

    $this->loadCredentials();

    $credentials = [
      'http_method' => 'DELETE',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::IMAGE_ENDPOINT, "/images/{$params['ImageId']}"),
    ];

    $response = $this->getClient($credentials);

    return $response['status_code'] === 204 ? ['Success' => TRUE] : NULL;
  }

  /**
   * Get key pairs List.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The key pairs list API response.
   */
  public function describeKeyPairs(array $params = []) {
    if ($this->isWorkerResource()) {
      return ['SendToWorker' => TRUE];
    }

    $this->loadCredentials();

    $credentials = [
      'http_method' => 'GET',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::COMPUTE_ENDPOINT, '/os-keypairs'),
      'params' => $params,
    ];
    $response = $this->getClient($credentials);

    $key_pairs = $response['body']['keypairs'] ?? [];

    $results['KeyPairs'] = [];
    foreach ($key_pairs ?: [] as $key => $key_pair) {
      $results['KeyPairs'][$key]['KeyName'] = $key_pair['keypair']['name'];
      $results['KeyPairs'][$key]['KeyFingerprint'] = $key_pair['keypair']['fingerprint'];
      $results['KeyPairs'][$key]['KeyMaterial'] = $key_pair['keypair']['public_key'];
    }

    return $results;
  }

  /**
   * Create key pairs List.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The key pairs create API response.
   */
  public function createKeyPair(array $params = []) {
    if ($this->isWorkerResource()) {
      $this->saveOperation(__FUNCTION__, [$params], 'createKeyPairEntity');

      return ['SendToWorker' => TRUE];
    }

    $this->loadCredentials();

    $body = [
      'keypair' => [
        'name' => $params['KeyName'],
      ],
    ];

    $credentials = [
      'http_method' => 'POST',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::COMPUTE_ENDPOINT, '/os-keypairs'),
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    return !empty($response)
      && !empty($response['body'])
      && !empty($response['body']['keypair'])
      ? [
        'KeyPairId' => NULL,
        'KeyName' => $response['body']['keypair']['name'],
        'KeyFingerprint' => $response['body']['keypair']['fingerprint'],
        'KeyMaterial' => $response['body']['keypair']['public_key'],
      ]
      : [];
  }

  /**
   * {@inheritdoc}
   */
  public function createKeyPairEntity(array $params): void {
    $result = $params['result'];
    $uid = $this->getUidTagValue($result, 'openstack_key_pair');
    $timestamp = time();

    $entity = OpenStackKeyPair::create([
      'cloud_context' => $this->cloudContext,
      'key_pair_name' => $result['KeyName'],
      'key_pair_id' => $result['KeyPairId'],
      'key_fingerprint' => $result['KeyFingerprint'],
      'key_material' => $result['KeyMaterial'],
      'created' => $timestamp,
      'changed' => $timestamp,
      'refreshed' => $timestamp,
      'uid' => $uid,
    ]);
    $entity->save();
  }

  /**
   * Import key pairs List.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The key pairs import API response.
   */
  public function importKeyPair(array $params = []) {
    if ($this->isWorkerResource()) {
      $this->saveOperation(__FUNCTION__, [$params]);

      return ['SendToWorker' => TRUE];
    }

    $this->loadCredentials();

    $body = [
      'keypair' => [
        'name' => $params['KeyName'],
        'public_key' => $params['PublicKeyMaterial'],
      ],
    ];

    $credentials = [
      'http_method' => 'POST',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::COMPUTE_ENDPOINT, '/os-keypairs'),
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    return $response['status_code'] === 200
      && !empty($response)
      && !empty($response['body'])
      && !empty($response['body']['keypair'])
      ? [
        'KeyPairId' => NULL,
        'KeyName' => $response['body']['keypair']['name'],
        'KeyFingerprint' => $response['body']['keypair']['fingerprint'],
        'KeyMaterial' => $response['body']['keypair']['public_key'],
      ]
      : [];
  }

  /**
   * Delete key pairs List.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array|null
   *   The result.
   */
  public function deleteKeyPair(array $params = []): ?array {
    if ($this->isWorkerResource()) {
      $this->saveOperation(__FUNCTION__, [$params]);

      return ['SendToWorker' => TRUE];
    }

    $this->loadCredentials();

    $credentials = [
      'http_method' => 'DELETE',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::COMPUTE_ENDPOINT, "/os-keypairs/{$params['KeyName']}"),
      'params' => $params,
    ];
    $response = $this->getClient($credentials);

    return ($response['status_code'] === 202 || $response['status_code'] === 204) ? ['Success' => TRUE] : NULL;
  }

  /**
   * Get volume list.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The volume list API response.
   */
  public function describeVolumes(array $params = []) {
    $this->loadCredentials();
    $project_id = $this->credentials['project_id'];

    $credentials = [
      'http_method' => 'get',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::VOLUME_ENDPOINT, "/{$project_id}/volumes/detail"),
    ];
    $response = $this->getClient($credentials);

    $volumes = $response['body']['volumes'] ?? [];

    $results['Volumes'] = [];

    foreach ($volumes ?: [] as $key => $value) {
      $attachments = [];
      if (!empty($value['attachments'])) {
        foreach ($value['attachments'] as $attachment_key => $attachment) {
          $attachments[$attachment_key]['Device'] = $attachment['device'];
          $attachments[$attachment_key]['InstanceId'] = $attachment['server_id'];
          $attachments[$attachment_key]['VolumeId'] = $attachment['volume_id'];
        }
      }
      $tags = [];
      if (!empty($value['metadata'])) {
        foreach ($value['metadata'] as $tag_key => $tag_value) {
          $tags[] = "${tag_key}.${tag_value}";
        }
      }

      $results['Volumes'][$key]['Attachments'] = $attachments;
      $results['Volumes'][$key]['Name'] = $value['name'];
      $results['Volumes'][$key]['Description'] = $value['description'];
      $results['Volumes'][$key]['VolumeId'] = $value['id'];
      $results['Volumes'][$key]['State'] = $value['status'];
      $results['Volumes'][$key]['CreateTime'] = str_replace("T", " ", $value['created_at']);
      $results['Volumes'][$key]['SnapshotId'] = $value['snapshot_id'];
      $results['Volumes'][$key]['VolumeType'] = $value['volume_type'];
      $results['Volumes'][$key]['Size'] = $value['size'];
      $results['Volumes'][$key]['VirtualizationType'] = $value['replication_status'];
      $results['Volumes'][$key]['AvailabilityZone'] = $value['availability_zone'];
      $results['Volumes'][$key]['Encrypted'] = $value['encrypted'];
      $results['Volumes'][$key]['Tags'] = $tags ?? [];
    }

    return $results;
  }

  /**
   * Get volume types.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The volume list API response.
   */
  public function describeVolumeTypes(array $params = []) {
    $this->loadCredentials();
    $project_id = $this->credentials['project_id'];

    $credentials = [
      'http_method' => 'get',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::VOLUME_ENDPOINT, "/{$project_id}/types"),
    ];
    $response = $this->getClient($credentials);

    $types = [];
    array_map(static function ($type) use (&$types) {
      return $types[$type] = $type;
    }, array_column($response['body']['volume_types'], 'name'));

    return $types;
  }

  /**
   * Get a default volume type.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return string
   *   The volume list API response.
   */
  public function getDefaultVolumeType(array $params = []) {
    $this->loadCredentials();
    $project_id = $this->credentials['project_id'];

    $credentials = [
      'http_method' => 'get',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::VOLUME_ENDPOINT, "/{$project_id}/types/default"),
    ];
    $response = $this->getClient($credentials);

    return !empty($response['body']['volume_type']['name'])
      ? $response['body']['volume_type']['name']
      : '';
  }

  /**
   * Create volume.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The volume create API response.
   */
  public function createVolume(array $params = []) {
    if ($this->isWorkerResource()) {
      $this->saveOperation(__FUNCTION__, [$params], 'updateVolumes');

      return ['SendToWorker' => TRUE];
    }

    $this->loadCredentials();

    $body['volume'] = [
      'name' => $params['Name'],
      'size' => $params['Size'],
      'volume_type' => $params['VolumeType'],
      'availability_zone' => $params['AvailabilityZone'],
    ];

    if (!empty($params['SnapshotId'])) {
      $body['volume'] += ['snapshot_id' => $params['SnapshotId']];
    }

    $project_id = $this->credentials['project_id'];

    $credentials = [
      'http_method' => 'post',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::VOLUME_ENDPOINT, "/{$project_id}/volumes"),
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    $volume = $response['body']['volume'] ?? [];

    $attachments = [];

    if (!empty($volume['attachments'])) {
      foreach ($volume['attachments'] ?: [] as $attachment_key => $attachment) {
        $attachments[$attachment_key]['Device'] = $attachment['device'];
        $attachments[$attachment_key]['InstanceId'] = $attachment['server_id'];
        $attachments[$attachment_key]['VolumeId'] = $attachment['volume_id'];
      }
    }

    return [
      'Attachments'        => $attachments ?? [],
      'Name'               => $volume['name'] ?? NULL,
      'Description'        => $volume['description'] ?? NULL,
      'VolumeId'           => $volume['id'] ?? NULL,
      'State'              => $volume['status'] ?? NULL,
      'CreateTime'         => !empty($volume['created_at']) ? str_replace('T', ' ', $volume['created_at']) : NULL,
      'SnapshotId'         => $volume['snapshot_id'] ?? NULL,
      'VolumeType'         => $volume['volume_type'] ?? NULL,
      'Size'               => $volume['size'] ?? NULL,
      'VirtualizationType' => $volume['replication_status'] ?? NULL,
      'AvailabilityZone'   => $volume['availability_zone'] ?? NULL,
      'Encrypted'          => $volume['encrypted'] ?? NULL,
    ];
  }

  /**
   * Update volume.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The volume update API response.
   */
  public function updateVolume(array $params = []) {
    $this->loadCredentials();

    $body['volume'] = [
      'name' => $params['Name'],
    ];

    $project_id = $this->credentials['project_id'];

    $credentials = [
      'http_method' => 'put',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::VOLUME_ENDPOINT, "/{$project_id}/volumes/{$params['VolumeId']}"),
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    $volume = $response['body']['volume'] ?? [];

    $attachments = [];

    if (!empty($volume['attachments'])) {
      foreach ($volume['attachments'] ?: [] as $attachment_key => $attachment) {
        $attachments[$attachment_key]['Device'] = $attachment['device'];
        $attachments[$attachment_key]['InstanceId'] = $attachment['server_id'];
        $attachments[$attachment_key]['VolumeId'] = $attachment['volume_id'];
      }
    }

    return [
      'Attachments'        => $attachments ?? [],
      'Name'               => $volume['name'] ?? NULL,
      'Description'        => $volume['description'] ?? NULL,
      'VolumeId'           => $volume['id'] ?? NULL,
      'State'              => $volume['status'] ?? NULL,
      'CreateTime'         => !empty($volume['created_at']) ? str_replace('T', ' ', $volume['created_at']) : NULL,
      'SnapshotId'         => $volume['snapshot_id'] ?? NULL,
      'VolumeType'         => $volume['volume_type'] ?? NULL,
      'Size'               => $volume['size'] ?? NULL,
      'VirtualizationType' => $volume['replication_status'] ?? NULL,
      'AvailabilityZone'   => $volume['availability_zone'] ?? NULL,
      'Encrypted'          => $volume['encrypted'] ?? NULL,
    ];
  }

  /**
   * Delete volume.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array|null
   *   Succeeded or failed.
   */
  public function deleteVolume(array $params = []): ?array {
    if ($this->isWorkerResource()) {
      $this->saveOperation(__FUNCTION__, [$params], 'updateVolumes');

      return ['SendToWorker' => TRUE];
    }

    $this->loadCredentials();

    $project_id = $this->credentials['project_id'];

    $credentials = [
      'http_method' => 'delete',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::VOLUME_ENDPOINT, "/{$project_id}/volumes/{$params['VolumeId']}"),
    ];

    $response = $this->getClient($credentials);

    return $response['status_code'] === 202 ? ['Success' => TRUE] : NULL;
  }

  /**
   * Attach volume.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array|null
   *   Succeeded or failed.
   */
  public function attachVolume(array $params = []): ?array {
    if ($this->isWorkerResource()) {
      $this->saveOperation(__FUNCTION__, [$params], 'updateVolumes');

      return ['SendToWorker' => TRUE];
    }

    $this->loadCredentials();

    $body['os-attach'] = [
      'instance_uuid' => $params['InstanceId'],
      'mountpoint' => $params['Device'],
    ];

    $project_id = $this->credentials['project_id'];

    $credentials = [
      'http_method' => 'post',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::VOLUME_ENDPOINT, "/{$project_id}/volumes/{$params['VolumeId']}/action"),
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    return $response['status_code'] === 202 ? ['Success' => TRUE] : NULL;
  }

  /**
   * Detach volume.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array|null
   *   Succeeded or failed.
   */
  public function detachVolume(array $params = []): ?array {
    if ($this->isWorkerResource()) {
      $this->saveOperation(__FUNCTION__, [$params], 'updateVolumes');

      return ['SendToWorker' => TRUE];
    }

    $this->loadCredentials();

    $body['os-detach'] = [
      'attachment_id' => $params['InstanceId'] ?? NULL,
    ];

    $project_id = $this->credentials['project_id'];

    $credentials = [
      'http_method' => 'post',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::VOLUME_ENDPOINT, "/{$project_id}/volumes/{$params['VolumeId']}/action"),
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    return $response['status_code'] === 202 ? ['Success' => TRUE] : NULL;
  }

  /**
   * Get snapshot list.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The snapshot list API response.
   */
  public function describeSnapshots(array $params = []) {
    $this->loadCredentials();
    $project_id = $this->credentials['project_id'];

    $credentials = [
      'http_method' => 'get',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::VOLUME_ENDPOINT, "/{$project_id}/snapshots/detail"),
    ];
    $response = $this->getClient($credentials);

    $snapshots = $response['body']['snapshots'] ?? [];

    $results['Snapshots'] = [];

    foreach ($snapshots ?: [] as $key => $snapshot) {
      $tags = [];
      if (!empty($snapshot['metadata'])) {
        foreach ($snapshot['metadata'] as $tag_key => $tag_value) {
          $tags[] = "${tag_key}.${tag_value}";
        }
      }
      $results['Snapshots'][$key]['Description'] = $snapshot['description'] ?? NULL;
      $results['Snapshots'][$key]['OwnerId'] = $snapshot['os-extended-snapshot-attributes:project_id'] ?? NULL;
      $results['Snapshots'][$key]['Progress'] = $snapshot['os-extended-snapshot-attributes:progress'] ?? NULL;
      $results['Snapshots'][$key]['name'] = $snapshot['name'] ?? NULL;
      $results['Snapshots'][$key]['SnapshotId'] = $snapshot['id'] ?? NULL;
      $results['Snapshots'][$key]['StartTime'] = $snapshot['created_at'] ?? NULL;
      $results['Snapshots'][$key]['State'] = $snapshot['status'] ?? NULL;
      $results['Snapshots'][$key]['VolumeId'] = $snapshot['volume_id'] ?? NULL;
      $results['Snapshots'][$key]['VolumeSize'] = $snapshot['size'] ?? NULL;
      $results['Snapshots'][$key]['Tags'] = $tags ?? [];
    }

    return $results;
  }

  /**
   * Create snapshot.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The snapshot create API response.
   */
  public function createSnapshot(array $params = []) {
    if ($this->isWorkerResource()) {
      $this->saveOperation(__FUNCTION__, [$params], 'updateSnapshots');

      return ['SendToWorker' => TRUE];
    }

    $this->loadCredentials();

    $body['snapshot'] = [
      'name' => $params['Name'],
      'description' => $params['Description'],
      'volume_id' => $params['VolumeId'],
    ];

    $project_id = $this->credentials['project_id'];

    $credentials = [
      'http_method' => 'post',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::VOLUME_ENDPOINT, "/{$project_id}/snapshots"),
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    $snapshot = $response['body']['snapshot'] ?? [];

    return [
      'Description' => $snapshot['description'] ?? NULL,
      'OwnerId'     => $snapshot['os-extended-snapshot-attributes:project_id'] ?? NULL,
      'Progress'    => $snapshot['os-extended-snapshot-attributes:progress'] ?? NULL,
      'name'        => $snapshot['name'] ?? NULL,
      'SnapshotId'  => $snapshot['id'] ?? NULL,
      'StartTime'   => !empty($snapshot['created_at']) ? $snapshot['created_at'] : NULL,
      'State'       => $snapshot['status'] ?? NULL,
      'VolumeId'    => $snapshot['volume_id'] ?? NULL,
      'VolumeSize'  => $snapshot['size'] ?? NULL,
      'Tags'        => $snapshot['metadata'] ?? [],
    ];
  }

  /**
   * Update snapshot.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The snapshot update API response.
   */
  public function updateSnapshot(array $params = []) {
    if ($this->isWorkerResource()) {
      $this->saveOperation(__FUNCTION__, [$params], 'updateSnapshots');

      return ['SendToWorker' => TRUE];
    }

    $this->loadCredentials();

    $body['snapshot'] = [
      'name' => $params['Name'],
      'description' => $params['Description'],
    ];

    $project_id = $this->credentials['project_id'];

    $credentials = [
      'http_method' => 'put',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::VOLUME_ENDPOINT, "/{$project_id}/snapshots/{$params['SnapshotId']}"),
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    $snapshot = $response['body']['snapshot'] ?? [];

    return [
      'Description' => $snapshot['description'] ?? NULL,
      'OwnerId'     => $snapshot['os-extended-snapshot-attributes:project_id'] ?? NULL,
      'Progress'    => $snapshot['os-extended-snapshot-attributes:progress'] ?? NULL,
      'name'        => $snapshot['name'] ?? NULL,
      'SnapshotId'  => $snapshot['id'] ?? NULL,
      'StartTime'   => $snapshot['created_at'] ?? NULL,
      'State'       => $snapshot['status'] ?? NULL,
      'VolumeId'    => $snapshot['volume_id'] ?? NULL,
      'VolumeSize'  => $snapshot['size'] ?? NULL,
      'Tags'        => $snapshot['metadata'] ?? [],
    ];
  }

  /**
   * Delete snapshot.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array|null
   *   Succeeded or failed.
   */
  public function deleteSnapshot(array $params = []): ?array {
    if ($this->isWorkerResource()) {
      $this->saveOperation(__FUNCTION__, [$params], 'updateSnapshots');

      return ['SendToWorker' => TRUE];
    }

    $this->loadCredentials();

    $project_id = $this->credentials['project_id'];

    $credentials = [
      'http_method' => 'delete',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::VOLUME_ENDPOINT, "/{$project_id}/snapshots/{$params['SnapshotId']}"),
    ];

    $response = $this->getClient($credentials);

    return $response['status_code'] === 202 ? ['Success' => TRUE] : NULL;
  }

  /**
   * Get flavors.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The Flavor list API response.
   */
  public function describeFlavors(array $params = []) {
    if ($this->isWorkerResource()) {
      return ['SendToWorker' => TRUE];
    }

    $this->loadCredentials();

    $credentials = [
      'http_method' => 'get',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::COMPUTE_ENDPOINT, '/flavors'),
      'params' => $params,
    ];

    $response = $this->getClient($credentials);

    return $response['body']['flavors'] ?? [];
  }

  /**
   * Get Flavor Detail of specific flavor Id.
   *
   * @param array $params
   *   Parameters to pass to the API.
   * @param string $flavor_id
   *   The Flavor Id.
   *
   * @return array
   *   The Flavor list API response.
   */
  public function describeFlavorById(array $params = [], $flavor_id = ''): array {
    $this->loadCredentials();

    $credentials = [
      'http_method' => 'get',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::COMPUTE_ENDPOINT, "/flavors/{$flavor_id}"),
      'params' => $params,
    ];

    $response = $this->getClient($credentials);

    return !empty($response['body']['flavor']) ? $response['body']['flavor'] : [];
  }

  /**
   * Get volume Detail of specific volume Id.
   *
   * @param array $params
   *   Parameters to pass to the API.
   * @param string $volume_id
   *   The Volume Id.
   *
   * @return array
   *   The volume list API response.
   */
  public function describeVolumeById(array $params = [], $volume_id = '') {
    $this->loadCredentials();
    $project_id = $this->credentials['project_id'];

    $credentials = [
      'http_method' => 'get',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::VOLUME_ENDPOINT, "/{$project_id}/volumes/{$volume_id}"),
      'params' => $params,
    ];

    $response = $this->getClient($credentials);

    return $response['body']['volume'] ?? [];
  }

  /**
   * Get security group list.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The security group list API response.
   */
  public function describeSecurityGroups(array $params = []) {
    $this->loadCredentials();
    $project_id = $this->credentials['project_id'] ?? NULL;

    $endpoint = !empty($params['id'])
      ? $this->getEndpoint(OpenStackServiceInterface::NETWORK_ENDPOINT, "/security-groups?project_id={$project_id}&id={$params['id']}")
      : $this->getEndpoint(OpenStackServiceInterface::NETWORK_ENDPOINT, "/security-groups?project_id={$project_id}");

    $credentials = [
      'http_method' => 'get',
      'endpoint' => $endpoint,
      'params' => $params,
    ];
    $response = $this->getClient($credentials);

    $security_groups = $response['body']['security_groups'] ?? [];

    $results['SecurityGroups'] = [];

    foreach ($security_groups ?: [] as $key => $value) {

      $ip_permission_cnt = $ip_permission_egress_cnt = 0;

      $ip_permissions = $ip_permissions_egress = [];

      $security_group_rules = [];
      if (!empty($value['security_group_rules'])) {
        $security_group_rules = $value['security_group_rules'];
      }

      foreach ($security_group_rules ?: [] as $value_rules) {
        if (!empty($value_rules['direction']) && $value_rules['direction'] === 'egress') {

          $ip_permissions_egress[$ip_permission_egress_cnt] = [
            'RuleId'     => $value_rules['id'] ?? NULL,
            'FromPort'   => $value_rules['port_range_min'] ?? 1,
            'IpProtocol' => $value_rules['protocol'] ?? '-1',
            'ToPort'     => $value_rules['port_range_max'] ?? 65535,
          ];

          if (!empty($value_rules['remote_group_id'])) {
            $ip_permissions_egress[$ip_permission_egress_cnt] += [
              'UserIdGroupPairs' => [
                [
                  'GroupId' => $value_rules['remote_group_id'] ?? NULL,
                  'UserId'  => $value_rules['tenant_id'] ?? NULL,
                ],
              ],
            ];
          }

          if ((!empty($value_rules['ethertype'])
            && $value_rules['ethertype'] === 'IPv6')
            && empty($value_rules['remote_group_id'])) {
            $ip_permissions_egress[$ip_permission_egress_cnt] += [
              'Ipv6Ranges' => [
                [
                  'CidrIpv6'    => !empty($value_rules['remote_ip_prefix']) ? $value_rules['remote_ip_prefix'] : '::/0',
                  'Description' => !empty($value_rules['description']) ? $value_rules['description'] : '',
                ],
              ],
            ];
          }

          if ((!empty($value_rules['ethertype'])
            && $value_rules['ethertype'] === 'IPv4')
            && empty($value_rules['remote_group_id'])) {
            $ip_permissions_egress[$ip_permission_egress_cnt] += [
              'IpRanges' => [
                [
                  'CidrIp'      => !empty($value_rules['remote_ip_prefix']) ? $value_rules['remote_ip_prefix'] : '0.0.0.0/0',
                  'Description' => !empty($value_rules['description']) ? $value_rules['description'] : '',
                ],
              ],
            ];
          }

          $ip_permission_egress_cnt++;

        }
        else {
          $ip_permissions[$ip_permission_cnt] = [
            'RuleId'     => $value_rules['id'] ?? NULL,
            'FromPort'   => $value_rules['port_range_min'] ?? 1,
            'IpProtocol' => $value_rules['protocol'] ?? '-1',
            'ToPort'     => $value_rules['port_range_max'] ?? 65535,
          ];

          if (!empty($value_rules['remote_group_id'])) {
            $ip_permissions[$ip_permission_cnt] += [
              'UserIdGroupPairs' => [
                [
                  'GroupId' => $value_rules['remote_group_id'] ?? NULL,
                  'UserId'  => $value_rules['tenant_id'] ?? NULL,
                ],
              ],
            ];
          }

          if (!empty($value_rules['ethertype'])
            && $value_rules['ethertype'] === 'IPv6'
            && empty($value_rules['remote_group_id'])) {
            $ip_permissions[$ip_permission_cnt] += [
              'Ipv6Ranges' => [
                [
                  'CidrIpv6'    => !empty($value_rules['remote_ip_prefix']) ? $value_rules['remote_ip_prefix'] : '::/0',
                  'Description' => !empty($value_rules['description']) ? $value_rules['description'] : '',
                ],
              ],
            ];
          }
          if (!empty($value_rules['ethertype'])
            && $value_rules['ethertype'] === 'IPv4'
            && empty($value_rules['remote_group_id'])) {
            $ip_permissions[$ip_permission_cnt] += [
              'IpRanges' => [
                [
                  'CidrIp'      => !empty($value_rules['remote_ip_prefix']) ? $value_rules['remote_ip_prefix'] : '0.0.0.0/0',
                  'Description' => !empty($value_rules['description']) ? $value_rules['description'] : '',
                ],
              ],
            ];
          }

          $ip_permission_cnt++;
        }
      }

      $results['SecurityGroups'][$key] = [
        'Description'         => $value['description'] ?? NULL,
        'GroupName'           => $value['name'] ?? NULL,
        'IpPermissions'       => $ip_permissions ?? [],
        'OwnerId'             => $value['tenant_id'] ?? NULL,
        'GroupId'             => $value['id'] ?? NULL,
        'IpPermissionsEgress' => $ip_permissions_egress ?? [],
        'Tags'                => $value['tags'] ?? [],
      ];
    }

    return $results;
  }

  /**
   * Create security group list.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The security group create API response.
   */
  public function createSecurityGroup(array $params = []): array {
    if ($this->isWorkerResource()) {
      $this->saveOperation(__FUNCTION__, [$params], 'updateSecurityGroups');

      return ['SendToWorker' => TRUE];
    }

    $this->loadCredentials();
    $project_id = $this->credentials['project_id'];

    $body['security_group'] = [
      'name' => $params['GroupName'],
      'description' => $params['Description'],
    ];

    $credentials = [
      'http_method' => 'post',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::NETWORK_ENDPOINT, "/security-groups?project_id={$project_id}"),
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    $security_group = $response['body']['security_group'] ?? [];

    return [
      'Description'         => $security_group['description'] ?? NULL,
      'GroupName'           => $security_group['name'] ?? NULL,
      'IpPermissions'       => [],
      'OwnerId'             => $security_group['tenant_id'] ?? NULL,
      'GroupId'             => $security_group['id'] ?? NULL,
      'IpPermissionsEgress' => [],
      'Tags'                => $security_group['tags'] ?? [],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function updateSecurityGroup(array $params = []): array {
    if ($this->isWorkerResource()) {
      $this->saveOperation(__FUNCTION__, [$params], 'updateSecurityGroups');

      return ['SendToWorker' => TRUE];
    }

    $this->loadCredentials();
    $project_id = $this->credentials['project_id'];

    $body['security_group'] = [
      'name' => $params['GroupName'],
    ];

    $credentials = [
      'http_method' => 'put',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::NETWORK_ENDPOINT, "/security-groups/{$params['GroupId']}?project_id={$project_id}"),
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    $security_group = $response['body']['security_group'] ?? [];

    return [
      'Description'         => $security_group['description'] ?? NULL,
      'GroupName'           => $security_group['name'] ?? NULL,
      'IpPermissions'       => [],
      'OwnerId'             => $security_group['tenant_id'] ?? NULL,
      'GroupId'             => $security_group['id'] ?? NULL,
      'IpPermissionsEgress' => [],
      'Tags'                => $security_group['tags'] ?? [],
    ];
  }

  /**
   * Delete security group List.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array|null
   *   Succeeded or failed.
   */
  public function deleteSecurityGroup(array $params = []): ?array {
    if ($this->isWorkerResource()) {
      $this->saveOperation(__FUNCTION__, [$params], 'updateSecurityGroups');

      return ['SendToWorker' => TRUE];
    }

    $this->loadCredentials();

    $credentials = [
      'http_method' => 'delete',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::NETWORK_ENDPOINT, "/security-groups/{$params['GroupId']}"),
    ];

    $response = $this->getClient($credentials);

    return $response['status_code'] === 204 ? ['Success' => TRUE] : NULL;
  }

  /**
   * {@inheritdoc}
   */
  public function createSecurityGroupRule(array $params = []) {
    if ($this->isWorkerResource()) {
      $this->saveOperation(__FUNCTION__, [$params], 'updateSecurityGroups', ['GroupIds' => [$params['GroupId']]]);

      return ['SendToWorker' => TRUE];
    }

    $this->loadCredentials();
    $project_id = $this->credentials['project_id'];

    $cidr_ip = !empty($params['CidrIp']) ? $params['CidrIp'] : NULL;
    $remote_ip = !empty($params['CidrIpv6']) ? $params['CidrIpv6'] : $cidr_ip;
    $remote_group_id = !empty($params['RemoteGroupId']) ? $params['RemoteGroupId'] : NULL;

    $body['security_group_rule'] = [
      'direction'         => $params['Direction'],
      'port_range_min'    => $params['FromPort'],
      'port_range_max'    => $params['ToPort'],
      'ethertype'         => $params['Source'],
      'protocol'          => $params['IpProtocol'],
      'remote_group_id'   => $remote_group_id,
      'security_group_id' => $params['GroupId'],
      'remote_ip_prefix'  => $remote_ip,
      'description'       => $params['Description'],
    ];

    $credentials = [
      'http_method' => 'post',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::NETWORK_ENDPOINT, "/security-group-rules?project_id={$project_id}"),
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    $security_group_rule = $response['body']['security_group_rule'] ?? [];

    $ip_permissions = $ip_permissions_egress = [];

    if (!empty($security_group_rule['direction'])
      && $security_group_rule['direction'] === 'egress') {
      $ip_permissions_egress = [
        'RuleId'     => $security_group_rule['id'] ?? NULL,
        'FromPort'   => $security_group_rule['port_range_min'] ?? 1,
        'IpProtocol' => $security_group_rule['protocol'] ?? '-1',
        'ToPort'     => $security_group_rule['port_range_max'] ?? 65535,
      ];

      if (!empty($security_group_rule['remote_group_id'])) {
        $ip_permissions_egress += [
          'UserIdGroupPairs' => [
            [
              'GroupId' => $security_group_rule['remote_group_id'] ?? NULL,
              'UserId'  => $security_group_rule['tenant_id'] ?? NULL,
            ],
          ],
        ];
      }

      if ((!empty($security_group_rule['ethertype'])
        && $security_group_rule['ethertype'] === 'IPv6')
        && empty($security_group_rule['remote_group_id'])) {
        $ip_permissions_egress += [
          'Ipv6Ranges' => [
            [
              'CidrIpv6'    => $security_group_rule['remote_ip_prefix'] ?? NULL,
              'Description' => $security_group_rule['description'] ?? NULL,
            ],
          ],
        ];
      }
      if ((!empty($security_group_rule['ethertype'])
        && $security_group_rule['ethertype'] === 'IPv4')
        && empty($security_group_rule['remote_group_id'])) {
        $ip_permissions_egress += [
          'IpRanges' => [
            [
              'CidrIp'      => $security_group_rule['remote_ip_prefix'] ?? NULL,
              'Description' => $security_group_rule['description'] ?? NULL,
            ],
          ],
        ];
      }
    }
    else {
      $ip_permissions = [
        'RuleId'     => $security_group_rule['id'] ?? NULL,
        'FromPort'   => $security_group_rule['port_range_min'] ?? 1,
        'IpProtocol' => $security_group_rule['protocol'] ?? '-1',
        'ToPort'     => $security_group_rule['port_range_max'] ?? 65535,
      ];

      if (!empty($security_group_rule['remote_group_id'])) {
        $ip_permissions += [
          'UserIdGroupPairs' => [
            [
              'GroupId' => $security_group_rule['remote_group_id'] ?? NULL,
              'UserId'  => $security_group_rule['tenant_id'] ?? NULL,
            ],
          ],
        ];
      }

      if ((!empty($security_group_rule['ethertype'])
        && $security_group_rule['ethertype'] === 'IPv6')
        && empty($security_group_rule['remote_group_id'])) {
        $ip_permissions += [
          'Ipv6Ranges' => [
            [
              'CidrIpv6'    => $security_group_rule['remote_ip_prefix'] ?? NULL,
              'Description' => $security_group_rule['description'] ?? NULL,
            ],
          ],
        ];
      }
      if ((!empty($security_group_rule['ethertype'])
        && $security_group_rule['ethertype'] === 'IPv4')
        && empty($security_group_rule['remote_group_id'])) {
        $ip_permissions += [
          'IpRanges' => [
            [
              'CidrIp'      => $security_group_rule['remote_ip_prefix'] ?? NULL,
              'Description' => $security_group_rule['description'] ?? NULL,
            ],
          ],
        ];
      }
    }

    return [
      'RuleId'              => $security_group_rule['id'] ?? NULL,
      'SecurityGroups'      => ['ip_permissions' => $ip_permissions ?? []],
      'IpPermissionsEgress' => $ip_permissions_egress ?? [],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function revokeSecurityGroupRule(array $params = []) {
    if ($this->isWorkerResource()) {
      $this->saveOperation(__FUNCTION__, [$params], 'updateSecurityGroups');

      return ['SendToWorker' => TRUE];
    }

    $this->loadCredentials();

    $credentials = [
      'http_method' => 'delete',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::NETWORK_ENDPOINT, "/security-group-rules/{$params['RuleId']}"),
    ];

    $response = $this->getClient($credentials);

    return $response['status_code'] === 204;
  }

  /**
   * Get Subnet Detail.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The Subnet Detail list API response.
   */
  public function describeSubnetDetail(array $params = []) {
    $this->loadCredentials();

    $credentials = [
      'http_method' => 'get',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::NETWORK_ENDPOINT, "/subnets/{$params['SubnetId']}"),
    ];

    $response = $this->getClient($credentials);

    return $response['body']['subnet'] ?? [];
  }

  /**
   * Get network interface list.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The network interface list API response.
   */
  public function describeNetworkInterfaces(array $params = []) {
    if ($this->isWorkerResource()) {
      return ['SendToWorker' => TRUE];
    }

    $this->loadCredentials();

    $credentials = [
      'http_method' => 'get',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::NETWORK_ENDPOINT, '/networks'),
      'params' => $params,
    ];
    $response = $this->getClient($credentials);

    $networks = $response['body']['networks'] ?? [];

    $results['NetworkInterfaces'] = [];

    foreach ($networks ?: [] as $key => $value) {
      $subnets = [];
      foreach ($value['subnets'] ?: [] as $subnet_id) {
        $subnet_result = $this->describeSubnetDetail(['SubnetId' => $subnet_id]);
        $subnets[] = $subnet_result['name'];

      }
      $subnet_names = implode(', ', $subnets);

      $results['NetworkInterfaces'][$key] = [
        'Name'               => $value['name'] ?? NULL,
        'Description'        => $value['description'] ?? NULL,
        'Groups'             => [],
        'MacAddress'         => NULL,
        'OwnerId'            => $value['tenant_id'] ?? NULL,
        'NetworkInterfaceId' => $value['id'] ?? NULL,
        'PrivateIpAddress'   => NULL,
        'PrivateIpAddresses' => [],
        'RequesterManaged'   => NULL,
        'SourceDestCheck'    => NULL,
        'Status'             => $value['status'] === 'ACTIVE' ? 'available' : '',
        'SubnetId'           => $subnet_names ?? NULL,
        'TagSet'             => $value['tags'] ?? [],
        'VpcId'              => NULL,
        'External'           => $value['router:external'],
      ];
    }

    return $results;
  }

  /**
   * Create network interface.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The network interface create API response.
   */
  public function createNetworkInterface(array $params = []): array {
    if ($this->isWorkerResource()) {
      $this->saveOperation(__FUNCTION__, [$params], 'updateNetworkInterfaces');

      return ['SendToWorker' => TRUE];
    }

    $this->loadCredentials();
    $project_id = $this->credentials['project_id'];

    $body['network'] = [
      'name' => $params['Name'],
      'description' => $params['Description'],
    ];

    $credentials = [
      'http_method' => 'post',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::NETWORK_ENDPOINT, "/networks?project_id={$project_id}"),
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    $network_interface = $response['body']['network'] ?? [];

    $subnets = [];
    $response_subnets = array_key_exists('subnets', $network_interface) ? $network_interface['subnets'] : [];
    foreach ($response_subnets ?: [] as $subnet_id) {
      $subnet_result = $this->describeSubnetDetail(['SubnetId' => $subnet_id]);
      $subnets[] = $subnet_result['name'];

    }
    $subnet_names = implode(', ', $subnets);

    return [
      'NetworkInterface' => [
        'Name'               => $network_interface['name'] ?? NULL,
        'Description'        => $network_interface['description'] ?? NULL,
        'Groups'             => [],
        'MacAddress'         => NULL,
        'OwnerId'            => $network_interface['tenant_id'] ?? NULL,
        'NetworkInterfaceId' => $network_interface['id'] ?? NULL,
        'PrivateIpAddress'   => NULL,
        'PrivateIpAddresses' => [],
        'RequesterManaged'   => NULL,
        'SourceDestCheck'    => NULL,
        'Status'             => $network_interface['status'] ?? NULL,
        'SubnetId'           => $subnet_names ?? NULL,
        'TagSet'             => $network_interface['tags'] ?? [],
        'VpcId'              => NULL,
      ],
    ];
  }

  /**
   * Update network interface.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The network interface update API response.
   */
  public function updateNetworkInterface(array $params = []): array {
    if ($this->isWorkerResource()) {
      $this->saveOperation(__FUNCTION__, [$params], 'updateNetworkInterfaces');

      return ['SendToWorker' => TRUE];
    }

    $this->loadCredentials();
    $project_id = $this->credentials['project_id'];

    $body['network'] = [
      'name' => $params['Name'],
      'description' => $params['Description'],
    ];

    $credentials = [
      'http_method' => 'put',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::NETWORK_ENDPOINT, "/networks/{$params['NetworkInterfaceId']}?project_id={$project_id}"),
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    $network_interface = $response['body']['network'] ?? [];

    $subnets = [];
    $response_subnets = array_key_exists('subnets', $network_interface) ? $network_interface['subnets'] : [];
    foreach ($response_subnets ?: [] as $subnet_id) {
      $subnet_result = $this->describeSubnetDetail(['SubnetId' => $subnet_id]);
      $subnets[] = $subnet_result['name'];

    }
    $subnet_names = implode(', ', $subnets);

    return [
      'NetworkInterface' => [
        'Name'               => $network_interface['name'] ?? NULL,
        'Description'        => $network_interface['description'] ?? NULL,
        'Groups'             => [],
        'MacAddress'         => NULL,
        'OwnerId'            => $network_interface['tenant_id'] ?? NULL,
        'NetworkInterfaceId' => $network_interface['id'] ?? NULL,
        'PrivateIpAddress'   => NULL,
        'PrivateIpAddresses' => [],
        'RequesterManaged'   => NULL,
        'SourceDestCheck'    => NULL,
        'Status'             => $network_interface['status'] ?? NULL,
        'SubnetId'           => $subnet_names ?? NULL,
        'TagSet'             => $network_interface['tags'] ?? [],
        'VpcId'              => NULL,
      ],
    ];
  }

  /**
   * Delete network interface.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array|null
   *   Succeeded or failed.
   */
  public function deleteNetworkInterface(array $params = []): ?array {
    if ($this->isWorkerResource()) {
      $this->saveOperation(__FUNCTION__, [$params], 'updateNetworkInterfaces');

      return ['SendToWorker' => TRUE];
    }

    $this->loadCredentials();

    $credentials = [
      'http_method' => 'delete',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::NETWORK_ENDPOINT, "/networks/{$params['NetworkInterfaceId']}"),
    ];

    $response = $this->getClient($credentials);

    return $response['status_code'] === 204 ? ['Success' => TRUE] : NULL;
  }

  /**
   * Get Floating IP List.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The Floating IP list API response.
   */
  public function describeAddresses(array $params = []): array {
    if ($this->isWorkerResource()) {
      return ['SendToWorker' => TRUE];
    }

    $this->loadCredentials();
    $project_id = $this->credentials['project_id'];

    $credentials = [
      'http_method' => 'get',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::NETWORK_ENDPOINT, "/floatingips?project_id={$project_id}"),
      'params' => $params,
    ];
    $response = $this->getClient($credentials);

    $floating_ips = $response['body']['floatingips'] ?? [];

    $results['Addresses'] = [];

    foreach ($floating_ips ?: [] as $key => $value) {
      $results['Addresses'][$key] = [
        'FloatingIpId'            => $value['id'],
        'PublicIp'                => $value['floating_ip_address'] ?? NULL,
        'PrivateIpAddress'        => $value['fixed_ip_address'] ?? NULL,
        'NetworkInterfaceOwnerId' => $value['tenant_id'] ?? NULL,
        'NetworkInterfaceId'      => $value['port_details'] ? $value['port_details']['network_id'] : NULL,
        'InstanceId'              => $value['port_details'] ? $value['port_details']['device_id'] : NULL,
        'AssociationId'           => $value['port_id'] ?? NULL,
        'AllocationId'            => $value['floating_network_id'] ?? NULL,
        'Tags'                    => $value['tags'] ?? [],
      ];
    }
    return $results;
  }

  /**
   * Create Floating IP.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The Floating IP create API response.
   */
  public function allocateAddress(array $params = []) {
    if ($this->isWorkerResource()) {
      $this->saveOperation(__FUNCTION__, [$params]);
      return ['SendToWorker' => TRUE];
    }

    $this->loadCredentials();

    $body['floatingip'] = [
      'floating_network_id' => $params['AllocationId'],
    ];

    $credentials = [
      'http_method' => 'post',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::NETWORK_ENDPOINT, '/floatingips'),
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    $floating_ip = $response['body']['floatingip'] ?? [];

    return [
      'FloatingIpId'            => $floating_ip['id'] ?? NULL,
      'PublicIp'                => $floating_ip['floating_ip_address'] ?? NULL,
      'PrivateIpAddress'        => $floating_ip['fixed_ip_address'] ?? NULL,
      'NetworkInterfaceOwnerId' => $floating_ip['tenant_id'] ?? NULL,
      'NetworkInterfaceId'      => $floating_ip['port_details'] ? $floating_ip['port_details']['network_id'] : NULL,
      'InstanceId'              => $floating_ip['port_details'] ? $floating_ip['port_details']['device_id'] : NULL,
      'Domain'                  => $floating_ip['dns_domain'] ?? NULL,
      'AssociationId'           => $floating_ip['port_id'] ?? NULL,
      'AllocationId'            => $floating_ip['floating_network_id'] ?? NULL,
      'Tags'                    => $floating_ip['tags'] ?? [],
    ];
  }

  /**
   * Associate Floating IP.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The Floating IP associate API response.
   */
  public function associateAddress(array $params = []): array {
    if ($this->isWorkerResource()) {
      $this->saveOperation(__FUNCTION__, [$params]);
      return ['SendToWorker' => TRUE];
    }

    $this->loadCredentials();

    $body['floatingip'] = [
      'port_id' => $params['PortId'],
    ];

    $credentials = [
      'http_method' => 'put',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::NETWORK_ENDPOINT, "/floatingips/{$params['FloatingIpId']}"),
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    $floating_ip = $response['body']['floatingip'] ?? [];

    return !empty($response) && !empty($response['body'])
      && $response['status_code'] === 200 ?
      [
        'FloatingIpId'            => $floating_ip['id'] ?? NULL,
        'PublicIp'                => $floating_ip['floating_ip_address'] ?? NULL,
        'PrivateIpAddress'        => $floating_ip['fixed_ip_address'] ?? NULL,
        'NetworkInterfaceOwnerId' => $floating_ip['tenant_id'] ?? NULL,
        'NetworkInterfaceId'      => !empty($floating_ip['port_details']) ? $floating_ip['port_details']['network_id'] : NULL,
        'InstanceId'              => !empty($floating_ip['port_details']) ? $floating_ip['port_details']['device_id'] : NULL,
        'Domain'                  => $floating_ip['dns_domain'] ?? NULL,
        'AssociationId'           => $floating_ip['port_id'] ?? NULL,
        'Tags'                    => $floating_ip['tags'] ?? [],
      ]
    : [];
  }

  /**
   * Disassociate Floating IP.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The Floating IP associate API response.
   */
  public function disassociateAddress(array $params = []): array {
    if ($this->isWorkerResource()) {
      $this->saveOperation(__FUNCTION__, [$params]);
      return ['SendToWorker' => TRUE];
    }

    $this->loadCredentials();

    $body['floatingip'] = [
      'port_id' => NULL,
    ];

    $credentials = [
      'http_method' => 'put',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::NETWORK_ENDPOINT, "/floatingips/{$params['FloatingIpId']}"),
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    $floating_ip = $response['body']['floatingip'] ?? [];

    return [
      'FloatingIpId'            => $floating_ip['id'] ?? NULL,
      'PublicIp'                => $floating_ip['floating_ip_address'] ?? NULL,
      'PrivateIpAddress'        => $floating_ip['fixed_ip_address'] ?? NULL,
      'NetworkInterfaceOwnerId' => $floating_ip['tenant_id'] ?? NULL,
      'NetworkInterfaceId'      => $floating_ip['port_details'] ? $floating_ip['port_details']['network_id'] : NULL,
      'InstanceId'              => $floating_ip['port_details'] ? $floating_ip['port_details']['device_id'] : NULL,
      'Domain'                  => $floating_ip['dns_domain'] ?? NULL,
      'AssociationId'           => $floating_ip['port_id'] ?? NULL,
      'Tags'                    => $floating_ip['tags'] ?? [],
    ];
  }

  /**
   * Instance Port Interface.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The Instance Port Interface list API response.
   */
  public function instancePortInterface(array $params = []): array {
    $this->loadCredentials();

    $credentials = [
      'http_method' => 'get',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::COMPUTE_ENDPOINT, "/servers/{$params['InstanceId']}/os-interface"),
    ];

    $response = $this->getClient($credentials);

    return $response['body']['interfaceAttachments'] ?? [];
  }

  /**
   * Get NetworkInterface.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The NetworkInterface list API response.
   */
  public function describeNetworkInterfacesPorts(array $params = []) {
    $this->loadCredentials();
    $project_id = $this->credentials['project_id'];

    $credentials = [
      'http_method' => 'get',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::NETWORK_ENDPOINT, "/ports?project_id={$project_id}"),
      'params' => $params,
    ];
    $response = $this->getClient($credentials);

    $ports = $response['body']['ports'] ?? [];

    $results = [];

    foreach ($ports ?: [] as $port) {
      $results[$port['network_id']][$port['id']] = $port['id'];
    }

    return $results;
  }

  /**
   * Delete Floating IP.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array|null
   *   The Floating IP delete API response.
   */
  public function releaseAddress(array $params = []): ?array {
    if ($this->isWorkerResource()) {
      $this->saveOperation(__FUNCTION__, [$params]);
      return ['SendToWorker' => TRUE];
    }

    $this->loadCredentials();

    $credentials = [
      'http_method' => 'delete',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::NETWORK_ENDPOINT, "/floatingips/{$params['FloatingIpId']}"),
    ];

    $response = $this->getClient($credentials);

    return $response['status_code'] === 204 ? ['Success' => TRUE] : NULL;
  }

  /**
   * {@inheritdoc}
   */
  public function describeAvailabilityZones(array $params = []) {
    $this->loadCredentials();

    $credentials = [
      'http_method' => 'get',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::COMPUTE_ENDPOINT, '/os-availability-zone'),
      'params' => $params,
    ];

    $response = $this->getClient($credentials);

    $availability_zones = $response['body']['availabilityZoneInfo'] ?? [];

    $results['AvailabilityZones'] = [];

    foreach ($availability_zones ?: [] as $key => $value) {
      $results['AvailabilityZones'][$key] = [
        'ZoneName' => $value['zoneName'] ?? NULL,
        'State'   => $value['zoneState']['available'] === TRUE ? 'available' : '',
      ];
    }
    return $results;
  }

  /**
   * Call the API for updated entities and store them as Instance entities.
   *
   * @param array $params
   *   Optional parameters array.
   * @param bool $clear
   *   TRUE to clear stale security groups.
   * @param string $entity_type
   *   Entity type string.
   * @param string $cloud_context
   *   Cloud context string.
   *
   * @return bool
   *   Indicates success so failure.
   */
  public function updateInstanceEntities(array $params = [], $clear = TRUE, $entity_type = '', $cloud_context = ''): bool {
    if ($this->isWorkerResource()) {
      return TRUE;
    }

    $updated = FALSE;
    // Call the api and get all instances.
    $result = $this->describeInstances($params);
    if ($result !== NULL) {
      $all_instances = $this->cloudService->loadAllEntities($this->cloudContext, $entity_type);
      $stale = [];
      // Make it easier to lookup the instances by setting up
      // the array with the instance_id.
      foreach ($all_instances ?: [] as $instance) {
        $stale[$instance->getInstanceId()] = $instance;
      }

      /** @var \Drupal\Core\Batch\BatchBuilder $batch_builder */
      $batch_builder = $this->initBatch('Instance Update');
      // Loop through the reservations and store each one as an Instance entity.
      // Testcase throws an undefined index: Reservations error to resolve
      // that error needs to be add isset condition.
      if (isset($result['Reservations'])) {
        foreach ($result['Reservations'] ?: [] as $reservation) {

          foreach ($reservation['Instances'] ?: [] as $instance) {
            // Keep track of instances that do not exist anymore
            // delete them after saving the rest of the instances.
            if (!empty($stale[$instance['InstanceId']])) {
              unset($stale[$instance['InstanceId']]);
            }
            // Store the Reservation OwnerId in instance so batch
            // callback has access.
            $instance['reservation_ownerid'] = $reservation['OwnerId'] ?? NULL;
            $instance['reservation_id'] = $reservation['ReservationId'] ?? NULL;

            $batch_builder->addOperation([
              OpenStackBatchOperations::class,
              'updateInstance',
            ], [$cloud_context, $instance]);

          }
        }
      }
      $batch_builder->addOperation([
        OpenStackBatchOperations::class,
        'finished',
      ], [$entity_type, $stale, $clear]);
      $this->runBatch($batch_builder);
      $updated = TRUE;
    }
    return $updated;
  }

  /**
   * Update the Instances.
   *
   * @param array $params
   *   Parameters array to send to API.
   * @param bool $clear
   *   TRUE to clear stale instances.
   *
   * @return bool
   *   Indicates success or failure.
   */
  public function updateInstances(array $params = [], $clear = TRUE): bool {
    $entity_type = 'openstack_instance';
    $lock_name = $this->getLockKey($entity_type);

    if (!$this->lock->acquire($lock_name)) {
      return FALSE;
    }

    $updated = $this->updateInstanceEntities($params, $clear, $entity_type, $this->cloudContext);
    $this->lock->release($lock_name);

    return $updated;
  }

  /**
   * Update the OpenStackImages.
   *
   * @param array $params
   *   Optional parameters array.
   * @param bool $clear
   *   TRUE to clear stale security groups.
   * @param bool $save_operation
   *   TRUE to save operation and not to execute.
   *
   * @return bool
   *   Indicates success so failure.
   */
  public function updateImages(array $params = [], $clear = TRUE, $save_operation = FALSE): bool {
    $entity_type = 'openstack_image';
    $lock_name = $this->getLockKey($entity_type);
    if (!$this->lock->acquire($lock_name)) {
      return FALSE;
    }

    $updated = $this->updateImageEntities($params, $clear, $entity_type, $this->cloudContext);

    $this->lock->release($lock_name);
    return $updated;
  }

  /**
   * Call the API for updated entities and store them as Image entities.
   *
   * @param array $params
   *   Optional parameters array.
   * @param bool $clear
   *   TRUE to clear stale security groups.
   * @param string $entity_type
   *   Entity type string.
   * @param string $cloud_context
   *   Cloud context string.
   *
   * @return bool
   *   Indicates success so failure.
   */
  public function updateImageEntities(array $params = [], $clear = TRUE, $entity_type = '', $cloud_context = ''): bool {
    if ($this->isWorkerResource()) {
      return TRUE;
    }

    $updated = FALSE;

    // Load all entities by $this->cloudContext.
    $image_entities = $this->entityTypeManager->getStorage($entity_type)->loadByProperties(
      ['cloud_context' => $this->cloudContext]
    );

    $result = $this->describeImages($params);

    // Keep isset(), and do not replace !empty() here. We need to check for
    // instance testcase notice because the instance testcase throws an error
    // of undefined index notice.
    if (isset($result)) {

      $stale = [];
      // Make it easier to lookup the images by setting up
      // the array with the image_id.
      foreach ($image_entities ?: [] as $image) {
        $stale[$image->getImageId()] = $image;
      }

      /** @var \Drupal\Core\Batch\BatchBuilder $batch_builder */
      $batch_builder = $this->initBatch('Image Update');

      if (!empty($result['Images'])) {
        foreach ($result['Images'] ?: [] as $image) {
          // Keep track of images that do not exist anymore
          // delete them after saving the rest of the images.
          if (!empty($stale[$image['ImageId']])) {
            unset($stale[$image['ImageId']]);
          }

          $batch_builder->addOperation([
            OpenStackBatchOperations::class,
            'updateImage',
          ], [$cloud_context, $image]);
        }
      }

      $batch_builder->addOperation([
        OpenStackBatchOperations::class,
        'finished',
      ], [$entity_type, $stale, $clear]);
      $this->runBatch($batch_builder);
      $updated = TRUE;
    }
    return $updated;
  }

  /**
   * Call the API for updated entities and store them as key pair entities.
   *
   * @param string $entity_type
   *   Entity type string.
   * @param string $cloud_context
   *   Cloud context string.
   * @param array $params
   *   Key Name info.
   * @param bool $update
   *   TRUE to update Key info.
   *
   * @return bool
   *   Indicates success so failure.
   */
  public function updateKeyPairEntities($entity_type = '', $cloud_context = '', array $params = [], $update = TRUE): bool {
    if ($this->isWorkerResource()) {
      return TRUE;
    }

    $updated = FALSE;
    $result = $this->describeKeyPairs();

    if (!empty($result)) {

      $all_keys = $this->cloudService->loadAllEntities($this->cloudContext, $entity_type);
      $stale = [];
      // Make it easier to lookup the groups by setting up
      // the array with the group_id.
      foreach ($all_keys ?: [] as $key) {
        $stale[$key->getKeyPairName()] = $key;
      }
      /** @var \Drupal\Core\Batch\BatchBuilder $batch_builder */
      $batch_builder = $this->initBatch('Keypair Update');

      foreach ($result['KeyPairs'] ?: [] as $key_pair) {
        // Keep track of key pair that do not exist anymore
        // delete them after saving the rest of the key pair.
        if (!empty($stale[$key_pair['KeyName']])) {
          unset($stale[$key_pair['KeyName']]);
        }
        $batch_builder->addOperation([
          OpenStackBatchOperations::class,
          'updateKeyPair',
        ], [$cloud_context, $key_pair]);
      }

      $batch_builder->addOperation([
        OpenStackBatchOperations::class,
        'finished',
      ], [$entity_type, $stale, TRUE]);
      $this->runBatch($batch_builder);
      $updated = TRUE;
    }

    return $updated;
  }

  /**
   * Update the OpenStack key pairs.
   *
   * @return bool
   *   Indicates success so failure.
   */
  public function updateKeyPairs(): bool {
    $entity_type = 'openstack_key_pair';
    $lock_name = $this->getLockKey($entity_type);

    if (!$this->lock->acquire($lock_name)) {
      return FALSE;
    }

    $updated = $this->updateKeyPairEntities($entity_type, $this->cloudContext);
    $this->lock->release($lock_name);
    return $updated;
  }

  /**
   * Call the API for updated entities and store them as volume entities.
   *
   * @param string $entity_type
   *   Entity type string.
   * @param string $cloud_context
   *   Cloud context string.
   * @param array $params
   *   Volume info.
   * @param bool $update
   *   TRUE to update volume.
   *
   * @return bool
   *   Indicates success so failure.
   */
  public function updateVolumeEntities($entity_type = '', $cloud_context = '', array $params = [], $update = TRUE): bool {
    if ($this->isWorkerResource()) {
      return TRUE;
    }

    $updated = FALSE;
    $result = $this->describeVolumes();

    $properties = empty($params['VolumeIds'][0])
      ? ['cloud_context' => $cloud_context]
      : [
        'cloud_context' => $cloud_context,
        'volume_id' => $params['VolumeIds'][0],
      ];
    $all_volumes = $this->cloudService->loadAllEntities($this->cloudContext, $entity_type, $properties);
    $stale = [];
    // Make it easier to lookup the groups by setting up
    // the array with the group_id.
    foreach ($all_volumes ?: [] as $volume) {
      $stale[$volume->getVolumeId()] = $volume;
    }
    $snapshot_id_name_map = $this->getSnapshotIdNameMap(!empty($result['Volumes']) ? $result['Volumes'] : []);

    /** @var \Drupal\Core\Batch\BatchBuilder $batch_builder */
    $batch_builder = $this->initBatch('Volume Update');

    if (!empty($result)) {
      foreach ($result['Volumes'] ?: [] as $volume) {
        // Keep track of network interfaces that do not exist anymore
        // delete them after saving the rest of the network interfaces.
        if (!empty($stale[$volume['VolumeId']])) {
          unset($stale[$volume['VolumeId']]);
        }
        if ($update) {
          $batch_builder->addOperation([
            OpenStackBatchOperations::class,
            'updateVolume',
          ], [$cloud_context, $volume, $snapshot_id_name_map]);
        }
      }

      $batch_builder->addOperation([
        OpenStackBatchOperations::class,
        'finished',
      ], [$entity_type, $stale, TRUE]);
      $this->runBatch($batch_builder);
      $updated = empty($params['VolumeIds'][0]) ? TRUE : !empty($result['Volumes']);
    }
    elseif (empty($result) && !empty($params['VolumeIds'][0])) {
      $batch_builder->addOperation([
        OpenStackBatchOperations::class,
        'finished',
      ], [$entity_type, $stale, TRUE]);
      $this->runBatch($batch_builder);
    }
    return $updated;
  }

  /**
   * Update the volumes.
   *
   * @return bool
   *   Indicates success so failure.
   */
  public function updateVolumes(): bool {
    $entity_type = 'openstack_volume';
    $lock_name = $this->getLockKey($entity_type);

    if (!$this->lock->acquire($lock_name)) {
      return FALSE;
    }

    $updated = $this->updateVolumeEntities($entity_type, $this->cloudContext);

    $this->lock->release($lock_name);
    return $updated;
  }

  /**
   * {@inheritdoc}
   */
  public function updateAvailabilityZonesWithoutBatch(array $params = [], $clear = TRUE, $update = TRUE) {
    if ($this->isWorkerResource()) {
      return TRUE;
    }

    $updated = FALSE;
    $entity_type = 'openstack_availability_zone';
    $lock_name = $this->getLockKey($entity_type);

    if (!$this->lock->acquire($lock_name)) {
      return FALSE;
    }

    $properties = ['cloud_context' => $this->cloudContext];
    $all_availability_zones = $this->cloudService->loadAllEntities($this->cloudContext, $entity_type, $properties);
    $stale = [];
    // Make it easier to lookup the snapshot by setting up
    // the array with the zone_name.
    foreach ($all_availability_zones ?: [] as $availability_zone) {
      $stale[$availability_zone->getZoneName()] = $availability_zone;
    }

    $result = $this->describeAvailabilityZones($params);

    if (!empty($result)) {
      foreach ($result['AvailabilityZones'] ?: [] as $availability_zone) {
        // Keep track of snapshot that do not exist anymore
        // delete them after saving the rest of the snapshots.
        if (!empty($stale[$availability_zone['ZoneName']])) {
          unset($stale[$availability_zone['ZoneName']]);
        }
        if ($update) {
          OpenStackBatchOperations::updateAvailabilityZone($this->cloudContext, $availability_zone);
        }
      }

      if (count($stale) && $clear === TRUE) {
        $this->entityTypeManager->getStorage($entity_type)->delete($stale);
      }

      $updated = TRUE;
    }
    elseif (empty($result) && $clear) {
      $this->entityTypeManager->getStorage($entity_type)->delete($stale);
    }

    $this->lock->release($lock_name);
    return $updated;
  }

  /**
   * {@inheritdoc}
   */
  public function updateFlavorsWithoutBatch(array $params = [], $clear = TRUE, $update = TRUE) {
    if ($this->isWorkerResource()) {
      return TRUE;
    }

    $updated = FALSE;
    $entity_type = 'openstack_flavor';
    $lock_name = $this->getLockKey($entity_type);

    if (!$this->lock->acquire($lock_name)) {
      return FALSE;
    }

    $properties = ['cloud_context' => $this->cloudContext];
    $all_flavors = $this->cloudService->loadAllEntities($this->cloudContext, $entity_type, $properties);
    $stale = [];
    // Make it easier to lookup the flavor by setting up
    // the array with the flavor id.
    foreach ($all_flavors ?: [] as $flavor) {
      $stale[$flavor->getFlavorId()] = $flavor;
    }

    $result = $this->describeFlavors($params);

    if (!empty($result)) {
      foreach ($result ?? [] as $flavor) {
        // Keep track of snapshot that do not exist anymore
        // delete them after saving the rest of the snapshots.
        if (!empty($stale[$flavor['id']])) {
          unset($stale[$flavor['id']]);
        }
        if ($update) {
          OpenStackBatchOperations::updateFlavor($this->cloudContext, $flavor);
        }
      }

      if (count($stale) && $clear === TRUE) {
        $this->entityTypeManager->getStorage($entity_type)->delete($stale);
      }

      $updated = TRUE;
    }
    elseif (empty($result) && $clear) {
      $this->entityTypeManager->getStorage($entity_type)->delete($stale);
    }

    $this->lock->release($lock_name);
    return $updated;
  }

  /**
   * Get snapshot id and name.
   *
   * @param array $volumes
   *   The volumes array.
   *
   * @return array
   *   Return map array.
   */
  public function getSnapshotIdNameMap(array $volumes = []) {
    $snapshot_ids = array_filter(array_column($volumes, 'SnapshotId'));
    if (empty($snapshot_ids)) {
      return [];
    }

    $map = [];
    foreach ($snapshot_ids ?: [] as $snapshot_id) {
      $map[$snapshot_id] = $snapshot_id;
    }

    $result = $this->describeSnapshots();

    if (!empty($result)) {
      foreach ($result['Snapshots'] ?: [] as $snapshot) {
        $snapshot_id = $snapshot['SnapshotId'];
        if (!array_key_exists($snapshot_id, $map)) {
          continue;
        }

        $map[$snapshot_id] = $this->getTagName($snapshot, '');
      }
    }

    return $map;
  }

  /**
   * Call the API for updated entities and store them as snapshot entities.
   *
   * @param string $entity_type
   *   Entity type string.
   * @param string $cloud_context
   *   Cloud context string.
   * @param array $params
   *   Snapshot IDs.
   * @param bool $update
   *   TRUE to update info.
   *
   * @return bool
   *   Indicates success so failure.
   */
  public function updateSnapshotEntities($entity_type = '', $cloud_context = '', array $params = [], $update = TRUE): bool {
    if ($this->isWorkerResource()) {
      return TRUE;
    }

    $updated = FALSE;

    $result = $this->describeSnapshots();

    if (!empty($result)) {

      $all_snapshots = $this->cloudService->loadAllEntities($this->cloudContext, $entity_type);
      $stale = [];
      // Make it easier to lookup the snapshot by setting up
      // the array with the snapshot_id.
      foreach ($all_snapshots ?: [] as $snapshot) {
        $stale[$snapshot->getSnapshotId()] = $snapshot;
      }
      /** @var \Drupal\Core\Batch\BatchBuilder $batch_builder */
      $batch_builder = $this->initBatch('Snapshot Update');

      foreach ($result['Snapshots'] ?: [] as $snapshot) {
        // Keep track of snapshot that do not exist anymore
        // delete them after saving the rest of the snapshots.
        if (!empty($stale[$snapshot['SnapshotId']])) {
          unset($stale[$snapshot['SnapshotId']]);
        }

        $batch_builder->addOperation([
          OpenStackBatchOperations::class,
          'updateSnapshot',
        ], [$cloud_context, $snapshot]);
      }

      $batch_builder->addOperation([
        OpenStackBatchOperations::class,
        'finished',
      ], [$entity_type, $stale, TRUE]);
      $this->runBatch($batch_builder);
      $updated = TRUE;
    }

    return $updated;
  }

  /**
   * Update the snapshots.
   *
   * @return bool
   *   Indicates success so failure.
   */
  public function updateSnapshots(): bool {
    $entity_type = 'openstack_snapshot';
    $lock_name = $this->getLockKey($entity_type);

    if (!$this->lock->acquire($lock_name)) {
      return FALSE;
    }

    $updated = $this->updateSnapshotEntities($entity_type, $this->cloudContext);

    $this->lock->release($lock_name);
    return $updated;
  }

  /**
   * Update the OpenStackSecurityGroups.
   *
   * @param array $params
   *   Optional parameters array.
   * @param bool $clear
   *   TRUE to clear stale security groups.
   *
   * @return bool
   *   indicates success so failure.
   */
  public function updateSecurityGroups(array $params = [], $clear = TRUE): bool {
    $entity_type = 'openstack_security_group';
    $lock_name = $this->getLockKey($entity_type);

    if (!$this->lock->acquire($lock_name)) {
      return FALSE;
    }

    $updated = $this->updateSecurityGroupEntities($params, $clear, $entity_type, $this->cloudContext);

    $this->lock->release($lock_name);
    return $updated;
  }

  /**
   * Call the API for updated entities and store them as SecurityGroup entities.
   *
   * @param array $params
   *   Optional parameters array.
   * @param bool $clear
   *   TRUE to clear stale security groups.
   * @param string $entity_type
   *   Entity type string.
   * @param string $cloud_context
   *   Cloud context string.
   * @param bool $update
   *   TRUE to update SG info.
   *
   * @return bool
   *   indicates success so failure.
   */
  public function updateSecurityGroupEntities(array $params = [], $clear = TRUE, $entity_type = '', $cloud_context = '', $update = TRUE): bool {
    if ($this->isWorkerResource()) {
      return TRUE;
    }

    $cloud_configs = $this->entityTypeManager
      ->getStorage('cloud_config')
      ->loadByProperties([
        'type' => 'openstack',
        'cloud_context' => $cloud_context,
      ]);

    /** @var \Drupal\cloud\Entity\CloudConfig $cloud_config */
    $cloud_config = array_shift($cloud_configs);

    $params['project_id'] = $cloud_config->get('field_project_id')->value ?? NULL;

    $updated = FALSE;
    $result = $this->describeSecurityGroups($params);

    if (!empty($result)) {

      $all_groups = $this->cloudService->loadAllEntities($this->cloudContext, $entity_type);
      $stale = [];
      // Make it easier to lookup the groups by setting up
      // the array with the group_id.
      foreach ($all_groups ?: [] as $group) {
        $stale[$group->getGroupId()] = $group;
      }

      /** @var \Drupal\Core\Batch\BatchBuilder $batch_builder */
      $batch_builder = $this->initBatch('Security group update');

      foreach ($result['SecurityGroups'] ?: [] as $security_group) {

        // Keep track of instances that do not exist anymore
        // delete them after saving the rest of the instances.
        if (!empty($stale[$security_group['GroupId']])) {
          unset($stale[$security_group['GroupId']]);
        }
        $batch_builder->addOperation([
          OpenStackBatchOperations::class,
          'updateSecurityGroup',
        ], [$cloud_context, $security_group]);
      }

      $batch_builder->addOperation([
        OpenStackBatchOperations::class,
        'finished',
      ], [$entity_type, $stale, $clear]);
      $this->runBatch($batch_builder);
      $updated = TRUE;
    }

    return $updated;
  }

  /**
   * Update the OpenStackNetworkInterfaces.
   *
   * @param array $params
   *   Optional parameters array.
   * @param bool $clear
   *   TRUE to clear stale security groups.
   *
   * @return bool
   *   indicates success so failure.
   */
  public function updateNetworkInterfaces(array $params = [], $clear = TRUE): bool {
    $entity_type = 'openstack_network_interface';
    $lock_name = $this->getLockKey($entity_type);

    if (!$this->lock->acquire($lock_name)) {
      return FALSE;
    }

    $updated = $this->updateNetworkInterfaceEntities($params, $clear, $entity_type, $this->cloudContext);

    $this->lock->release($lock_name);
    return $updated;
  }

  /**
   * Call API for updated entities and store them as NetworkInterface entities.
   *
   * @param array $params
   *   Optional parameters array.
   * @param bool $clear
   *   TRUE to clear stale security groups.
   * @param string $entity_type
   *   Entity type string.
   * @param string $cloud_context
   *   Cloud context string.
   * @param bool $update
   *   TRUE to update info.
   *
   * @return bool
   *   indicates success so failure.
   */
  public function updateNetworkInterfaceEntities(array $params = [], $clear = TRUE, $entity_type = '', $cloud_context = '', $update = TRUE): bool {
    if ($this->isWorkerResource()) {
      return TRUE;
    }

    $cloud_configs = $this->entityTypeManager
      ->getStorage('cloud_config')
      ->loadByProperties([
        'type' => 'openstack',
        'cloud_context' => $cloud_context,
      ]);

    /** @var \Drupal\cloud\Entity\CloudConfig $cloud_config */
    $cloud_config = array_shift($cloud_configs);

    $params['project_id'] = $cloud_config->get('field_project_id')->value ?? NULL;

    $updated = FALSE;
    $result = $this->describeNetworkInterfaces($params);

    if (!empty($result)) {

      $all_interfaces = $this->cloudService->loadAllEntities($this->cloudContext, $entity_type);
      $stale = [];
      // Make it easier to lookup the groups by setting up
      // the array with the group_id.
      foreach ($all_interfaces ?: [] as $interface) {
        $stale[$interface->getNetworkInterfaceId()] = $interface;
      }

      /** @var \Drupal\Core\Batch\BatchBuilder $batch_builder */
      $batch_builder = $this->initBatch('Network interface update');

      foreach ($result['NetworkInterfaces'] ?: [] as $network_interface) {
        // Keep track of network interfaces that do not exist anymore
        // delete them after saving the rest of the network interfaces.
        if (!empty($stale[$network_interface['NetworkInterfaceId']])) {
          unset($stale[$network_interface['NetworkInterfaceId']]);
        }
        $batch_builder->addOperation([
          OpenStackBatchOperations::class,
          'updateNetworkInterface',
        ], [$cloud_context, $network_interface]);
      }

      $batch_builder->addOperation([
        OpenStackBatchOperations::class,
        'finished',
      ], [$entity_type, $stale, $clear]);
      $this->runBatch($batch_builder);
      $updated = TRUE;
    }
    return $updated;
  }

  /**
   * Create network.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The network interface create API response.
   */
  public function createNetwork(array $params = []): array {
    if ($this->isWorkerResource()) {
      $this->saveOperation(__FUNCTION__, [$params], 'updateNetworkInterfaces');

      return ['SendToWorker' => TRUE];
    }

    $this->loadCredentials();
    $project_id = $this->credentials['project_id'];

    $body['network'] = [
      'name' => $params['Name'],
      'admin_state_up' => $params['AdminStateUp'],
      'shared' => $params['Shared'],
    ];
    if (!empty($params['AvailabilityZones'])) {
      $body['network']['availability_zone_hints'] = $params['AvailabilityZones'];
    }

    $credentials = [
      'http_method' => 'post',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::NETWORK_ENDPOINT, "/networks?project_id={$project_id}"),
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    $network = $response['body']['network'] ?? [];

    return [
      'Network' => [
        'NetworkId' => $network['id'] ?? NULL,
      ],
    ];
  }

  /**
   * Get network list.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The network list API response.
   */
  public function describeNetworks(array $params = []) {
    if ($this->isWorkerResource()) {
      return ['SendToWorker' => TRUE];
    }

    $project_id = $params['project_id'];

    $this->loadCredentials();
    $credentials = [
      'http_method' => 'get',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::NETWORK_ENDPOINT, '/networks'),
      'params' => $params,
    ];
    $response = $this->getClient($credentials);

    $networks = $response['body']['networks'] ?? [];

    $results['Networks'] = [];
    foreach ($networks ?: [] as $network) {
      if ($network['project_id'] !== $project_id && empty($network['shared'])) {
        continue;
      }

      $subnets = [];
      foreach ($network['subnets'] ?: [] as $subnet_id) {
        $subnet_result = $this->describeSubnetDetail(['SubnetId' => $subnet_id]);
        $subnets[] = $subnet_result['name'];
      }

      $results['Networks'][] = [
        'NetworkId'          => $network['id'] ?? NULL,
        'Name'               => $network['name'] ?? NULL,
        'ProjectId'          => $network['project_id'],
        'Status'             => $network['status'] === 'ACTIVE' ? 'available' : '',
        'AdminStateUp'       => $network['admin_state_up'] ?? FALSE,
        'SubnetId'           => $subnet_names ?? NULL,
        'TagSet'             => empty($network['tags']) ? [] : $network['tags'],
        'External'           => $network['router:external'],
        'Shared'             => $network['shared'],
        'Mtu'                => $network['mtu'],
        'NetworkType'        => $network['provider:network_type'],
        'PhysicalNetwork'    => $network['provider:physical_network'],
        'SegmentationId'     => $network['provider:segmentation_id'],
        'Subnets'            => $subnets,
        'AvailabilityZones'  => $network['availability_zones'],
      ];
    }

    return $results;
  }

  /**
   * Delete network.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array|null
   *   Succeeded or failed.
   */
  public function deleteNetwork(array $params = []): ?array {
    if ($this->isWorkerResource()) {
      $this->saveOperation(__FUNCTION__, [$params], 'updateNetworks');

      return ['SendToWorker' => TRUE];
    }

    $this->loadCredentials();

    $credentials = [
      'http_method' => 'delete',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::NETWORK_ENDPOINT, "/networks/{$params['NetworkId']}"),
    ];

    $response = $this->getClient($credentials);

    return $response['status_code'] === 204 ? ['Success' => TRUE] : NULL;
  }

  /**
   * Update network.
   *
   * @param array $params
   *   Parameters to pass to the API.
   *
   * @return array
   *   The network interface update API response.
   */
  public function updateNetwork(array $params = []): array {
    if ($this->isWorkerResource()) {
      $this->saveOperation(__FUNCTION__, [$params], 'updateNetworks');

      return ['SendToWorker' => TRUE];
    }

    $this->loadCredentials();
    $project_id = $this->credentials['project_id'];

    $body['network'] = [
      'name' => $params['Name'],
      'admin_state_up' => $params['AdminStateUp'],
      'shared' => $params['Shared'],
    ];

    $credentials = [
      'http_method' => 'put',
      'endpoint' => $this->getEndpoint(OpenStackServiceInterface::NETWORK_ENDPOINT, "/networks/{$params['NetworkId']}?project_id={$project_id}"),
      'params' => $body,
    ];

    $response = $this->getClient($credentials);

    $network = $response['body']['network'] ?? [];

    return [
      'Network' => [
        'Name' => $network['name'] ?? NULL,
        'NetworkId' => $network['network_id'] ?? NULL,
      ],
    ];
  }

  /**
   * Update the OpenStackNetworks.
   *
   * @param array $params
   *   Optional parameters array.
   * @param bool $clear
   *   TRUE to clear stale security groups.
   *
   * @return bool
   *   indicates success so failure.
   */
  public function updateNetworks(array $params = [], $clear = TRUE): bool {
    $entity_type = 'openstack_network';
    $lock_name = $this->getLockKey($entity_type);

    if (!$this->lock->acquire($lock_name)) {
      return FALSE;
    }

    $updated = $this->updateNetworkEntities($params, $clear, $entity_type, $this->cloudContext);

    $this->lock->release($lock_name);
    return $updated;
  }

  /**
   * Call API for updated entities and store them as Network entities.
   *
   * @param array $params
   *   Optional parameters array.
   * @param bool $clear
   *   TRUE to clear stale security groups.
   * @param string $entity_type
   *   Entity type string.
   * @param string $cloud_context
   *   Cloud context string.
   * @param bool $update
   *   TRUE to update info.
   *
   * @return bool
   *   indicates success so failure.
   */
  public function updateNetworkEntities(array $params = [], $clear = TRUE, $entity_type = '', $cloud_context = '', $update = TRUE): bool {
    if ($this->isWorkerResource()) {
      return TRUE;
    }

    $cloud_configs = $this->entityTypeManager
      ->getStorage('cloud_config')
      ->loadByProperties([
        'type' => 'openstack',
        'cloud_context' => $cloud_context,
      ]);

    /** @var \Drupal\cloud\Entity\CloudConfig $cloud_config */
    $cloud_config = array_shift($cloud_configs);

    $params['project_id'] = $cloud_config->get('field_project_id')->value ?? NULL;

    $updated = FALSE;
    $result = $this->describeNetworks($params);
    if (!empty($result)) {

      $all_networks = $this->cloudService->loadAllEntities($this->cloudContext, $entity_type);
      $stale = [];

      foreach ($all_networks ?: [] as $network) {
        $stale[$network->getNetworkId()] = $network;
      }

      /** @var \Drupal\Core\Batch\BatchBuilder $batch_builder */
      $batch_builder = $this->initBatch('Network update');

      foreach ($result['Networks'] ?: [] as $network) {
        // Keep track of networks that do not exist anymore
        // delete them after saving the rest of the networks.
        if (!empty($stale[$network['NetworkId']])) {
          unset($stale[$network['NetworkId']]);
        }
        $batch_builder->addOperation([
          OpenStackBatchOperations::class,
          'updateNetwork',
        ], [$cloud_context, $network]);
      }

      $batch_builder->addOperation([
        OpenStackBatchOperations::class,
        'finished',
      ], [$entity_type, $stale, $clear]);
      $this->runBatch($batch_builder);
      $updated = TRUE;
    }
    return $updated;
  }

  /**
   * Update the OpenStackFloatingIps.
   *
   * @return bool
   *   indicates success so failure.
   */
  public function updateFloatingIps(): bool {
    $entity_type = 'openstack_floating_ip';
    $lock_name = $this->getLockKey($entity_type);

    if (!$this->lock->acquire($lock_name)) {
      return FALSE;
    }

    $updated = $this->updateFloatingIpEntities($entity_type, $this->cloudContext);

    $this->lock->release($lock_name);
    return $updated;
  }

  /**
   * Call the API for updated entities and store them as FloatingIp entities.
   *
   * @param string $entity_type
   *   Entity type string.
   * @param string $cloud_context
   *   Cloud context string.
   *
   * @return bool
   *   indicates success so failure.
   */
  public function updateFloatingIpEntities($entity_type = '', $cloud_context = ''): bool {
    if ($this->isWorkerResource()) {
      return TRUE;
    }

    $updated = FALSE;
    $result = $this->describeAddresses();

    if (!empty($result)) {

      $all_ips = $this->cloudService->loadAllEntities($this->cloudContext, $entity_type);
      $stale = [];
      // Make it easier to lookup the groups by setting up
      // the array with the group_id.
      foreach ($all_ips ?: [] as $ip) {
        $stale[$ip->getPublicIp()] = $ip;
      }

      /** @var \Drupal\Core\Batch\BatchBuilder $batch_builder */
      $batch_builder = $this->initBatch('FloatingIp Update');

      foreach ($result['Addresses'] ?: [] as $floating_ip) {
        // Keep track of IPs that do not exist anymore
        // delete them after saving the rest of the IPs.
        if (!empty($stale[$floating_ip['PublicIp']])) {
          unset($stale[$floating_ip['PublicIp']]);
        }
        $batch_builder->addOperation([
          OpenStackBatchOperations::class,
          'updateFloatingIp',
        ], [$cloud_context, $floating_ip]);
      }

      $batch_builder->addOperation([
        OpenStackBatchOperations::class,
        'finished',
      ], [$entity_type, $stale, TRUE]);
      $this->runBatch($batch_builder);
      $updated = TRUE;
    }

    return $updated;
  }

  /**
   * {@inheritdoc}
   */
  public function getAvailabilityZones($with_id = FALSE, array $credentials = []): array {
    $zones = [];
    $result = $this->describeAvailabilityZones([], $credentials);
    if (!empty($result)) {
      foreach ($result['AvailabilityZones'] ?: [] as $zone_info) {
        $zones[$zone_info['ZoneName']] = $zone_info['ZoneName'];
      }
    }
    return $zones;
  }

  /**
   * Helper method to do a Guzzle HTTP Request.
   *
   * @param string $http_method
   *   The http method in lower case.
   * @param string $endpoint
   *   The endpoint to access.
   * @param array $params
   *   Parameters to pass to the request.
   *
   * @return array
   *   Array interpretation of the response body.
   */
  private function request($http_method, $endpoint, array $params = []): array {
    $output = [];

    try {
      $response = $this->httpClient->$http_method($endpoint, $params);

      if (!empty($response)) {
        $status_code = $response->getStatusCode();
        $headers = $response->getHeaders();
        $output['auth_token'] = $headers['X-Subject-Token'][0] ?? '';
        $output['status_code'] = $status_code;
        $body = $response->getBody()->getContents();
      }
    }
    catch (GuzzleException $error) {
      /** @var \GuzzleHttp\Exception\RequestException $error */
      $response = $error->getResponse();
      $error_content = [];
      $content = '';
      if (!empty($response)) {
        $content = $response->getBody()->getContents();
        $status_code = $response->getStatusCode();
        $output['status_code'] = $status_code;
        $error_content = json_decode($content, TRUE);
      }

      $error_content = !empty($error_content) ? reset($error_content) : [];

      if (!empty($error_content)) {

        // If OpenStackServiceInterface::getDefaultVolumeType() returns 404,
        // we do not recognize it as the error.
        if ($error_content['message'] === 'Default volume type can not be found.'
          && strpos($endpoint, '/types/default')) {
          return [];
        }

        $this->messenger->addError($this->t('Status Code: @status_code', [
          '@status_code' => $error_content['code'] ?? '',
        ]));

        $this->messenger->addError($this->t('Error Message: @msg', [
          '@msg' => $error_content['message'],
        ]));
      }
      else {
        $this->messenger->addError($this->t('OpenStack API error. Error details are as follows: @response', [
          '@response' => $error->getMessage(),
        ]));
      }

      $message = new FormattableMarkup(
        'OpenStack API error for @endpoint. Error details are as follows:<pre>@response</pre>', [
          '@endpoint' => $endpoint,
          '@response' => !empty($content)
            ? $content
            : $error->getMessage(),
        ]
      );
      $this->logError('Remote API Connection', $error, $message, FALSE);
    }
    catch (\Exception $error) {
      $this->messenger->addError($this->t('An unknown error occurred while trying to connect to the API endpoint at: %endpoint (HTTP %http_method method). This is not a Guzzle error, not an error in the API endpoint, rather a generic local error occurred: "%error"', [
        '%endpoint' => $endpoint,
        '%error' => $error->getMessage(),
        '%http_method' => $http_method,
      ]));

      $this->logError('openstack_rest', $error, $this->t('An unknown error occurred while trying to connect to the API endpoint at: %endpoint (HTTP %http_method method). This is not a Guzzle error, not an error in the API endpoint, rather a generic local error occurred: "%error".', [
        '%endpoint' => $endpoint,
        '%error' => $error->getMessage(),
        '%http_method' => $http_method,
      ]
      ), FALSE);
    }
    $body = empty($body) ? [] : json_decode($body, TRUE);
    $output['body'] = $body;
    return $output;
  }

  /**
   * Initialize a new batch builder.
   *
   * @param string $batch_name
   *   The batch name.
   *
   * @return \Drupal\Core\Batch\BatchBuilder
   *   The initialized batch object.
   */
  protected function initBatch($batch_name) {
    return (new BatchBuilder())
      ->setTitle($batch_name);
  }

  /**
   * Run the batch job to process entities.
   *
   * @param \Drupal\Core\Batch\BatchBuilder $batch_builder
   *   The batch builder object.
   */
  protected function runBatch(BatchBuilder $batch_builder) {
    try {
      // Log the start time.
      $start = $this->getTimestamp();
      $batch_array = $batch_builder->toArray();
      batch_set($batch_array);
      // Reset the progressive so batch works with out a web head.
      $batch = &batch_get();
      $batch['progressive'] = FALSE;
      // Check if using Drush to process the queue.
      // using drush_backend_batch_process() will fix the "Route not found"
      // error.
      if (PHP_SAPI === 'cli' && function_exists('drush_backend_batch_process')) {
        drush_backend_batch_process();
      }
      else {
        batch_process();
      }
      // Log the end time.
      $end = $this->getTimestamp();
      $this->logger('openstack_rest')->info($this->t('@updater - @cloud_context: Batch operation took @time seconds.', [
        '@cloud_context' => $this->cloudContext,
        '@updater' => $batch_array['title'],
        '@time' => $end - $start,
      ]));
    }
    catch (\Exception $e) {
      $this->handleException($e);
    }
    finally {
      // Reset the batch otherwise this operation hangs when using Drush.
      // https://www.drupal.org/project/drupal/issues/3166042
      $batch = [];
    }
  }

  /**
   * Log errors to watchdog and throw an exception.
   *
   * @param string $type
   *   The error type.
   * @param \Exception $exception
   *   The exception object.
   * @param string $message
   *   The message to log.
   * @param bool $throw
   *   TRUE to throw exception.
   *
   * @throws \Drupal\openstack\Service\Rest\OpenStackServiceException
   */
  private function logError($type, \Exception $exception, $message, $throw = TRUE) {
    watchdog_exception($type, $exception, $message);
    if ($throw === TRUE) {
      throw new OpenStackServiceException($message);
    }
  }

  /**
   * Get Tag name.
   *
   * @param array $openstack_obj
   *   The openstack object.
   * @param string $default_value
   *   The default value.
   *
   * @return string
   *   Return name of corresponding resource.
   */
  public function getTagName(array $openstack_obj, $default_value) {
    $name = $default_value;

    if (!empty($openstack_obj['name'])) {
      return $openstack_obj['name'];
    }

    $tags = empty($openstack_obj['Tags']) ? [] : $openstack_obj['Tags'];
    foreach ($tags as $tag) {
      if (strpos($tag, 'Name__') === 0) {
        $name = substr($tag, strlen('Name__'));
        break;
      }
    }
    return $name;
  }

  /**
   * Setup IP Permissions.
   *
   * @param object $security_group
   *   The security group entity.
   * @param string $field
   *   Field to used for lookup.
   * @param array $ec2_permissions
   *   Permissions array from Ec2.
   */
  public function setupIpPermissions(&$security_group, $field, array $ec2_permissions) {
    // Permissions are always overwritten with the latest from
    // EC2.  The reason is that there is no way to guarantee a 1
    // to 1 mapping from the $security_group['IpPermissions'] array.
    // There is no IP permission ID coming back from EC2.
    // Clear out all items before re-adding them.
    $i = $security_group->$field->count() - 1;

    while ($i >= 0) {
      if ($security_group->$field->get($i)) {
        $security_group->$field->removeItem($i);
      }
      $i--;
    }

    // Setup all permission objects.
    $count = 0;
    foreach ($ec2_permissions ?: [] as $permissions) {
      $permission_objects = $this->setupIpPermissionObject($permissions);
      // Loop through the permission objects and add them to the
      // security group ip_permission field.
      foreach ($permission_objects ?: [] as $permission) {
        $security_group->$field->set($count, $permission);
        $count++;
      }
    }
  }

  /**
   * Setup the ip_permission field given the inbound security group array.
   *
   * The array comes from DescribeSecurityGroup Amazon EC2 API call.
   *
   * @param array $ec2_permission
   *   An array object of EC2permission.
   *
   * @return array
   *   An array of \Drupal\Core\Field\FieldItemInterface.
   */
  public function setupIpPermissionObject(array $ec2_permission) {
    $ip_permissions = [];

    // $file_path = \Drupal::service('file_system')->realpath("public://");
    // $handle = fopen($file_path . "/myfile.txt", "a+");
    // fwrite($handle, json_encode($ec2_permission));
    // fclose($handle);
    // Get the field definition for an IpPermission object.
    $definition = $this->entityFieldManager->getBaseFieldDefinitions('openstack_security_group');

    /* Setup the more global attributes.
     * NOTE: DO NOT change from isset() to !empty() here since
     * !empty($ec2_permission['FromPort']) is considered empty
     * in PHP when $ec2_permission['FromPort'] is a zero (0) value.
     */
    $rule_id = $ec2_permission['RuleId'] ?? NULL;
    $from_port = (string) ($ec2_permission['FromPort']);
    $to_port = (string) ($ec2_permission['ToPort']);

    $ip_protocol = $ec2_permission['IpProtocol'];

    // To keep things consistent, if ip_protocol is -1,
    // set from_port and to_port as 1-65535.
    if ($ip_protocol === '-1') {
      $from_port = '1';
      $to_port = '65535';
    }

    if (!empty($ec2_permission['IpRanges'])) {
      // Create a IPv4 permission object.
      foreach ($ec2_permission['IpRanges'] ?: [] as $ip_range) {
        $ip_range_permission = $this->fieldTypePluginManager->createInstance('ip_permission', [
          'field_definition' => $definition['ip_permission'],
          'parent' => NULL,
          'name' => NULL,
        ]);
        // Source is an internal identifier.  Does not come from EC2.
        $ip_range_permission->source = 'ip4';
        $ip_range_permission->cidr_ip = $ip_range['CidrIp'];
        $ip_range_permission->description = $ip_range['Description'] ?? NULL;

        $ip_range_permission->rule_id = $rule_id;
        $ip_range_permission->from_port = $from_port;
        $ip_range_permission->to_port = $to_port;
        $ip_range_permission->ip_protocol = $ip_protocol;

        $ip_permissions[] = $ip_range_permission;
      }
    }

    if (!empty($ec2_permission['Ipv6Ranges'])) {
      // Create IPv6 permissions object.
      foreach ($ec2_permission['Ipv6Ranges'] ?: [] as $ip_range) {
        $ip_v6_permission = $this->fieldTypePluginManager->createInstance('ip_permission', [
          'field_definition' => $definition['ip_permission'],
          'parent' => NULL,
          'name' => NULL,
        ]);
        // Source is an internal identifier.  Does not come from EC2.
        $ip_v6_permission->rule_id = $rule_id;
        $ip_v6_permission->source = 'ip6';
        $ip_v6_permission->cidr_ip_v6 = $ip_range['CidrIpv6'];
        $ip_v6_permission->description = $ip_range['Description'] ?? NULL;
        $ip_v6_permission->from_port = $from_port;
        $ip_v6_permission->to_port = $to_port;
        $ip_v6_permission->ip_protocol = $ip_protocol;
        $ip_permissions[] = $ip_v6_permission;
      }
    }

    if (!empty($ec2_permission['UserIdGroupPairs'])) {
      // Create Group permissions object.
      foreach ($ec2_permission['UserIdGroupPairs'] ?: [] as $group) {
        $group_permission = $this->fieldTypePluginManager->createInstance('ip_permission', [
          'field_definition' => $definition['ip_permission'],
          'parent' => NULL,
          'name' => NULL,
        ]);
        // Source is an internal identifier.  Does not come from EC2.
        $group_permission->rule_id = $rule_id;
        $group_permission->source = 'group';
        $group_permission->group_id = $group['GroupId'] ?? NULL;
        $group_permission->group_name = $group['GroupName'] ?? NULL;
        $group_permission->user_id = $group['UserId'] ?? NULL;
        $group_permission->peering_status = $group['PeeringStatus'] ?? NULL;
        $group_permission->vpc_id = $group['VpcId'] ?? NULL;
        $group_permission->peering_connection_id = $group['VpcPeeringConnectionId'] ?? NULL;
        $group_permission->description = $group['Description'] ?? NULL;
        $group_permission->from_port = $from_port;
        $group_permission->to_port = $to_port;
        $group_permission->ip_protocol = $ip_protocol;
        $ip_permissions[] = $group_permission;
      }
    }

    if (!empty($ec2_permission['PrefixListIds'])) {
      foreach ($ec2_permission['PrefixListIds'] ?: [] as $prefix) {
        $prefix_permission = $this->fieldTypePluginManager->createInstance('ip_permission', [
          'field_definition' => $definition['ip_permission'],
          'parent' => NULL,
          'name' => NULL,
        ]);
        // Source is an internal identifier.  Does not come from EC2.
        $prefix_permission->rule_id = $rule_id;
        $prefix_permission->source = 'prefix';
        $prefix_permission->prefix_list_id = $prefix['PrefixListId'];
        $prefix_permission->description = $prefix['Description'] ?? NULL;
        $prefix_permission->from_port = $from_port;
        $prefix_permission->to_port = $to_port;
        $prefix_permission->ip_protocol = $ip_protocol;
        $ip_permissions[] = $prefix_permission;
      }
    }

    return $ip_permissions;
  }

  /**
   * {@inheritdoc}
   */
  public function createResourceQueueItems(CloudConfigInterface $cloud_config): void {
    $queue_limit = $this->configFactory->get('openstack.settings')->get('openstack_queue_limit');
    $method_names = [
      'updateInstances',
      'updateImages',
      'updateKeyPairs',
      'updateVolumes',
      'updateSnapshots',
      'updateSecurityGroups',
      'updateFloatingIps',
      'updateNetworkInterfaces',
    ];
    self::updateResourceQueue($cloud_config, $method_names, $queue_limit);
  }

  /**
   * Get UID tag value.
   *
   * @param array $tags_array
   *   The tags array of API.
   * @param string $entity_type_id
   *   The entity type ID.
   *
   * @return int
   *   Return user id.
   */
  public function getUidTagValue(array $tags_array, string $entity_type_id): int {
    $bundle = $this->entityTypeManager->getDefinition($entity_type_id)->getProvider();

    $key = $this->getTagKeyCreatedByUid(
      $bundle,
      $this->cloudContext
    );

    $uid = 0;
    if (!empty($tags_array['Tags'])) {
      foreach ($tags_array['Tags'] ?: [] as $tag) {
        if (strpos($tag, $key . '.') === 0) {
          $uid = substr($tag, strlen($key . '.'));
          break;
        }
      }
    }
    return is_numeric($uid) ? $uid : 0;
  }

  /**
   * Get InstanceUID.
   *
   * @param string $instance_id
   *   The instance id.
   *
   * @return int
   *   Return user id.
   */
  public function getInstanceUid($instance_id) {
    $uid = 0;
    $instances = $this->entityTypeManager
      ->getStorage('openstack_instance')
      ->loadByProperties([
        'instance_id' => $instance_id,
      ]);

    if (count($instances) > 0) {
      /** @var \Drupal\aws_cloud\Entity\Ec2\InstanceInterface $instance */
      $instance = array_shift($instances);
      $uid = $instance->getOwnerId();
    }
    return $uid;
  }

  /**
   * Get InstanceUID.
   *
   * @param string $service
   *   The OpenStack service name such as 'identity', 'compute', 'image',
   *   'network', and 'volume'.
   * @param string $path
   *   The endpoint path such as `/servers/detail'.
   *
   * @return string
   *   The full endpoint.
   */
  private function getEndpoint(string $service, string $path): string {
    $credentials = $this->loadCredentials();

    return rtrim($credentials["${service}_api_endpoint"], '/')
      . '/' . trim(($credentials["${service}_version"] ?? ''), '/')
      . '/' . trim($path, '/');
  }

  /**
   * {@inheritdoc}
   */
  public function isWorkerResource(): bool {
    if (empty($this->cloudContext)) {
      return FALSE;
    }

    $cloud_config = $this->cloudConfigPluginManager->loadConfigEntity();
    return $cloud_config->isRemote();
  }

  /**
   * Save operation to queue.
   *
   * @param string $operation
   *   The operation to perform.
   * @param array $params
   *   An array of parameters.
   * @param string $update_entities_method
   *   The method name of updating entities.
   * @param array $update_entities_method_params
   *   The method parameters of updating entities.
   * @param array $refs
   *   The array of references.
   */
  private function saveOperation(string $operation, array $params, string $update_entities_method = '', array $update_entities_method_params = NULL, array $refs = []): void {
    if (empty($this->cloudContext)) {
      return;
    }

    // Skip describe operations.
    if (strpos($operation, 'Describe') === 0) {
      return;
    }

    // Skip dry run operations.
    if (!empty($params[0]['DryRun'])) {
      return;
    }

    $cloud_config = $this->cloudConfigPluginManager->loadConfigEntity();
    if (!$cloud_config->isRemote()) {
      return;
    }

    if (empty($update_entities_method)) {
      $update_entities_method = preg_replace('/^[[:lower:]]+/', 'update', lcfirst($operation));
    }

    $update_entities_method_params = [
      $update_entities_method_params ?? $params[0],
      FALSE,
    ];

    $queue = $this->queueFactory->get('operation_queue_' . $cloud_config->get('cloud_cluster_name')->value . $cloud_config->get('cloud_cluster_worker_name')->value);
    $queue->createItem([
      'cloud_context' => $this->cloudContext,
      'operation' => $operation,
      'params' => $params,
      'service_name' => 'openstack.rest',
      'update_entities_method' => $update_entities_method,
      'update_entities_method_params' => $update_entities_method_params,
      'refs' => $refs,
    ]);
  }

  /**
   * {@inheritdoc}
   */
  public function getNetworkPrivateIps($network_interface_id, $cloud_context): array {
    $this->setCloudContext($cloud_context);

    $network_interface_ports = $this->describeNetworkInterfacesPorts();

    $network_interface_list = $this->entityTypeManager
      ->getStorage('openstack_network_interface')
      ->loadByProperties(['id' => $network_interface_id]);

    /** @var \Drupal\openstack\Entity\OpenStackNetworkInterface $network_interface */
    $network_interface = count($network_interface_list) === 1
      ? array_shift($network_interface_list)
      : $network_interface_list;

    $association_id = $network_interface->getAssociationId();
    $ips = [];
    if (empty($association_id)) {
      foreach ($network_interface_ports ?: [] as $key_network_interface_id => $port_id) {
        if ($key_network_interface_id == $network_interface->getNetworkInterfaceId()) {
          $ips = $port_id;
        }
      }
    }

    return $ips;
  }

  /**
   * {@inheritdoc}
   */
  public function getTagKeyCreatedByUid(
    string $bundle,
    string $cloud_context,
    int $uid = 0
  ): string {
    $tag_name = $this->cloudService->getTagKeyCreatedByUid($bundle, $cloud_context);
    return strlen($tag_name . ".{$uid}") >= OpenStackServiceInterface::MAX_TAG_LENGTH
      ? substr($tag_name, 0, OpenStackServiceInterface::MAX_TAG_LENGTH - strlen(".{$uid}") - 1)
      : $tag_name;
  }

  /**
   * {@inheritdoc}
   */
  public function getTagValueCreatedByUid(
    string $bundle,
    string $cloud_context,
    int $uid
  ): string {
    $tag_format = '{tag}.{uid}';

    $context = [
      '{tag}' => $this->getTagKeyCreatedByUid($bundle, $cloud_context, $uid),
      '{uid}' => $uid,
    ];

    return strtr($tag_format, $context);
  }

}
